(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory LinearPolyCarrier imports SN_Order_Carrier2 LinearPolyOrder MatrixPoly 
begin

instantiation arctic :: "show" begin
fun shows_arctic :: "arctic \<Rightarrow> shows"
  where "shows_arctic(Num_arc i) = shows i"
  | "shows_arctic(MinInfty)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ai = shows_arctic ai"
definition "shows_list (as::arctic list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic: "shows_prec d (a::arctic) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_def)
lemma assoc_arctic_elt: "\<forall>(a::arctic)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic by auto
instance 
  by (intro_classes,rule assoc_arctic,unfold shows_list_arctic_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_elt])
end

instantiation arctic_rat :: "show" begin
fun shows_arctic_rat :: "arctic_rat \<Rightarrow> shows"
  where "shows_arctic_rat (Num_arc_rat i) = shows i"
  | "shows_arctic_rat(MinInfty_rat)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ari = shows_arctic_rat ari"
definition "shows_list (as::arctic_rat list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic_rat: "shows_prec d (a::arctic_rat) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_rat_def)
lemma assoc_arctic_rat_elt: "\<forall>(a::arctic_rat)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic_rat by auto
instance 
  by (intro_classes,rule assoc_arctic_rat,unfold shows_list_arctic_rat_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_rat_elt])
end

definition int_name :: "string" where "int_name = ''naturals with negative constants''"
definition rat_name :: "string" where "rat_name = ''rationals (with negative constants)''"
definition arctic_name :: "string" where "arctic_name = ''arctic semiring''"
definition arctic_rat_name :: "string" where "arctic_rat_name = ''arctic rationals''"


definition mat_mono :: "nat \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a mat \<Rightarrow> bool)"
where "mat_mono sd m \<equiv> mat_monoI m sd"

abbreviation int_components :: "int poly_components"
where "int_components \<equiv> ((\<lambda>x. True) , 0 , 1 , 1, op +, op *, max0, ge, op >, True, int_name, \<lambda>x. True, int_mono)"

abbreviation rat_components :: "rat \<Rightarrow> rat poly_components"
where "rat_components \<delta> \<equiv> ((\<lambda>x. True) , 0 , 1 , \<delta>, op +, op *, max0, ge, rat_gt \<delta>, True, rat_name, \<lambda>x. True, rat_mono)"

abbreviation arctic_components :: "arctic poly_components"
where "arctic_components \<equiv> ((\<lambda>x. True) , 0 , 1 , 1, op +, op *, id, ge, gt_arctic, False, arctic_name, pos_arctic, (\<lambda>x. False))"

abbreviation arctic_rat_components :: "rat \<Rightarrow> arctic_rat poly_components"
where "arctic_rat_components \<delta> \<equiv> ((\<lambda>x. True) , 0 , 1 , 1, op +, op *, id, ge, gt_arctic_rat \<delta>, False, arctic_rat_name, pos_arctic_rat, (\<lambda>x. False))"

abbreviation check_poly_mono_comp :: "('a :: show)poly_components \<Rightarrow> ('f :: show, 'a)lpoly_interL \<Rightarrow> shows check" 
  where "check_poly_mono_comp c \<equiv> check_poly_monoI (get_comp_0 c) (get_comp_ge c) (get_comp_check_mono c)"




locale mono_poly_interpretation = 
  fixes c :: "('a :: {show, max_ordered_semiring_1}) poly_components"
   and  check_carrier :: "shows check"
 assumes c: "c = (\<lambda>x. True, 0, 1, get_comp_def c, op + ,  op *, max0, ge, get_comp_gt c, True, get_comp_description c, \<lambda>x. True, get_comp_check_mono c)"
  and check_carrier: "isOK(check_carrier) \<Longrightarrow> mono_poly_order_carrier
  (get_comp_wf_elem c)
  (get_comp_0 c)
  (get_comp_plus c)
  (get_comp_times c)
  (get_comp_ge c)
  (get_comp_gt c)
  (get_comp_mono c)
  (get_comp_1 c)
  (get_comp_def c)
  (get_comp_max0 c)
  (get_comp_apos c)
  (get_comp_check_mono c)" 

locale poly_interpretation = 
  fixes c :: "('a :: {show, max_ordered_semiring_1}) poly_components"
   and  check_carrier :: "shows check"
 assumes c: "c = (\<lambda>x. True, 0, 1, get_comp_def c, op + ,  op *, id, ge, get_comp_gt c, False, get_comp_description c, get_comp_apos c, \<lambda>x. False)"
  and zero_leastIII: "\<And> x :: 'a. ge x 0"
  and check_carrier: "isOK(check_carrier) \<Longrightarrow> poly_order_carrier
  (get_comp_wf_elem c)
  (get_comp_0 c)
  (get_comp_plus c)
  (get_comp_times c)
  (get_comp_ge c)
  (get_comp_gt c)
  (get_comp_mono c)
  (get_comp_1 c)
  (get_comp_def c)
  (get_comp_max0 c)
  (get_comp_apos c)"


declare check_lpoly_coeffs.simps[simp del]

interpretation int_mono_poly: mono_poly_interpretation int_components whatever 
proof (rule mono_poly_interpretation.intro, simp_all)
  interpret mono_matrix_carrier int_mono "op >" 1 ..
  show "mono_poly_order_carrier (\<lambda>x. True) 0 op + op * ge op > True 1 1 max0 (\<lambda>x. True) int_mono" ..
qed

interpretation rat_mono_poly: mono_poly_interpretation "rat_components d" "check (d > 0) whatever"
proof (rule mono_poly_interpretation.intro, simp_all)
  assume "0 < d"
  from rat_interpretation[OF this] interpret SN_strict_mono_ordered_semiring_1 d "rat_gt d" rat_mono by simp
  interpret mono_matrix_carrier rat_mono "rat_gt d" d ..
  show "mono_poly_order_carrier (\<lambda>x. True) 0 op + op * ge (rat_gt d) True 1 d max0 (\<lambda>x. True) rat_mono" ..
qed

interpretation arctic_poly: poly_interpretation arctic_components whatever 
proof (rule poly_interpretation.intro, simp_all)
  fix x :: arctic
  show "x \<ge> 0" unfolding zero_arctic_def by simp
next
  interpret SN_both_mono_ordered_semiring_1 1 gt_arctic pos_arctic ..
  interpret both_mono_matrix_carrier pos_arctic gt_arctic 1  ..
  show "poly_order_carrier (\<lambda>x. True) 0 op + op * op \<ge> gt_arctic False 1 1 id pos_arctic" ..
qed


interpretation arctic_rat_poly: poly_interpretation "arctic_rat_components d" "check (d > 0) whatever"
proof (rule poly_interpretation.intro, simp_all)
  fix x :: arctic_rat
  show "x \<ge> 0" unfolding zero_arctic_rat_def by simp
next
  assume "0 < d" 
  from arctic_rat_interpretation[OF this] interpret SN_both_mono_ordered_semiring_1 1 "gt_arctic_rat d" pos_arctic_rat .
  interpret both_mono_matrix_carrier pos_arctic_rat "gt_arctic_rat d" 1  ..
  show "poly_order_carrier (\<lambda>x. True) 0 op + op * op \<ge> (gt_arctic_rat d) False 1 1 id pos_arctic_rat" ..
qed

sublocale mono_poly_interpretation \<subseteq> generic_redpair_impl "create_poly_redpair (\<lambda>I. check_carrier) c"
proof (rule create_poly_redpair)
  fix I
  assume "isOK check_carrier"
  from check_carrier[OF this] 
  interpret mono_poly_order_carrier "get_comp_wf_elem c" "get_comp_0 c" "get_comp_plus c" "get_comp_times c" "get_comp_ge c" "get_comp_gt c" "get_comp_mono c" "get_comp_1 c" "get_comp_def c" "get_comp_max0 c" "get_comp_apos c" "get_comp_check_mono c" .
  show "poly_order_carrier (get_comp_wf_elem c) (get_comp_0 c) (get_comp_plus c) (get_comp_times c) (get_comp_ge c) (get_comp_gt c) (get_comp_mono c) (get_comp_1 c) (get_comp_def c)
         (get_comp_max0 c) (get_comp_apos c)" ..
next
  fix I
  assume "isOK check_carrier"
  from check_carrier[OF this] 
  show "mono_poly_order_carrier (get_comp_wf_elem c) (get_comp_0 c) (get_comp_plus c) (get_comp_times c) (get_comp_ge c) (get_comp_gt c) (get_comp_mono c) (get_comp_1 c) (get_comp_def c) (get_comp_max0 c) (get_comp_apos c) (get_comp_check_mono c)" .
qed

sublocale poly_interpretation \<subseteq> generic_redpair_impl "create_poly_redpair (\<lambda>I. check_carrier) c"
proof (rule create_poly_redpair)
  fix I
  assume "isOK check_carrier"
  from check_carrier[OF this] 
  show "poly_order_carrier (get_comp_wf_elem c) (get_comp_0 c) (get_comp_plus c) (get_comp_times c) (get_comp_ge c) (get_comp_gt c) (get_comp_mono c) (get_comp_1 c) (get_comp_def c)
         (get_comp_max0 c) (get_comp_apos c)" .
next
  fix I
  assume "get_comp_mono c"
  with c have False by (cases c, auto) 
  thus "mono_poly_order_carrier (get_comp_wf_elem c) (get_comp_0 c) (get_comp_plus c) (get_comp_times c) (get_comp_ge c) (get_comp_gt c) (get_comp_mono c) (get_comp_1 c) (get_comp_def c) (get_comp_max0 c) (get_comp_apos c) (get_comp_check_mono c)" by simp
qed

abbreviation int_redpair :: "('f :: show,int)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "int_redpair \<equiv> create_poly_redpair (\<lambda>i. OK) int_components"

abbreviation arctic_redpair :: "('f :: show,arctic)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "arctic_redpair \<equiv> create_poly_redpair (\<lambda>i. OK) arctic_components"

abbreviation rat_redpair :: "rat \<Rightarrow> ('f :: show,rat)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "rat_redpair d \<equiv> create_poly_redpair (\<lambda>i. check (d > 0) (shows_string ''invalid delta'')) (rat_components d)"

abbreviation arctic_rat_redpair :: "rat \<Rightarrow> ('f :: show,arctic_rat)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "arctic_rat_redpair d \<equiv> create_poly_redpair (\<lambda>i. check (d > 0) (shows_string ''invalid delta'')) (arctic_rat_components d)"

lemma int_redpair: "generic_redpair_impl int_redpair" ..
lemma arctic_redpair: "generic_redpair_impl arctic_redpair" ..
lemma rat_redpair: "generic_redpair_impl (rat_redpair \<delta>)" ..
lemma arctic_rat_redpair: "generic_redpair_impl (arctic_rat_redpair \<delta>)" ..


abbreviation mat_components :: "nat \<Rightarrow> nat \<Rightarrow> ('a :: max_ordered_semiring_1) poly_components \<Rightarrow> 'a mat poly_components"
where "mat_components n sd c \<equiv> 
  (mat n n, mat0I 0 n n, mat1I 0 1 n, mat1I 0 (get_comp_def c) n, mat_plusI (op +), mat_multI 0 (op +) (op *) n,
   mat_max0, mat_ge, mat_gtI (get_comp_gt c) sd, get_comp_mono c, 
  concat [to_string n, ''x'',to_string n,'' matrices with strict dimension '',to_string sd,'' over '',get_comp_description c], 
  (\<lambda>x. True), mat_mono sd (get_comp_check_mono c))"

abbreviation mat_check :: "nat \<Rightarrow> nat \<Rightarrow> shows check"
where "mat_check n sd \<equiv> check (sd > 0 \<and> sd \<le> n) (shows_string ''invalid dimension'')"

sublocale mono_poly_interpretation \<subseteq> mono_matrix: generic_redpair_impl "create_poly_redpair (\<lambda>I. check_carrier \<guillemotright> mat_check n sd) (mat_components n sd c)"
proof (rule create_poly_redpair, simp_all, clarify)
  assume carrier: "isOK check_carrier" and sdpos: "0 < sd" and sd_n: "sd \<le> n"
  from c[symmetric] have id: 
    "get_comp_wf_elem c = (\<lambda>x. True) \<and> get_comp_0 c = 0 \<and> get_comp_1 c = 1 \<and> get_comp_max0 c = max0 \<and> get_comp_apos c = (\<lambda>x. True) \<and> get_comp_plus c = op +
     \<and> get_comp_times c = op * \<and> get_comp_ge c = ge \<and> get_comp_mono c = True" by (cases c, auto)
  from check_carrier[OF carrier, simplified id]
  interpret mono_poly_order_carrier "\<lambda>x. True" 0 "op +" "op *" ge "get_comp_gt c" True 1 "get_comp_def c" max0 "\<lambda>x. True" "get_comp_check_mono c" .
  interpret foo: mono_matrix_carrier "get_comp_check_mono c" "get_comp_gt c" "get_comp_def c"
    by (unfold_locales, 
      simp only: compat[simplified], 
      simp only: compat2[simplified], 
      rule default_geq_zero,
      rule plus_gt_left_mono[simplified], 
      simp,
      rule default_gt_zero,
      rule SN[simplified],
      rule check_mono[simplified], auto
    )
  from foo.mat_mono_poly_order_carrier[OF sd_n sdpos] id
  show "poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gtI (get_comp_gt c) sd) (get_comp_mono c) (mat1 n) (mat1I (0\<Colon>'a) (get_comp_def c) n)
     mat_max0 (\<lambda>x. True)" by simp (rule mono_poly_order_carrier.axioms(1))
next
  assume "isOK check_carrier \<and> 0 < sd \<and> sd \<le> n" and mono: "get_comp_mono c"
  hence carrier: "isOK check_carrier" and sdpos: "0 < sd" and sd_n: "sd \<le> n" by auto
  from c[symmetric] have id: 
    "get_comp_wf_elem c = (\<lambda>x. True) \<and> get_comp_0 c = 0 \<and> get_comp_1 c = 1 \<and> get_comp_max0 c = max0 \<and> get_comp_apos c = (\<lambda>x. True) \<and> get_comp_plus c = op +
     \<and> get_comp_times c = op * \<and> get_comp_ge c = ge \<and> get_comp_mono c = True" by (cases c, auto)
  from check_carrier[OF carrier, simplified id]
  interpret mono_poly_order_carrier "\<lambda>x. True" 0 "op +" "op *" ge "get_comp_gt c" True 1 "get_comp_def c" max0 "\<lambda>x. True" "get_comp_check_mono c" .
  interpret foo: mono_matrix_carrier "get_comp_check_mono c" "get_comp_gt c" "get_comp_def c"
    by (unfold_locales, 
      simp only: compat[simplified], 
      simp only: compat2[simplified], 
      rule default_geq_zero,
      rule plus_gt_left_mono[simplified], 
      simp,
      rule default_gt_zero,
      rule SN[simplified],
      rule check_mono[simplified], auto
    )
  from foo.mat_mono_poly_order_carrier[OF sd_n sdpos] id
  show "mono_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gtI (get_comp_gt c) sd) True (mat1 n) (mat1I (0\<Colon>'a) (get_comp_def c) n)
     mat_max0 (\<lambda>x. True) (mat_mono sd (get_comp_check_mono c))" unfolding mat_mono_def by simp 
qed

abbreviation mat_arctic_components :: "nat \<Rightarrow> ('a :: max_ordered_semiring_1) poly_components \<Rightarrow> 'a mat poly_components"
where "mat_arctic_components n c \<equiv>  
  (mat n n, mat0I 0 n n, mat1I 0 1 n, mat1I 0 (get_comp_def c) n, mat_plusI (op +), mat_multI 0 (op +) (op *) n,
   id, mat_ge, mat_comp_all (get_comp_gt c), False,
  concat [to_string n, ''x'',to_string n,'' matrices over '',get_comp_description c], 
  mat_arc_posI (get_comp_apos c), (\<lambda>x. False))"

abbreviation mat_arctic_check :: "nat \<Rightarrow> shows check"
where "mat_arctic_check n \<equiv> check (n > 0) (shows_string ''invalid dimension'')"



sublocale poly_interpretation \<subseteq> matrix: generic_redpair_impl "create_poly_redpair (\<lambda>I. check_carrier \<guillemotright> mat_arctic_check n) (mat_arctic_components n c)"
proof (rule create_poly_redpair, simp_all, clarify)
  assume carrier: "isOK check_carrier" and n_pos: "0 < n"
  from c[symmetric] have id: 
    "get_comp_wf_elem c = (\<lambda>x. True) \<and> get_comp_0 c = 0 \<and> get_comp_1 c = 1 \<and> get_comp_max0 c = id \<and> get_comp_plus c = op +
     \<and> get_comp_times c = op * \<and> get_comp_ge c = ge \<and> get_comp_mono c = False" by (cases c, auto)
  from check_carrier[OF carrier, simplified id]
  interpret poly_order_carrier "\<lambda>x. True" 0 "op +" "op *" ge "get_comp_gt c" False 1 "get_comp_def c" id "get_comp_apos c" .
  interpret foo: both_mono_matrix_carrier "get_comp_apos c" "get_comp_gt c" "get_comp_def c"
    by(unfold_locales, 
      simp only: compat[simplified],
      simp only: compat2[simplified], 
      rule default_geq_zero,
      rule plus_gt_both_mono[simplified], simp+,
      rule times_gt_left_mono[simplified], simp+,
      rule zero_leastI[simplified], 
      rule zero_leastII[simplified], simp+,
      rule zero_leastIII,
      rule SN[simplified zero_leastIII, simplified],
      rule arc_pos_plus[simplified], simp+,
      rule arc_pos_mult[simplified], simp+,
      rule arc_pos_default)
  from foo.mat_poly_order_carrier[OF n_pos] id
  show "poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_comp_all (get_comp_gt c)) False (mat1 n) (mat1I (0\<Colon>'a) (get_comp_def c) n) id
        (mat_arc_posI (get_comp_apos c))" by simp
qed

abbreviation int_mat_redpair :: "nat \<Rightarrow> nat \<Rightarrow> ('f :: show,int mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "int_mat_redpair n sd \<equiv> create_poly_redpair (\<lambda>i. OK \<guillemotright> mat_check n sd) (mat_components n sd int_components)"

abbreviation rat_mat_redpair :: "nat \<Rightarrow> nat \<Rightarrow> rat \<Rightarrow>  ('f :: show,rat mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "rat_mat_redpair n sd d \<equiv> create_poly_redpair (\<lambda>i. check (d > 0) (shows_string ''invalid delta'') \<guillemotright> mat_check n sd) (mat_components n sd (rat_components d))"

abbreviation arctic_mat_redpair :: "nat \<Rightarrow> ('f :: show,arctic mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "arctic_mat_redpair n \<equiv> create_poly_redpair (\<lambda>i. OK \<guillemotright> mat_arctic_check n) (mat_arctic_components n arctic_components)"

abbreviation arctic_rat_mat_redpair :: "nat \<Rightarrow> rat \<Rightarrow> ('f :: show,arctic_rat mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redpair_impl" 
where "arctic_rat_mat_redpair n d \<equiv> create_poly_redpair (\<lambda>i. check (d > 0) (shows_string ''invalid delta'') \<guillemotright> mat_arctic_check n) (mat_arctic_components n (arctic_rat_components d))"


end
