(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory LinearPolynomial
imports Show Monadic_Check Abstract_Rewriting 
begin

type_synonym ('v,'a)p_ass = "'v \<Rightarrow> 'a"
type_synonym ('v,'a)p_vars = "('v \<times> 'a)list"
datatype ('v,'a)l_poly = LPoly 'a "('v,'a)p_vars"

(* naturally, all of the following functions would be more easy to 
   define inside the poly_carrier locale (such that one does not have to pass zero, plus, ... around
   however, then code-generation does not work 
*)
fun add_varI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'v \<Rightarrow> 'a \<Rightarrow> ('v,'a)p_vars \<Rightarrow> ('v,'a)p_vars"
  where "add_varI _ _ x a [] = [(x,a)]"
    |   "add_varI ze pl x a ((y,b) # vas) = (if x = y then (let s = pl a b in if s = ze then vas else (x,s) # vas) else ((y,b) # add_varI ze pl x a vas))"

fun sum_pvarsI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('v,'a)p_vars \<Rightarrow> ('v,'a)p_vars \<Rightarrow> ('v,'a)p_vars"
  where "sum_pvarsI _ _  [] vbs = vbs"
     |  "sum_pvarsI ze pl ((x,a) # vas) vbs = (if a = ze then sum_pvarsI ze pl vas vbs else sum_pvarsI ze pl vas (add_varI ze pl x a vbs))"

fun mul_pvarsI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a \<Rightarrow> ('v,'a)p_vars \<Rightarrow> ('v,'a)p_vars"
  where "mul_pvarsI _ _ a [] = []"
      | "mul_pvarsI ze ti a ((x,b) # vas) = (let p = ti a b in let res = mul_pvarsI ze ti a vas in if p = ze then res else (x,p) # res)"

fun list_sumI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a list \<Rightarrow> 'a"
where "list_sumI ze pl [] = ze"
  | "list_sumI ze pl (x # xs) = pl x (list_sumI ze pl xs)"

locale poly_carrier =  
  fixes wf_elem :: "'a::show \<Rightarrow> bool"
  and zero :: "'a" ("0")
  and plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "+" 65)
  and times :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "*" 70)
  and geq :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<ge>" 21)
  and gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix ">" 21)
  and plus_single_mono :: "bool"
  assumes wf_zero: "wf_elem 0"
  and wf_plus: "\<lbrakk>wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> wf_elem (x+y)"
  and plus_comm: "\<lbrakk>wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> x+y = y+x"
  and plus_ass: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x+(y+z) = (x+y)+z"
  and zero_left: "wf_elem x \<Longrightarrow> 0+x = x"
  and wf_times: "\<lbrakk>wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> wf_elem (x*y)"
  and times_ass: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x*(y*z) = (x*y)*z"
  and left_mult_zero: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> 0*x = 0"
  and right_mult_zero: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> x*0 = 0"
  and distrib_left: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x*(y+z) = (x*y)+(x*z)"
  and distrib_right: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> (x+y)*z = (x*z)+(y*z)"
  and compat: "\<lbrakk>s \<ge> (t :: 'a); t > u; wf_elem s; wf_elem t; wf_elem u\<rbrakk> \<Longrightarrow> s > u"
  and compat2: "\<lbrakk>s > (t :: 'a); t \<ge> u; wf_elem s; wf_elem t; wf_elem u\<rbrakk> \<Longrightarrow> s > u"
  and plus_left_mono: "\<lbrakk>x \<ge> y; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x + z \<ge> y + z"
  and times_left_mono: "\<lbrakk>z \<ge> 0; x \<ge> y; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x * z \<ge> y * z"
  and times_right_mono: "\<lbrakk>x \<ge> 0; y \<ge> z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x * y \<ge> x * z"
  and geq_refl: "wf_elem x \<Longrightarrow> x \<ge> x"
  and geq_trans[trans]: "\<lbrakk>x \<ge> y; y \<ge> z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x \<ge> z"
  and plus_gt_left_mono: "\<lbrakk>x > y; wf_elem x; wf_elem y; wf_elem z; plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + z"
  and plus_gt_both_mono: "\<lbrakk>x > y; z > u; wf_elem x; wf_elem y; wf_elem z; wf_elem u; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + u"
  and times_gt_left_mono: "\<lbrakk>x > y; wf_elem x; wf_elem y; wf_elem z; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x * z > y * z"
  and zero_leastI: "\<lbrakk>wf_elem x; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x > 0" 
  and zero_leastII: "\<lbrakk>0 > x; wf_elem x; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x = 0" 
begin

lemma plus_gt_left_mono2: assumes xy: "(x :: 'a) > y" and zu: "(z :: 'a) \<ge> u" and x: "wf_elem x" and y: "wf_elem y" and z: "wf_elem z" and u: "wf_elem u" and psm: plus_single_mono
  shows "(x :: 'a) + z > y + u"
proof -
  have "x + z > y + z" by (rule plus_gt_left_mono[OF xy x y z psm])
  also have "\<dots> = z + y" by (rule plus_comm[OF y z])
  finally have one: "x + z > z + y" .
  have "z+ y \<ge> u + y" by (rule plus_left_mono[OF zu z u y])
  also have "\<dots> = y + u" by (rule plus_comm[OF u y])
  finally have two: "z+y \<ge> y+u" .
  show ?thesis
    by (rule compat2[OF one two wf_plus wf_plus wf_plus], auto simp: x y z u)
qed


lemmas poly_simps = wf_zero wf_plus plus_ass zero_left wf_times times_ass left_mult_zero right_mult_zero distrib_left distrib_right geq_refl
declare poly_simps[simp]

lemma plus_right_mono: "\<lbrakk>y \<ge> z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x + y \<ge> x + z"
using plus_comm[where x = x] plus_left_mono by auto

lemma plus_left_right_mono: "\<lbrakk>x \<ge> y; z \<ge> u; wf_elem x; wf_elem y; wf_elem z; wf_elem u\<rbrakk> \<Longrightarrow> x + z \<ge> y + u"
  by (rule geq_trans[where y="y+z"], auto, rule plus_left_mono, auto, rule plus_right_mono, auto)

abbreviation add_var where "add_var \<equiv> add_varI (0 :: 'a) (op +)"
abbreviation sum_pvars where "sum_pvars \<equiv> sum_pvarsI (0 :: 'a) (op +)"
abbreviation mul_pvars where "mul_pvars \<equiv> mul_pvarsI (0 :: 'a) (op *)"

lemma zero_right[simp]: "wf_elem x \<Longrightarrow> x + 0 = x"
proof -
  assume x: "wf_elem x"
  from this wf_zero have id: "x + 0 = 0 + x" by (rule plus_comm)
  show ?thesis by (simp add: id x)
qed

abbreviation list_sum :: "'a list \<Rightarrow> 'a"
where "list_sum \<equiv> list_sumI zero plus"


lemma wf_list_sum[simp]: assumes wf_as: "\<forall> a \<in> set as. wf_elem a"
  shows "wf_elem (list_sum as)"
using wf_as
by (induct as, auto)

lemma pos_list_sum: assumes "\<forall> a \<in> set as. wf_elem a \<and> geq a 0"
  shows "list_sum as \<ge> 0"
using assms
proof (induct as, simp)
  case (Cons a as)
  have "list_sum (a # as) \<ge> 0 + 0"
    by (simp only: list_sumI.simps, rule plus_left_right_mono, auto simp: Cons)
  thus ?case by auto
qed

lemma list_sum_mono: assumes ge: "\<forall> t \<in> set ts. (f t \<ge> g t) \<and> wf_elem (f t) \<and> wf_elem (g t)"
  shows "list_sum (map f ts) \<ge> list_sum (map g ts)"
using ge
proof (induct ts)
  case (Cons t ts)
  show ?case by (simp, rule plus_left_right_mono, auto simp: Cons)
qed simp

lemma sum_pos: assumes "geq a 0" and "geq b 0" and "wf_elem a" and "wf_elem b" shows "geq (a + b) 0"
proof -
  have "a + b \<ge> 0 + 0" by (rule plus_left_right_mono, auto simp: assms)
  thus ?thesis by auto
qed

definition wf_ass :: "('v,'a)p_ass \<Rightarrow> bool"
  where "wf_ass \<alpha> \<equiv> \<forall> x. wf_elem (\<alpha> x)"

definition wf_pvars :: "('v,'a)p_vars \<Rightarrow> bool"
  where "wf_pvars vas \<equiv> \<forall> a \<in> set (map snd vas). wf_elem a"

lemma wf_mul_pvars[simp]: 
  assumes wf_pvars: "wf_pvars vas"
  and wf_elem: "wf_elem a"
  shows "wf_pvars (mul_pvars a vas)"
using wf_pvars  
unfolding wf_pvars_def
proof (induct vas)
  case (Cons yb vas)
  show ?case
  proof (cases yb)
    case (Pair y b)
    show ?thesis 
    proof (cases "a * b = 0")
      case True
      with Pair Cons show ?thesis by (auto simp: Let_def)
    next
      case False
      with Pair Cons wf_elem show ?thesis by (auto simp: Let_def)
    qed
  qed
qed simp

lemma wf_add_var[simp]: 
  assumes wf_pvars: "wf_pvars vas"
  and wf_elem: "wf_elem a"
  shows "wf_pvars (add_var x a vas)"
using wf_pvars  
unfolding wf_pvars_def
proof (induct vas)
  case Nil
  with wf_elem show ?case by auto
next
  case (Cons yb vas)
  show ?case
  proof (cases yb)
    case (Pair y b)
    show ?thesis 
    proof (cases "x = y")
      case True
      with Pair Cons wf_elem show ?thesis 
	by (cases "a + b = 0", auto simp: Let_def) 
    next
      case False
      with Pair Cons show ?thesis by auto
    qed
  qed
qed

lemma wf_sum_pvars[simp]: 
  assumes wf_pvars_a: "wf_pvars vas"
  and wf_pvars_b: "wf_pvars vbs"
  shows "wf_pvars (sum_pvars vas vbs)"
using wf_pvars_a  wf_pvars_b
unfolding wf_pvars_def[where vas = vas]
proof (induct vas arbitrary: vbs)
  case (Cons xa vas)
  show ?case
  proof (cases xa)
    case (Pair x a)
    thus ?thesis using Cons by auto
  qed
qed simp


fun eval_pvars :: "('v,'a)p_ass \<Rightarrow> ('v,'a)p_vars \<Rightarrow> 'a"
where "eval_pvars _ [] = zero"
  |   "eval_pvars \<alpha> ((x,a) # vas) = (a * \<alpha> x + eval_pvars \<alpha> vas)"

lemma wf_eval_pvars[simp]: assumes wf_ass: "wf_ass \<alpha>"
  and wf_pvars: "wf_pvars vas"
  shows "wf_elem (eval_pvars \<alpha> vas)"
using wf_pvars
unfolding wf_pvars_def
proof (induct vas)
  case (Cons xa vas)
  show ?case
  proof (cases xa)
    case (Pair x a)
    from wf_ass have "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
    with Pair Cons show ?thesis by auto
  qed
qed simp

lemma mul_pvars_sound[simp]: 
  assumes wf_pvars: "wf_pvars vas"
  and wf_a: "wf_elem a"
  and wf_ass: "wf_ass \<alpha>"
  shows "eval_pvars \<alpha> (mul_pvars a vas) = a * (eval_pvars \<alpha> vas)"
using wf_pvars  
proof (induct vas)
  case Nil thus ?case by (simp add: wf_a)
next
  case (Cons xb vas)
  show ?case
  proof (cases xb)
    case (Pair x b)
    with Cons have wf_b: "wf_elem b" unfolding wf_pvars_def by auto
    from Cons have wf_vas: "wf_pvars vas" unfolding wf_pvars_def by auto
    from wf_ass have wf_x: "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
    show ?thesis 
    proof (cases "a * b = 0")
      case True
      with Pair Cons have "eval_pvars \<alpha> (mul_pvars a (xb # vas)) = eval_pvars \<alpha> (mul_pvars a vas)" by (simp add: Let_def)
      also have "\<dots> = a * eval_pvars \<alpha> vas" using Cons wf_vas wf_a by auto
      also have "\<dots> = a * (b * \<alpha> x + eval_pvars \<alpha> vas)" using True wf_vas wf_a wf_b wf_x wf_ass by auto
      finally show ?thesis using Pair by auto
    next
      case False
      with Cons Pair have "eval_pvars \<alpha> (mul_pvars a (xb # vas)) = eval_pvars \<alpha> ((x,a*b) # mul_pvars a vas)" by (simp add: Let_def)
      also have "\<dots> = a * (b * \<alpha> x + eval_pvars \<alpha> vas)" using Cons wf_vas wf_vas wf_a wf_b wf_x wf_ass by auto
      finally show ?thesis using Pair by auto
    qed
  qed
qed

definition pos_pvars :: "('v,'a)p_vars \<Rightarrow> bool"
where "pos_pvars vas \<equiv> \<forall> va \<in> set vas. snd va \<ge> 0"

lemma pos_mul_pvars:
  assumes pos_vas: "pos_pvars vas"
  and     wf_vas: "wf_pvars vas"
  and     pos_a: "geq a 0"
  and     wf_a: "wf_elem a"
shows "pos_pvars (mul_pvars a vas)"
using pos_vas wf_vas unfolding pos_pvars_def
proof (induct vas)
  case (Cons xb vas)
  show ?case
  proof (cases xb)
    case (Pair x b)
    with Cons have wf_b: "wf_elem b" and pos_b: "b \<ge> 0" unfolding wf_pvars_def by auto
    from Cons have "wf_pvars vas" unfolding wf_pvars_def by auto
    with Cons have rec: "\<forall> va \<in> set (mul_pvars a vas). snd va \<ge> 0" by auto
    show ?thesis 
    proof (cases "a * b = 0")
      case True
      with Cons Pair have "mul_pvars a (xb # vas) = mul_pvars a vas" by (simp add: Let_def)
      with rec show ?thesis by auto
    next
      case False
      have "a * b \<ge> 0 * 0" by (rule geq_trans[where y = "a * 0"], rule times_right_mono, auto simp: wf_a pos_a pos_b wf_b)
      hence pos_ab: "a * b \<ge> 0" by simp            
      from False Pair Cons have "mul_pvars a (xb # vas) = (x,a*b) # mul_pvars a vas" by (simp add: Let_def)
      with rec pos_ab show ?thesis by auto
    qed
  qed
qed simp

declare mul_pvarsI.simps[simp del]

lemma add_var_sound[simp]: 
  assumes wf_pvars: "wf_pvars vas"
  and wf_a: "wf_elem a"
  and wf_ass: "wf_ass \<alpha>"
  shows "eval_pvars \<alpha> (add_var x a vas) = a * \<alpha> x + eval_pvars \<alpha> vas"
using wf_pvars  
proof (induct vas)
  case (Cons yb vas)
  show ?case
  proof (cases yb)
    case (Pair y b)
    with Cons have wf_b: "wf_elem b" unfolding wf_pvars_def by auto
    from Cons have wf_vas: "wf_pvars vas" unfolding wf_pvars_def by auto
    from wf_ass have wf_x: "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
    show ?thesis 
    proof (cases "x = y")
      case True note oTrue = this
      show ?thesis 
      proof (cases "a + b = 0")
	case True
	with oTrue Pair have "eval_pvars \<alpha> (add_var x a (yb # vas)) = eval_pvars \<alpha> (vas)" by (simp add: Let_def)
	also have "\<dots> = (a + b) * \<alpha> x + eval_pvars \<alpha> vas" using True wf_x wf_ass wf_vas by auto
	also have "\<dots> = a * \<alpha> x + eval_pvars \<alpha> ((y,b) # vas)" using `x = y` wf_ass wf_x wf_vas wf_a wf_b  by auto
	finally show ?thesis using Pair by auto
      next
	case False
	with oTrue Pair have "eval_pvars \<alpha> (add_var x a (yb # vas)) = eval_pvars \<alpha> ((x,a+b) # vas)" by (simp add: Let_def)
	also have "\<dots> = (a + b) * \<alpha> x + eval_pvars \<alpha> vas" using True wf_x wf_ass wf_vas by auto
	also have "\<dots> = a * \<alpha> x + eval_pvars \<alpha> ((y,b) # vas)" using `x = y` wf_ass wf_x wf_vas wf_a wf_b  by auto
	finally show ?thesis using Pair by auto
      qed
    next
      case False
      from wf_ass have wf_y: "wf_elem (\<alpha> y)" unfolding wf_ass_def by auto
      with Pair  False have "eval_pvars \<alpha> (add_var x a (yb # vas)) = eval_pvars \<alpha> ((y,b) # (add_var x a vas))" by (simp add: Let_def)
      also have "\<dots> = b * \<alpha> y + eval_pvars \<alpha> (add_var x a vas)" by auto
      also have "\<dots> = b * \<alpha> y + (a * \<alpha> x + eval_pvars \<alpha> vas)" using Cons wf_vas by auto
      also have "\<dots> = a * \<alpha> x + eval_pvars \<alpha> ((y,b) # vas)" using wf_ass wf_x wf_y wf_vas wf_a wf_b plus_comm[where x = "b * \<alpha> y" and y = "a * \<alpha> x"] by (auto)
      finally show ?thesis using Pair by auto
    qed
  qed
qed simp

lemma pos_add_var: 
  assumes wf_pvars: "wf_pvars vas"
  and pos_pvars: "pos_pvars vas"
  and wf_a: "wf_elem a"
  and pos_a: "geq a 0"
  shows "pos_pvars (add_var x a vas)"
using wf_pvars  pos_pvars unfolding pos_pvars_def
proof (induct vas)
  case Nil thus ?case by (simp add: pos_a)
next
  case (Cons yb vas)
  show ?case
  proof (cases yb)
    case (Pair y b)
    with Cons have wf_b: "wf_elem b" and pos_b: "b \<ge> 0" unfolding wf_pvars_def by auto
    from Cons have "wf_pvars vas" unfolding wf_pvars_def by auto
    with Cons have rec: "\<forall> va \<in> set (add_var x a vas). snd va \<ge> 0" by auto
    show ?thesis 
    proof (cases "x = y")
      case True note oTrue = this
      show ?thesis 
      proof (cases "a + b = 0")
	case True
	with oTrue Pair have "add_var x a (yb # vas) = vas" by (simp add: Let_def)
	with Cons show ?thesis by auto
      next
	case False
	with True Pair have id: "add_var x a (yb # vas) = (x,a+b) # vas" by (simp add: Let_def)
	have "a + b \<ge> 0 + 0" by (rule plus_left_right_mono, auto simp: pos_a pos_b wf_a wf_b)
	with Cons id show ?thesis by auto
      qed
    next
      case False
      with Pair have "add_var x a (yb # vas) = (y,b) # (add_var x a vas)" by (simp add: Let_def)
      with pos_b rec show ?thesis by auto 
    qed
  qed
qed

declare add_varI.simps[simp del]
  

lemma sum_pvars_sound[simp]: assumes wf_ass: "wf_ass \<alpha>"
  and wf_pvars_a: "wf_pvars vas"
  and wf_pvars_b: "wf_pvars vbs" 
  shows "eval_pvars \<alpha> (sum_pvars vas vbs) = eval_pvars \<alpha> vas + eval_pvars \<alpha> vbs"
using wf_pvars_a wf_pvars_b
proof (induct vas arbitrary: vbs)
  case Nil thus ?case using wf_ass by auto
next
  case (Cons xa vas)
  hence wf_vas: "wf_pvars vas" unfolding wf_pvars_def by auto
  show ?case
  proof (cases xa)
    case (Pair x a)
    with Cons have wf_a: "wf_elem a" unfolding wf_pvars_def by auto
    have wf_x: "wf_elem (\<alpha> x)" using wf_ass unfolding wf_ass_def by auto
    have wf_evas: "wf_elem (eval_pvars \<alpha> vas)" using wf_eval_pvars wf_ass wf_vas by auto
    show ?thesis 
    proof (cases "a = 0")
      case True
      thus ?thesis using Pair Cons wf_vas wf_x wf_ass by auto
    next
      case False
      from wf_a Cons have wf_xvbs: "wf_pvars (add_var x a vbs)" by auto
      with False show ?thesis 
	using Pair Cons wf_vas wf_x wf_ass wf_a plus_comm[where x = "a * \<alpha> x" and y = "eval_pvars \<alpha> vas"] by auto
    qed
  qed
qed

lemma pos_sum_pvars: 
  assumes pos_vas: "pos_pvars vas"
  and     wf_vas: "wf_pvars vas"
  and     pos_vbs: "pos_pvars vbs"
  and     wf_vbs: "wf_pvars vbs"
shows "pos_pvars (sum_pvars vas vbs)"
using assms 
proof (induct vas arbitrary: vbs)
  case (Cons xa vas)
  from Cons(2) Cons(3)
  have wf_vas: "wf_pvars vas" and pos_vas: "pos_pvars vas" unfolding wf_pvars_def pos_pvars_def by auto
  show ?case
  proof (cases xa)
    case (Pair x a)
    with Cons(2) Cons(3) have wf_a: "wf_elem a" and pos_a: "a \<ge> 0" unfolding wf_pvars_def pos_pvars_def by auto
    show ?thesis 
    proof (cases "a = 0")
      case True
      thus ?thesis using Pair Cons(1)[OF pos_vas wf_vas Cons(4) Cons(5)] unfolding pos_pvars_def by auto
    next
      case False
      from wf_a Cons have wf_xvbs: "wf_pvars (add_var x a vbs)" by auto
      have "pos_pvars (add_var x a vbs)" by (rule pos_add_var, auto simp: Cons wf_a pos_a)
      with Cons wf_xvbs wf_vas pos_vas have "pos_pvars (sum_pvars vas (add_var x a vbs))" by auto      
      with False show ?thesis using Pair by auto
    qed
  qed
qed simp

declare sum_pvarsI.simps[simp del]
end

fun mul_lpolyI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a \<Rightarrow> ('v,'a)l_poly \<Rightarrow> ('v,'a)l_poly"
where "mul_lpolyI ze ti a (LPoly b vas) = LPoly (ti a b) (mul_pvarsI ze ti a vas)"

fun sum_lpolyI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('v,'a)l_poly \<Rightarrow> ('v,'a)l_poly \<Rightarrow> ('v,'a)l_poly"
where "sum_lpolyI ze pl (LPoly a vas) (LPoly b vbs) = LPoly (pl a b) (sum_pvarsI ze pl vas vbs)"

fun get_nc_lpoly :: "('v,'a)l_poly \<Rightarrow> ('v,'a)p_vars" 
where "get_nc_lpoly (LPoly _ nc) = nc"

context poly_carrier
begin

abbreviation mul_lpoly :: "'a \<Rightarrow> ('v,'a)l_poly \<Rightarrow> ('v,'a)l_poly"
where "mul_lpoly \<equiv> mul_lpolyI zero times"

abbreviation sum_lpoly :: "('v,'a)l_poly \<Rightarrow> ('v,'a)l_poly \<Rightarrow> ('v,'a)l_poly"
where "sum_lpoly \<equiv> sum_lpolyI zero plus"

fun wf_lpoly :: "('v,'a)l_poly \<Rightarrow> bool" 
where "wf_lpoly (LPoly a vas) = (wf_elem a \<and> wf_pvars vas)"

fun eval_lpoly :: "('v,'a)p_ass \<Rightarrow> ('v,'a)l_poly \<Rightarrow> 'a"
where "eval_lpoly \<alpha> (LPoly a vas) = a + eval_pvars \<alpha> vas"


definition pos_coeffs :: "('v,'a)l_poly \<Rightarrow> bool"
where "pos_coeffs p \<equiv> pos_pvars (get_nc_lpoly p)"

lemma pos_sum_lpoly: 
  assumes wf_p: "wf_lpoly p" 
  and wf_q: "wf_lpoly q" 
  and pos_p: "pos_coeffs p"
  and pos_q: "pos_coeffs q"
  shows "pos_coeffs (sum_lpoly p q)"
using assms unfolding pos_coeffs_def
  by (cases p, cases q, simp add: pos_sum_pvars)

lemma wf_sum_lpoly: 
  assumes wf_p: "wf_lpoly p" 
  and wf_q: "wf_lpoly q" 
  shows "wf_lpoly (sum_lpoly p q)"
using assms
by (cases p, cases q, auto)

lemma pos_mul_lpoly: 
  assumes wf_a: "wf_elem a"
  and pos_a: "geq a 0"
  and wf_p: "wf_lpoly p" 
  and pos_p: "pos_coeffs p"
  shows "pos_coeffs (mul_lpoly a p)"
using assms unfolding pos_coeffs_def
  by (cases p, simp add: pos_mul_pvars)

lemma wf_mul_lpoly: 
  assumes wf_a: "wf_elem a" 
  and wf_p: "wf_lpoly p" 
  shows "wf_lpoly (mul_lpoly a p)"
using assms
by (cases p, auto)

lemma wf_eval_lpoly: 
  assumes wf_ass: "wf_ass \<alpha>"
  and wf_p: "wf_lpoly p"
  shows "wf_elem (eval_lpoly \<alpha> p)"
using assms by (cases p, auto)

lemma sum_poly_sound: assumes wf_ass: "wf_ass \<alpha>"
  and wf_p: "wf_lpoly p" 
  and wf_q: "wf_lpoly q" 
  shows "eval_lpoly \<alpha> (sum_lpoly p q) = eval_lpoly \<alpha> p + eval_lpoly \<alpha> q"
proof (cases p)
  case (LPoly a vas) note oLPoly = this
  show ?thesis
  proof (cases q)
    case (LPoly b vbs)
    with oLPoly assms show ?thesis by (auto simp: plus_comm[where x = "a + eval_pvars \<alpha> vas"] plus_comm[where x = a and y = b])
  qed
qed

lemma mul_poly_sound: assumes wf_ass: "wf_ass \<alpha>"
  and wf_a: "wf_elem a"
  and wf_p: "wf_lpoly p" 
  shows "eval_lpoly \<alpha> (mul_lpoly a p) = a * eval_lpoly \<alpha> p"
using assms by (cases p, auto)

text {* let us head for comparing polynomials now *}

definition pos_ass :: "('v,'a)p_ass \<Rightarrow> bool"
where "pos_ass \<alpha> \<equiv> \<forall> x. \<alpha> x \<ge> 0"

definition zero_ass :: "('v,'a)p_ass"
where "zero_ass x = 0"

lemma pos_zero_ass[simp]: "pos_ass zero_ass"
unfolding pos_ass_def zero_ass_def by auto 

lemma wf_zero_ass[simp]: "wf_ass zero_ass"
unfolding wf_ass_def zero_ass_def by auto 

definition poly_s :: "('v,'a)l_poly ars"
where "poly_s \<equiv> {(p,q) . \<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> gt (eval_lpoly \<alpha> p) (eval_lpoly \<alpha> q)}"     

definition poly_ns :: "('v,'a)l_poly ars"
where "poly_ns \<equiv> {(p,q) . \<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> geq (eval_lpoly \<alpha> p) (eval_lpoly \<alpha> q)}"     
end

fun
  lookup_rest :: "'a \<Rightarrow> ('a \<times> 'b) list \<Rightarrow> ('b \<times> ('a \<times> 'b) list) option"
where
  "lookup_rest x [] = None" |
  "lookup_rest x ((y,c) # ycs) = (if x = y then Some (c, ycs) else 
     (case lookup_rest x ycs of None \<Rightarrow> None | Some (d,yccs) \<Rightarrow> Some (d, (y,c) # yccs)))"

fun
  check_pvarsI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('v,'a)p_vars \<Rightarrow> ('v,'a)p_vars \<Rightarrow> 'v check"
where
  "check_pvarsI ze ge vas [] = check_allm (\<lambda>va. check (ge (snd va)  ze) (fst va)) vas" |
  "check_pvarsI ze ge vas ((x,b) # vbs) = do {
     let (a', vas') = 
       (case lookup_rest x vas of None \<Rightarrow> (ze,vas) | Some (a, vas'') \<Rightarrow> (a,vas''));
     check (ge a' b) x;
     check_pvarsI ze ge vas' vbs
   }"

fun
  show_pvars :: "'a \<Rightarrow> ('v::show, 'a::show) p_vars \<Rightarrow> string list"
where
  "show_pvars on [] = []" |
  "show_pvars on ((x,c) # vas) =
     ((if on = c then [] else to_string c) @ to_string x) # show_pvars on vas"

fun
  shows_lpoly :: "'a::show \<Rightarrow> 'a \<Rightarrow> ('v::show,'a)l_poly \<Rightarrow> shows"
where
  "shows_lpoly ze on (LPoly c cs) =
     (case show_pvars on cs of 
       [] \<Rightarrow> shows c
     | ss \<Rightarrow> (if c = ze
       then id
       else (shows c +@+ shows_string '' + '')) +@+ shows_list_gen shows [] [] '' + '' [] ss
   )"

fun
  check_lpoly_nsI
    :: "'a::show \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('v::show,'a)l_poly \<Rightarrow> ('v,'a)l_poly \<Rightarrow> shows check"
where
  "check_lpoly_nsI ze on ge (LPoly a vas) (LPoly b vbs) = do {
     check (ge a b) (shows_string ''problem when comparing constant parts'');
     check_pvarsI ze ge vas vbs
       <+? (\<lambda>e. ''problem when comparing coefficients of variable '' +#+ shows e)
   } <+? (\<lambda>e. ''problem when comparing '' +#+ shows_lpoly ze on (LPoly a vas)
           +@+ '' >= '' +#+ shows_lpoly ze on (LPoly b vbs) +@+ shows_nl +@+ e)"

fun
  check_lpoly_sI
  :: "bool \<Rightarrow> 'a::show \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool)
      \<Rightarrow> ('v::show,'a)l_poly \<Rightarrow> ('v,'a)l_poly \<Rightarrow> shows check"
where
  "check_lpoly_sI mode ze on ge gt (LPoly a vas) (LPoly b vbs) = do {
     check (gt a b) (shows_string ''problem when comparing constant part'');
     check_pvarsI ze (if mode then ge else gt) vas vbs
       <+? (\<lambda>x. ''problem when comparing coefficients of variable '' +#+ shows x) 
   } <+? (\<lambda>s. ''problem when comparing '' +#+ shows_lpoly ze on (LPoly a vas)
           +@+ '' > '' +#+ shows_lpoly ze on (LPoly b vbs) +@+ shows_nl +@+ s)"

context poly_carrier
begin

lemma lookup_rest_sound: 
  assumes lookup: "lookup_rest y vas = Some (a,vas')"
  and wf_ass: "wf_ass \<alpha>"
  and wf_vas: "wf_pvars vas"
  shows "eval_pvars \<alpha> vas = a * \<alpha> y + eval_pvars \<alpha> vas' \<and> wf_pvars vas' \<and> wf_elem a" 
using lookup wf_vas
proof (induct vas arbitrary: vas')
  case (Cons xb vas)
  show ?case 
  proof (cases xb)
    case (Pair x b)
    with Cons have wf_b: "wf_elem b" and wf_vas: "wf_pvars vas" and wf_x: "wf_elem (\<alpha> x)" using wf_ass unfolding wf_ass_def wf_pvars_def by auto
    show ?thesis 
    proof (cases "x = y")
      case True
      with Pair Cons have "a = b \<and> vas' = vas" by auto
      with wf_b wf_vas Pair True show ?thesis by auto
    next
      case False
      show ?thesis
      proof (cases "lookup_rest y vas")
	case None
	with Pair False have "lookup_rest y ((x,b) # vas) = None" by auto
	with Cons Pair show ?thesis by simp
      next
	case (Some res)
	show ?thesis 
	proof (cases res)
	  case (Pair c vbs)
	  with Some False `xb = (x,b)` have "lookup_rest y (xb # vas) = Some (c,(x,b) # vbs)" by simp
	  with Cons have ca: "c = a" and vas': "vas' = (x,b) # vbs" by auto
	  from Some Pair Cons ca wf_vas have rec: "eval_pvars \<alpha> vas = a * \<alpha> y + eval_pvars \<alpha> vbs \<and> wf_pvars vbs \<and> wf_elem a" by simp
	  from wf_ass have wf_y: "wf_elem (\<alpha> y)" unfolding wf_ass_def by auto
	  show ?thesis using `xb = (x,b)` wf_b vas' rec wf_x wf_y wf_ass 
	    by (auto simp: wf_pvars_def plus_comm[where x = "b * \<alpha> x" and y = "a * \<alpha> y"])
	qed
      qed
    qed
  qed
qed simp


abbreviation
  check_pvars :: "('v,'a)p_vars \<Rightarrow> ('v,'a)p_vars \<Rightarrow> 'v check"
where
  "check_pvars \<equiv> check_pvarsI zero geq"

lemma check_pvars_sound: assumes wf_ass: "wf_ass \<alpha>"
  and wf_vas: "wf_pvars vas"
  and wf_vbs: "wf_pvars vbs"
  and pos_ass: "pos_ass \<alpha>"
  and check: "isOK(check_pvars vas vbs)"
  shows "eval_pvars \<alpha> vas \<ge> eval_pvars \<alpha> vbs"
using check wf_vas wf_vbs
proof (induct vbs arbitrary: vas)
  case Nil
  hence "\<forall> va \<in> set vas. (snd va \<ge> 0)" by auto
  with `wf_pvars vas` have "eval_pvars \<alpha> vas \<ge> 0" 
  proof (induct vas)
    case (Cons xa vas)
    show ?case
    proof (cases xa)
      case (Pair x a)
      with Cons have wf_a: "wf_elem a" and wf_vas: "wf_pvars vas" and pos: "a \<ge> 0" unfolding wf_pvars_def by auto
      from wf_ass have wf_x: "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
      have "a * \<alpha> x + eval_pvars \<alpha> vas \<ge> 0 + eval_pvars \<alpha> vas" 
      proof (rule plus_left_mono)
	have posx: "\<alpha> x \<ge> 0" using pos_ass unfolding pos_ass_def by auto
	have ge1: "a * \<alpha> x \<ge> 0 * \<alpha> x" by (rule times_left_mono, auto simp: pos posx wf_x wf_a)
	have ge2: "\<dots> \<ge> 0 * 0" by (rule times_right_mono, auto simp: posx wf_x)
	have "a * \<alpha> x \<ge> 0 * 0" by (rule geq_trans[where y = "0 * \<alpha> x"], auto simp only: ge1 ge2, auto simp: wf_x wf_a)
	thus "a * \<alpha> x \<ge> 0" by auto 
      qed (auto simp: wf_a wf_ass wf_x wf_vas)
      hence ge1: "a * \<alpha> x + eval_pvars \<alpha> vas \<ge> eval_pvars \<alpha> vas" using wf_vas wf_ass by auto 
      have ge2: "\<dots> \<ge> 0" using Pair Cons wf_vas by auto
      have "a * \<alpha> x + eval_pvars \<alpha> vas \<ge> 0" by (rule geq_trans[where y = "eval_pvars \<alpha> vas"], auto simp only: ge1 ge2,
	auto simp: wf_a wf_x wf_ass wf_vas)
      thus ?thesis using Pair by auto
    qed
  qed simp
  thus ?case by auto
next
  case (Cons yb vbs)
  show ?case
  proof (cases yb)
    case (Pair y b) note oPair = this
    with Cons have wf_b: "wf_elem b" and wf_y: "wf_elem (\<alpha> y)" and pos_y: "\<alpha> y \<ge> 0" and wf_vbs: "wf_pvars vbs" 
      using wf_ass pos_ass unfolding wf_ass_def pos_ass_def wf_pvars_def by auto    
    show ?thesis 
    proof (cases "lookup_rest y vas")
      case None
      with Cons Pair wf_vbs have rec: "geq (eval_pvars \<alpha> vas) (eval_pvars \<alpha> vbs)" (is "geq ?l ?r") and small_b: "0 \<ge> b" by auto
      have "0 * \<alpha> y \<ge> b * \<alpha> y" by (rule times_left_mono, auto simp: wf_b wf_y small_b pos_y)
      hence ge: "0 \<ge> b * \<alpha> y" using wf_y by auto
      have "0 + ?l \<ge> b * \<alpha> y + ?r" by (rule plus_left_right_mono, auto simp: rec wf_ass wf_vbs wf_b wf_y Cons ge)
      with Cons wf_ass Pair None show ?thesis by auto
    next
      case (Some avas)
      show ?thesis 
      proof (cases avas)
	case (Pair a vas')
	with Some Cons oPair have lookup: "lookup_rest y vas = Some (a,vas')" and check: "isOK(check_pvars vas' vbs)" and ab: "a \<ge> b" by auto
	have lookupRes: "eval_pvars \<alpha> vas = a * \<alpha> y + eval_pvars \<alpha> vas' \<and> wf_pvars vas' \<and> wf_elem a"
	  by (rule lookup_rest_sound[OF lookup wf_ass `wf_pvars vas`])
	with Cons check wf_vbs have rec: "geq (eval_pvars \<alpha> vas')  (eval_pvars \<alpha> vbs)" (is "geq ?l ?r") by auto
	have "a * \<alpha> y + ?l \<ge> b * \<alpha> y + ?r" 
          by (rule plus_left_right_mono, rule times_left_mono, auto simp: rec lookupRes wf_ass wf_vbs wf_y wf_b ab pos_y)
	with lookupRes Pair Some oPair show ?thesis by auto 
      qed
    qed
  qed
qed

abbreviation
  check_pvars_gt :: "('v,'a)p_vars \<Rightarrow> ('v,'a)p_vars \<Rightarrow> 'v check"
where
  "check_pvars_gt \<equiv> check_pvarsI zero gt"

lemma check_pvars_gt_sound: assumes wf_ass: "wf_ass \<alpha>"
  and wf_vas: "wf_pvars vas"
  and wf_vbs: "wf_pvars vbs"
  and pos_ass: "pos_ass \<alpha>"
  and check: "isOK(check_pvars_gt vas vbs)"
  and mode: "\<not> plus_single_mono"
  shows "eval_pvars \<alpha> vas > eval_pvars \<alpha> vbs"
using check wf_vas wf_vbs
proof (induct vbs arbitrary: vas)
  case Nil
  have "eval_pvars \<alpha> Nil = 0" by auto
  thus ?case by (auto simp: zero_leastI mode wf_ass `wf_pvars vas`)
next
  case (Cons yb vbs)
  show ?case
  proof (cases yb)
    case (Pair y b) note oPair = this
    with Cons have wf_b: "wf_elem b" and wf_y: "wf_elem (\<alpha> y)" and pos_y: "\<alpha> y \<ge> 0" and wf_vbs: "wf_pvars vbs" 
      using wf_ass pos_ass unfolding wf_ass_def pos_ass_def wf_pvars_def by auto    
    show ?thesis 
    proof (cases "lookup_rest y vas")
      case None
      with Cons Pair wf_vbs have rec: "gt (eval_pvars \<alpha> vas) (eval_pvars \<alpha> vbs)" and small_b: "0 > b" by auto
      from small_b wf_b have "b = 0" using zero_leastII mode by auto
      with rec Pair wf_b wf_vbs wf_y wf_ass show ?thesis by auto 
    next
      case (Some avas)
      show ?thesis 
      proof (cases avas)
	case (Pair a vas')
	with Some oPair Cons  have lookup: "lookup_rest y vas = Some (a,vas')" and check: "isOK(check_pvars_gt vas' vbs)" and ab: "a > b" by auto
	have lookupRes: "eval_pvars \<alpha> vas = a * \<alpha> y + eval_pvars \<alpha> vas' \<and> wf_pvars vas' \<and> wf_elem a"
	  by (rule lookup_rest_sound[OF lookup wf_ass `wf_pvars vas`])
	with Cons check wf_vbs have gt: "eval_pvars \<alpha> vas' > eval_pvars \<alpha> vbs" (is "gt ?l ?r") by auto
	show ?thesis 
	proof (cases "\<alpha> y = 0")
	  case False
	  have "a * \<alpha> y + ?l > b * \<alpha> y + ?r"
	    by (rule plus_gt_both_mono, rule times_gt_left_mono[OF ab], auto simp: lookupRes wf_b wf_y wf_ass wf_vbs mode gt)
	  with lookupRes Some Pair oPair Cons show ?thesis by auto
	next
	  case True
	  hence "a * \<alpha> y + ?l = ?l" and "b * \<alpha> y + ?r = ?r" by (auto simp: wf_y lookupRes wf_ass wf_b wf_vbs)
	  with gt lookupRes True Some Pair oPair Cons show ?thesis by auto
	qed
      qed
    qed
  qed
qed

abbreviation
  check_lpoly_ns :: "'a \<Rightarrow> ('v::show,'a)l_poly \<Rightarrow> ('v,'a)l_poly \<Rightarrow> shows check"
where
  "check_lpoly_ns \<equiv> \<lambda> on. check_lpoly_nsI zero on geq"

lemma check_lpoly_ns_sound: assumes check: "isOK(check_lpoly_ns on p q)"
  and wf_p: "wf_lpoly p"
  and wf_q: "wf_lpoly q"
  shows "(p,q) \<in> poly_ns"
proof (cases p)
  case (LPoly a vas) note oLPoly = this
  show ?thesis
  proof (cases q)
    case (LPoly b vbs)
    with oLPoly assms have ab: "a \<ge> b" and pvars: "isOK(check_pvars vas vbs)" and
      wf_a: "wf_elem a" and wf_b: "wf_elem b" and wf_vas: "wf_pvars vas" and wf_vbs: "wf_pvars vbs" by auto
    {
      fix \<alpha>
      have "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> geq (plus a (eval_pvars \<alpha> vas)) (plus b (eval_pvars \<alpha> vbs))" (is "?ass \<longrightarrow> ?goal")
      proof 
	assume ass: ?ass
	show ?goal
	  by (rule plus_left_right_mono, auto simp: wf_a wf_b ass wf_vas wf_vbs ab,
	      rule check_pvars_sound, auto simp: ass pvars wf_vas wf_vbs)
      qed
    }
    hence "(LPoly a vas, LPoly b vbs) \<in> poly_ns" unfolding poly_ns_def
      by auto
    with LPoly oLPoly show ?thesis by auto
  qed
qed

abbreviation
  check_lpoly_s :: "'a \<Rightarrow> ('v::show,'a)l_poly \<Rightarrow> ('v,'a)l_poly \<Rightarrow> shows check"
where
  "check_lpoly_s \<equiv> \<lambda> on. check_lpoly_sI plus_single_mono zero on geq gt"

lemma check_lpoly_s_sound: assumes check: "isOK(check_lpoly_s on p q)"
  and wf_p: "wf_lpoly p"
  and wf_q: "wf_lpoly q"
  shows "(p,q) \<in> poly_s"
proof (cases p)
  case (LPoly a vas) note oLPoly = this
  show ?thesis
  proof (cases q)
    case (LPoly b vbs)
    with oLPoly assms have ab: "a > b" and ppvars: "isOK(check_pvarsI zero (if plus_single_mono then geq else gt) vas vbs)" and
      wf_a: "wf_elem a" and wf_b: "wf_elem b" and wf_vas: "wf_pvars vas" and wf_vbs: "wf_pvars vbs" by auto    
    {
      fix \<alpha>
      have "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> gt (plus a (eval_pvars \<alpha> vas)) (plus b (eval_pvars \<alpha> vbs))" (is "?ass \<longrightarrow> ?goal")
      proof 
	assume ass: ?ass
	show ?goal 
	proof (cases plus_single_mono)
	  case True
	  with ppvars have pvars: "isOK(check_pvars vas vbs)" by auto
	  show ?goal 
	    by (rule plus_gt_left_mono2[OF _ check_pvars_sound], auto simp: ass pvars wf_vas wf_vbs True wf_a wf_b ab)
	next 
	  case False
	  with ppvars have pvars: "isOK(check_pvars_gt vas vbs)" by auto
	  have gt: "eval_pvars \<alpha> vas > eval_pvars \<alpha> vbs" (is "gt ?l ?r") 
	    by (rule check_pvars_gt_sound, auto simp: ass wf_vas wf_vbs False pvars)
	  show ?thesis 
	    by (rule plus_gt_both_mono[OF ab gt], auto simp: wf_a wf_b ass wf_vas wf_vbs False)
	qed
      qed
    }
    hence "(LPoly a vas, LPoly b vbs) \<in> poly_s" unfolding poly_s_def by auto
    with LPoly oLPoly show ?thesis by auto
  qed
qed
end


locale poly_order_carrier = poly_carrier + 
  fixes one :: "'a" ("1")
   and  default :: "'a"
   and  max0 :: "'a \<Rightarrow> 'a"
   and  arc_pos :: "'a \<Rightarrow> bool"
  assumes SN: "SN {(x,y) . wf_elem x \<and> wf_elem y \<and> geq y zero \<and> arc_pos y \<and> gt x y}"
    and max0_id_pos: "\<lbrakk>geq x  zero; wf_elem x\<rbrakk> \<Longrightarrow> max0 x = x"
    and max0_pos: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> geq (max0 x) zero"
    and max0_mono: "\<lbrakk>x \<ge> y; wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> geq (max0 x) (max0 y)"
    and max0_x: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> geq (max0 x) x"
    and wf_max0: "wf_elem x \<Longrightarrow> wf_elem (max0 x)"
    and wf_one: "wf_elem 1"
    and wf_default: "wf_elem default"
    and left_one: "wf_elem x \<Longrightarrow> 1 * x = x"
    and right_one: "wf_elem x \<Longrightarrow> x * 1 = x"
    and one_geq_zero: "one \<ge> zero"
    and default_geq_zero: "default \<ge> zero"
    and arc_pos_plus: "\<lbrakk>arc_pos x; wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> arc_pos (x + y)"
    and arc_pos_mult: "\<lbrakk>arc_pos x; arc_pos y; wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> arc_pos (x * y)"
    and arc_pos_max0: "wf_elem x \<Longrightarrow> arc_pos (max0 x) = arc_pos x"
    and arc_pos_default: "arc_pos default"
begin


lemmas poly_order_simps = wf_one wf_default left_one right_one one_geq_zero default_geq_zero arc_pos_plus arc_pos_mult arc_pos_max0 arc_pos_default
end

locale mono_poly_order_carrier = poly_order_carrier +
  fixes check_mono :: "'a \<Rightarrow> bool"
  assumes plus_single_mono: "plus_single_mono"
  and check_mono: "\<lbrakk>check_mono x; gt y z; geq x 0; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> gt (times x y) (times x z)" 
  and default_gt_zero: "default > zero"

context poly_carrier
begin
declare poly_simps[simp del]
end


end
