(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SubtermCriterionImpl
imports SubtermCriterion DpTransImpl Substitution
begin

datatype 'f projL = Projection "('f \<times> nat) list"

type_synonym  ('f, 'v) prseq = "(pos \<times> ('f, 'v) rule \<times> bool \<times> ('f, 'v) term) list"
type_synonym  ('f, 'v) rseq = "(pos \<times> ('f, 'v) rule \<times> ('f, 'v) term) list"

instantiation pos :: "show" begin
  fun
    shows_pos :: "pos \<Rightarrow> shows"
  where
    "shows_pos (i <# \<epsilon>) = shows i" |
    "shows_pos (i <# ps) = (shows i +@+ shows_pos ps)" |
    "shows_pos \<epsilon> = shows_string ''e''"
  definition "shows_prec (d::nat) (p::pos) = shows_pos p"
  definition "shows_list_pos (ps::pos list) = shows_list_aux shows ps"

  lemma assoc_pos: "shows_prec d (p::pos) r @ s = shows_prec d p (r @ s)"
  unfolding shows_prec_pos_def
  proof (induct p)
    case (PCons i ps) thus ?case by (cases ps) simp_all
  qed simp

  lemma assoc_pos_elt: "\<forall>(p::pos)\<in>set ps. shows p r @ s = shows p (r @ s)"
  using assoc_pos by auto

  instance by
    (intro_classes, rule assoc_pos, unfold shows_list_pos_def)
    (rule shows_list_aux_assoc[OF assoc_pos_elt])
end

definition
  check_rstep ::
    "('f::show, 'v::show) rules \<Rightarrow> pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term
     \<Rightarrow> shows check"
where
  "check_rstep R p r s t \<equiv> do {
     let sps = poss_list s;
     let tps = poss_list t;
     check (list_ex (\<lambda> rule. r =v rule) R)
       (shows_rule r +@+ '' is not a rule of'' +#+ shows_nl +@+ shows_trs R +@+ shows_nl);
     check (p \<in> set sps) (shows p +@+ '' is not a position of '' +#+ shows s +@+ shows_nl);
     check (p \<in> set tps) (shows p +@+ '' is not a position of '' +#+ shows t +@+ shows_nl);
     let C = ctxt_of_pos_term p s;
     let D = ctxt_of_pos_term p t;
     let u = subt_at s p;
     let v = subt_at t p;
     (case match_list [(u,fst r), (v, snd r)] of
       Some \<sigma> \<Rightarrow>
         check (C = D)
           (''the term '' +#+ shows t
             +@+ '' does not result from a proper application of rule ''
             +#+ shows_nl +@+ shows_rule r +@+ '' at position '' +#+ shows p +@+ shows_nl)
     | None \<Rightarrow>
       error (''the term '' +#+ shows t
             +@+ '' does not result from a proper application of rule ''
             +#+ shows_nl +@+ shows_rule r +@+ '' at position '' +#+ shows p +@+ shows_nl))
   }"

lemma hole_pos_ctxt_of_pos_term: assumes "p \<in> poss t"
  shows "hole_pos (ctxt_of_pos_term p t) = p"
  using assms
proof (induct t arbitrary: p, simp)
  case (Fun f ts)
  show ?case 
  proof (cases p)
    case Empty
    thus ?thesis using Fun by auto
  next
    case (PCons i q)
    with Fun(2) have i: "i < length ts" and q: "q \<in> poss (ts ! i)" by auto
    hence "ts ! i \<in> set ts" by auto
    from Fun(1)[OF this q] PCons i show ?thesis by simp
  qed
qed

lemma hole_pos_id_ctxt: assumes "C\<langle>s\<rangle> = t"
  shows "ctxt_of_pos_term (hole_pos C) t = C"
using assms 
proof (induct C arbitrary: t, simp)
  case (More f bef C aft)
  thus ?case
  proof (cases t, simp)
    case (Fun g ts)
    with More have [simp]: "g = f" by simp
    from Fun More have bef: "take (length bef) ts = bef" by auto    
    from Fun More have aft: "drop (Suc (length bef)) ts = aft" by auto
    from Fun More have Cs: "C\<langle>s\<rangle> = ts ! length bef" by auto
    from Fun More show ?thesis by (simp add: bef aft More(1)[OF Cs])
  qed
qed

lemma check_rstep_sound:
   "isOK (check_rstep R p r s t) \<Longrightarrow> (\<exists> \<sigma> r. (s,t) \<in> rstep_r_p_s (set R) r p \<sigma>)"
proof -
  assume check: "isOK (check_rstep R p r s t)"  
  let ?C = "ctxt_of_pos_term p s"
  let ?D = "ctxt_of_pos_term p t"
  let ?mlist = "[(subt_at s p, fst r), (subt_at t p, snd r)]"
  from check[unfolded check_rstep_def] have "p \<in> poss s" and "p \<in> poss t" and "\<exists> rule \<in> set R. r =v rule" by (auto simp: list_ex_iff)
  then obtain rule where rule: "rule \<in> set R" and inst: "r =v rule" by auto
  from inst obtain \<delta> where fst: "fst r = fst rule \<cdot> \<delta>" and snd: "snd r = snd rule \<cdot> \<delta>" unfolding eq_rule_mod_vars_def instance_rule_def by auto
  from check[unfolded check_rstep_def]
  obtain \<sigma> where some: "match_list [(subt_at s p, fst r), (subt_at t p, snd r)] = Some \<sigma>" by force
  from match_list_sound[OF this] have "fst r \<cdot> \<sigma> = s |_ p" and "snd r \<cdot> \<sigma> = t |_ p" by auto 
  hence "s = ?C\<langle>fst rule \<cdot> \<delta> \<circ>s \<sigma>\<rangle>" and "t = ?D\<langle>snd rule \<cdot> \<delta> \<circ>s \<sigma>\<rangle>" using ctxt_supt_id[OF `p \<in> poss s`] ctxt_supt_id[OF `p \<in> poss t`] fst snd by simp+
  moreover from check[unfolded check_rstep_def, simplified] and some
    have "?C = ?D" by simp 
  ultimately have "(s,t) \<in> rstep_r_p_s (set R) rule p (\<delta> \<circ>s \<sigma>)" using `p \<in> poss s` and rule unfolding rstep_r_p_s_def some Let_def by force
  thus ?thesis by blast
qed

lemma check_rstep_complete:
   "\<exists> \<sigma>. (s,t) \<in> rstep_r_p_s (set R) r p \<sigma> \<Longrightarrow> isOK (check_rstep R p r s t)"
proof -
  assume "\<exists> \<sigma>. (s,t) \<in> rstep_r_p_s (set R) r p \<sigma>"
  then obtain \<sigma> where step: "(s,t) \<in> rstep_r_p_s (set R) r p \<sigma>" ..
  let ?C = "ctxt_of_pos_term p s"
  let ?C' = "ctxt_of_pos_term p t"
  from step have p: "p \<in> poss s" and r: "r \<in> set R" and s: "?C\<langle>fst r \<cdot> \<sigma>\<rangle> = s" and t: "?C\<langle>snd r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)  
  from hole_pos_ctxt_of_pos_term[OF p] have Cp: "hole_pos ?C = p" .
  from hole_pos_id_ctxt[OF t, simplified Cp] have C: "?C = ?C'" by auto
  from t[symmetric] have p': "p \<in> poss t" using hole_pos_poss[of ?C, simplified Cp] by best
  from r have r: "r \<in> set R" by (simp)
  have rr: "r =v r" unfolding eq_rule_mod_vars_def instance_rule_def 
    by (simp, rule exI[of _ \<iota>], simp)
  have "s |_ p = ?C\<langle>fst r \<cdot> \<sigma>\<rangle> |_ hole_pos ?C" by (simp only: s Cp)
  also have "\<dots> = fst r \<cdot> \<sigma>" by (rule subt_at_hole_pos)
  finally have proj: "s |_ p = fst r \<cdot> \<sigma>" .
  have "t |_ p = ?C\<langle>snd r \<cdot> \<sigma>\<rangle> |_ hole_pos ?C" by (simp only: t Cp)
  also have "\<dots> = snd r \<cdot> \<sigma>" by (simp only: C, rule subt_at_hole_pos)
  finally have tp: "t |_ p = snd r \<cdot> \<sigma>" .
  from match_list_complete proj tp obtain \<sigma> where match: "match_list [(s |_ p, fst r), (t |_ p, snd r)] = Some \<sigma>" by force
  from r rr have list_ex: "list_ex (\<lambda> rule. r =v rule) R" by (auto simp: list_ex_iff)
  show "isOK(check_rstep R p r s t)"
    by (simp add: check_rstep_def r rr p' p match list_ex C del: match_list.simps)
qed

lemma check_rstep_sound_2[simp]:
  assumes "isOK (check_rstep R p r s t)"
  shows "(s, t) \<in> rstep (set R)"
using check_rstep_sound[OF assms] unfolding rstep_iff_rstep_r_p_s by auto

definition check_rrstep ::
  "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term
     \<Rightarrow> shows check"
where "check_rrstep R r s t \<equiv> check_rstep R \<epsilon> r s t"

lemma check_rrstep: "(\<exists> r \<in> set R. (isOK (check_rrstep R r s t))) = ((s, t) \<in> rrstep (set R))"
using check_rstep_sound[of R \<epsilon> _ s t] check_rstep_complete[of s t R _ \<epsilon>]
unfolding check_rrstep_def rrstep_def rstep_r_p_s_def 
by (auto, blast+)

lemma check_rrstep_sound: assumes "isOK(check_rrstep R r s t)" shows "(s,t) \<in> rrstep (set R)"
using check_rstep_sound[OF assms[unfolded check_rrstep_def]]
unfolding rrstep_def rstep_r_p_s_def
by auto

fun
  check_steps ::
    "('f::show, 'v::show) rules \<Rightarrow> ('f,'v) rules \<Rightarrow> ('f, 'v) prseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_steps P R [] s u = check (s = u) (
     ''the last term of the rewrite sequence'' +#+ shows_nl +@+ shows s +@+ shows_nl
     ''does not correspond to the goal term''  +#+ shows_nl +@+ shows u +@+ shows_nl
   )" |
  "check_steps P R ((_, r, True, t) # prts) s u = do {
     check_rrstep P r s t;
     check_steps P R prts t u
   }" |
  "check_steps P R ((p, r, False, t) # prts) s u = do {
     check_rstep R p r s t;
     check_steps P R prts t u
   }"

lemma check_steps_sound:
  assumes "isOK (check_steps P R rseq s t)"
  shows "(s, t) \<in> (rrstep (set P) \<union> rstep (set R))^^(length rseq)"
proof -
  obtain PR where PR: "PR = rrstep (set P) \<union> rstep (set R)" by auto
  from assms show ?thesis unfolding PR[symmetric] 
  proof (induct rseq arbitrary: s)
    case Nil thus ?case by simp
  next
    case (Cons prt rseq)
    obtain p r or u where prt: "prt = (p, r, or, u)" by (cases prt, blast) 
    from Cons[unfolded prt]
    have IH: "(u, t) \<in> PR ^^ (length rseq)" by (cases or, induct rseq) auto
    have "(s,u) \<in> PR"
    proof (cases or)
      case True
      with Cons[unfolded prt]
      have "isOK (check_rrstep P r s u)" by simp
      hence "(s,u) \<in> rrstep (set P)" by (rule check_rrstep_sound)
      thus ?thesis unfolding PR ..
    next
      case False
      with Cons[unfolded prt]
      have "isOK (check_rstep R p r s u)" by simp
      hence "(s,u) \<in> rstep (set R)" by (rule check_rstep_sound_2)
      thus ?thesis unfolding PR ..
    qed
    with IH show ?case unfolding prt o_def using rel_pow_Suc_I2[of s u PR] by simp
  qed
qed


definition
  check_rsteps ::
    "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_rsteps R prts s u \<equiv> check_steps [] R (map (\<lambda> (p,r,t). (p,r,False,t)) prts) s u"

lemma check_rsteps_sound[simp]:
  assumes "isOK (check_rsteps R rseq s t)"
  shows "(s, t) \<in> (rstep (set R))^^(length rseq)"
proof -
  have empty: "rrstep {} \<union> rstep (set R) = rstep (set R)" unfolding rrstep_def rstep_r_p_s_def by auto
  from check_steps_sound[OF assms[unfolded check_rsteps_def]] show ?thesis 
    by (auto simp: Let_def empty)
qed

definition
  check_supt :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supt s t \<equiv> check (s \<rhd> t) (shows t +@+ '' is not a proper subterm of '' +#+ shows s)"

definition
  check_supteq :: "('f::show, 'v::show) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_supteq s t \<equiv> check (s \<unrhd> t) (shows t +@+ '' is not a subterm of '' +#+ shows s)"

lemma isOK_check_supt[simp]:
  "isOK (check_supt s t) = ((s, t) \<in> op \<rhd>)"
unfolding check_supt_def using supt_def by auto

lemma isOK_check_supteq[simp]:
  "isOK (check_supteq s t) = ((s, t) \<in> op \<unrhd>)"
unfolding check_supteq_def using supteq_def by auto

definition
  check_strict ::
    "('f::show, 'v::show) rules \<Rightarrow>
     (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option) \<Rightarrow>
     'f proj \<Rightarrow> ('f, 'v) rule \<Rightarrow> shows check"
where
  "check_strict R rseqm p r = (
     let s = proj_term p (fst r); t = proj_term p (snd r) in
     case rseqm r of
       None \<Rightarrow> check_supt s t
     | Some rseq \<Rightarrow> (if length rseq = 0
       then check_supt s t
       else (
         let u = snd (snd (last rseq)) in (
           check_rsteps R rseq s u
         \<guillemotright> check_supteq u t
         )
       )
     )
   )"

lemma check_strict_sound[simp]:
  assumes "isOK (check_strict R rseqm p r)"
  shows "(proj_term p (fst r), proj_term p (snd r)) \<in> (op \<rhd> \<union> rstep (set R))^+"
    (is "(?s, ?t) \<in> (op \<rhd> \<union> rstep (set R))^+")
proof (cases "rseqm r")
  case None
  from assms[unfolded check_strict_def None]
    have "isOK (check_supt ?s ?t)" by simp
  hence "(?s, ?t) \<in> op \<rhd>" by simp
  thus ?thesis by auto
next
  case (Some rseq)
  show ?thesis
  proof (cases "length rseq = 0")
    case True
    with assms[unfolded check_strict_def Some]
      have "isOK (check_supt ?s ?t)" by simp
    hence "(?s, ?t) \<in> op \<rhd>" by simp
    thus ?thesis by auto
  next
    let ?u = "snd (snd (last rseq))"
    case False
    with assms[unfolded check_strict_def Some]
      have "isOK (check_rsteps R rseq ?s ?u \<guillemotright> check_supteq ?u ?t)" unfolding Let_def by simp
    hence "(?s, ?u) \<in> (rstep (set R))^^(length rseq)" and "(?u, ?t) \<in> op \<unrhd>" by simp_all
    with False obtain n where "(?s, ?u) \<in> (rstep (set R))^^(Suc n)" by (induct rseq) simp_all
    hence "(?s, ?u) \<in> (rstep (set R))^+" using trancl_power[of _ "rstep (set R)"] by blast
    from `(?u, ?t) \<in> op \<unrhd>` show ?thesis unfolding supteq_supt_conv
    proof
      assume "?u = ?t"
      from `(?s, ?u) \<in> (rstep (set R))^+`
        show ?thesis unfolding `?u = ?t` using trancl_union_right[of "rstep (set R)" "op \<rhd>"] by blast
    next
      assume "(?u, ?t) \<in> op \<rhd>"
      hence "(?u, ?t) \<in> (op \<rhd> \<union> rstep (set R))^+" by auto
      moreover from `(?s, ?u) \<in> (rstep (set R))^+`
        have "(?s, ?u) \<in> (op \<rhd> \<union> rstep (set R))^+" using trancl_union_right by blast
      ultimately show ?thesis by simp
    qed
  qed
qed

definition
  check_weak :: "'f proj \<Rightarrow> ('f::show, 'v::show) rule \<Rightarrow> shows check"
where
  "check_weak p r \<equiv> check (proj_term p (fst r) = proj_term p (snd r))
     (''the projected lhs is not equal to the projected rhs'' +#+ shows_nl)
   <+? (\<lambda>s. ''Could not orient rule '' +#+ shows_rule r +@+ '', since'' +#+ shows_nl +@+
          shows (proj_term p (fst r)) +@+ '' != '' +#+ shows (proj_term p (snd r)) +@+
          shows_nl +@+ s)"

lemma check_weak_sound[simp]:
  assumes "isOK (check_weak p r)" shows "proj_term p (fst r) = proj_term p (snd r)"
using assms unfolding check_weak_def by auto

fun
  check_subterm_criterion_proc ::
    "'f proj \<Rightarrow>
     (('f::show, 'v::show) rule \<Rightarrow> ('f::show, 'v::show) rseq option) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v) rule_map \<Rightarrow> shows check"
where
  "check_subterm_criterion_proc p rseqm P P' R None = OK"
| "check_subterm_criterion_proc p rseqm P P' R (Some rm) = do {
     check_allm (\<lambda>(l, r). check_no_var l \<guillemotright> check_no_var r \<guillemotright> check_no_defined_root rm r) P;
     check_allm (check_strict R rseqm p) (list_diff P P');
     check_allm (check_weak p) P'
   }"

lemma check_subterm_criterion_proc_sound[simp] :
  "ci_only_pairs_proc_impl (\<lambda>P R rm rrm P'. isOK (check_subterm_criterion_proc p rseqm P P' R rm))"
unfolding ci_only_pairs_proc_impl_def proof (intro allI impI)
  fix P R P' s t \<sigma> assume ok: "isOK (check_subterm_criterion_proc p rseqm P P' R (computeRuleMap R))"
    and mchain: "min_ichain (set P, set R) s t \<sigma>"
  hence "SN_elt (rstep (set R)) (t 0 \<cdot> \<sigma> 0)" by (simp add: minimal_cond_def)
  then obtain rm where rm: "computeRuleMap R = Some rm" using computeRuleMap_None_not_SN by (cases "computeRuleMap R", auto)
  let ?\<pi> = "\<lambda>t. proj_term p t"
  from ok[unfolded rm] 
    have cond: "\<forall>(l, r)\<in>set P. \<not> is_var l \<and> \<not> is_var r \<and> \<not> defined (set R) (the (root r), num_args r)" by (simp add: split_def check_no_defined_root_sound[OF rm])
  moreover have strict: "\<forall>(l, r)\<in>set P - set P'. (?\<pi> l, ?\<pi> r) \<in> (op \<rhd> \<union> rstep (set R))^+"
  proof -
    from ok[unfolded rm]
      have "isOK (check_allm (check_strict R rseqm p) (list_diff P P'))" by simp
      thus ?thesis unfolding split_def using check_strict_sound[of R rseqm p] by (simp, blast)
  qed
  moreover from ok[unfolded rm]
    have weak: "\<forall>(l, r)\<in>set P'. ?\<pi> l = ?\<pi> r" by (simp add: split_def)
  ultimately
    show "\<exists>i. ichain (set P', set R) (shift s i) (shift t i) (shift \<sigma> i)"
    using ci_subterm_criterion_proc[of p, unfolded ci_proc_def subterm_criterion_proc_def subterm_criterion_cond_def,
      THEN spec[of _ "set P"], THEN spec[of _ "set R"], THEN spec[of _ "set P'"], THEN spec[of _ "set R"]]
    using mchain by blast
qed

fun
  create_proj :: "'f projL \<Rightarrow> 'f proj"
where
  "create_proj (Projection p) f = (case lookup f p of None \<Rightarrow> 0 | Some n \<Rightarrow> n)"

fun
  create_rseq_map :: "(('f, 'v) rule \<times> ('f, 'v) rseq) list \<Rightarrow> (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option)"
where
  "create_rseq_map rseqs r = lookup r rseqs"

fun
  create_subterm_criterion_proc ::
    "'f projL \<Rightarrow> (('f, 'v) rule \<times> ('f, 'v) rseq) list \<Rightarrow>
     ('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v) rule_map
     \<Rightarrow> shows check"
where
  "create_subterm_criterion_proc p rseqs P rP R rm = (let \<pi> = create_proj p; rseqm = create_rseq_map rseqs in
     check_subterm_criterion_proc \<pi> rseqm P rP R rm)"

lemma subterm_criterion_proc_ci:
  "ci_only_pairs_proc_impl (\<lambda>P R rm rrm rP. isOK (create_subterm_criterion_proc p rseqs P rP R rm))"
proof -
  let ?p = "create_proj p"
  let ?rseqs = "create_rseq_map rseqs"
  have simpme: "create_subterm_criterion_proc p rseqs = check_subterm_criterion_proc ?p ?rseqs"
  by (rule ext, rule ext, rule ext, rule ext, simp add: Let_def)
  show ?thesis by (simp only: simpme, rule check_subterm_criterion_proc_sound)
qed

declare create_subterm_criterion_proc.simps[simp del]

end
