theory Bin_Tree
imports "../TermFun"
begin

hide_const (open) List.remdups

datatype tree = E | N tree nat tree

fun getmax where
  "getmax E = 0"
| "getmax (N l x E) = x"
| "getmax (N l x r) = getmax r"

fun delmax where
  "delmax E = E"
| "delmax (N l x E) = l"
| "delmax (N l x r) = N l x (delmax r)"

fun del where
  "del x E = E"
| "del x (N E y r) = 
   (if x < y then (N E y r)
    else if eq x y then del x r
    else N E y (del x r))"
| "del x (N l y r) = 
   (if x < y then N (del x l) y r
    else if x > y then N l y (del x r)
    else N (delmax (del x l)) (getmax (del x l)) (del x r))"

(* Induction proof that we are avoiding: *)

lemma size_delmax: "size (delmax t) \<le> size t"
by (induct t rule: delmax.induct) auto

lemma size_del: "size (del x t) \<le> size t"
proof (induct t arbitrary: x)
  case E show ?case by simp
next
  case (N l x r y) note IH = this
  note size_delmax[simp] = size_delmax
  show ?case
  proof (cases l)
    case E 
    with IH(2)[of y]
    show ?thesis by auto
  next
    case N
    have "size (delmax (del x l)) \<le> size (del x l)" by simp
    also have "... \<le> size l" by (rule IH)
    finally 
    show ?thesis
      unfolding N
      apply (unfold del.simps, fold N)
      apply (auto simp: IH)
      apply (insert IH(2)[of x]) 
      by auto
  qed
qed

function remdups where
  "remdups E = E"
| "remdups (N l x r) = 
   N (remdups (del x l)) x (remdups (del x r))"
by pat_completeness auto
(* termination by (size_change) *)

setup {* Term_Fun.register_datatypes ["Bin_Tree.tree"] *}
define_trs getmax.simps
define_trs delmax.simps
define_trs del.simps
define_trs (no_sim) remdups.psimps

termination remdups
apply (trs_reduce remdups)
apply (simp only: all_trs_defs set_list_conv)
export_termination_problem "Examples/remdups.xml" 
(* call termination prover to generate proof *)
apply (ceta_import_proof "Examples/proofs/remdups.proof.xml")
apply ceta_eval
done


end
