theory Examples
imports "../TermFun" Loga
begin

section {* exmaple from some IsaFoR talks *}

function evenodd :: "bool \<Rightarrow> nat \<Rightarrow> bool"
where
  "evenodd True 0 = False" |
  "evenodd False x = (\<not> evenodd True x)" |
  "evenodd True (Suc x) = evenodd False x"
  by (pat_completeness) auto

define_trs not_True_eq_False not_False_eq_True (* order is important, why? *)
define_trs (no_sim) evenodd.psimps
thm evenodd.trs_def

termination evenodd
  apply (trs_reduce evenodd)
  apply (simp only: all_trs_defs set_list_conv)
  export_termination_problem "Examples/evenodd.xml"
  apply (ceta_import_proof "Examples/proofs/evenodd.proof.xml")
  apply ceta_eval
done

section {* normalization of conditionals (cf. Paulson 1986) *}

datatype exp = AT nat | IF exp exp exp
setup {* Term_Fun.register_datatypes [@{type_name exp}] *}

function norm
where
  "norm (AT a) = AT a"
| "norm (IF (AT a) y z) = IF (AT a) (norm y) (norm z)"
| "norm (IF (IF u v w) y z) = norm (IF u (IF v y z) (IF w y z))"
by pat_completeness auto

define_trs (no_sim) norm.psimps

termination norm 
apply (trs_reduce norm)
apply (simp only: all_trs_defs set_list_conv)
export_termination_problem "Examples/norm.xml"
apply (ceta_import_proof "Examples/proofs/norm.proof.xml") 
apply ceta_eval
(*apply ceta_simp*)
done

section {* Preprocessing of formulas (Chaieb) *}

datatype num = C int | Bound nat | CN nat int num | Neg num | Add num num| Sub num num 
  | Mul int num | Floor num| CF int num num

setup {* Term_Fun.register_datatypes [@{type_name num}] *}

datatype fm  = 
  T| F| Lt num| Le num| Gt num| Ge num| Eq num| NEq num| Dvd int num| NDvd int num|
  NOT fm| And fm fm|  Or fm fm| Imp fm fm| Iff fm fm| E fm| A fm

setup {* Term_Fun.register_datatypes [@{type_name fm}] *}

function (sequential) prep :: "fm \<Rightarrow> fm"
where
  "prep (E T) = T"
| "prep (E F) = F"
| "prep (E (Or p q)) = Or (prep (E p)) (prep (E q))"
| "prep (E (Imp p q)) = Or (prep (E (NOT p))) (prep (E q))"
| "prep (E (Iff p q)) = Or (prep (E (And p q))) (prep (E (And (NOT p) (NOT q))))" 
| "prep (E (NOT (And p q))) = Or (prep (E (NOT p))) (prep (E(NOT q)))"
| "prep (E (NOT (Imp p q))) = prep (E (And p (NOT q)))"
| "prep (E (NOT (Iff p q))) = Or (prep (E (And p (NOT q)))) (prep (E(And (NOT p) q)))"
| "prep (E p) = E (prep p)"
| "prep (A (And p q)) = And (prep (A p)) (prep (A q))"
| "prep (A p) = prep (NOT (E (NOT p)))"
| "prep (NOT (NOT p)) = prep p"
| "prep (NOT (And p q)) = Or (prep (NOT p)) (prep (NOT q))"
| "prep (NOT (A p)) = prep (E (NOT p))"
| "prep (NOT (Or p q)) = And (prep (NOT p)) (prep (NOT q))"
| "prep (NOT (Imp p q)) = And (prep p) (prep (NOT q))"
| "prep (NOT (Iff p q)) = Or (prep (And p (NOT q))) (prep (And (NOT p) q))"
| "prep (NOT p) = NOT (prep p)"
| "prep (Or p q) = Or (prep p) (prep q)"
| "prep (And p q) = And (prep p) (prep q)"
| "prep (Imp p q) = prep (Or (NOT p) q)"
| "prep (Iff p q) = Or (prep (And p q)) (prep (And (NOT p) (NOT q)))"
| "prep p = p"
by pat_completeness simp_all

define_trs (no_sim) prep.psimps

termination prep
apply (trs_reduce prep)
apply (simp only: all_trs_defs set_list_conv)
export_termination_problem "Examples/prep.xml" 
apply (ceta_import_proof "Examples/proofs/prep.proof.xml", ceta_eval)
done

section {* Merge sort *}

function merge :: "nat list \<Rightarrow> nat list \<Rightarrow> nat list"
where
  "merge (x#xs) (y#ys) = (case x \<le> y of
    True  \<Rightarrow> x # merge xs (y#ys)
  | False \<Rightarrow> y # merge (x#xs) ys)"
| "merge xs [] = xs"
| "merge [] ys = ys"
by pat_completeness auto

define_trs (no_sim) merge.psimps

function (sequential) msort :: "nat list \<Rightarrow> nat list"
where
  "msort [] = []"
| "msort [x] = [x]"
| "msort xs = merge (msort (take (half (length xs)) xs))
	                  (msort (drop (half (length xs)) xs))"
by pat_completeness auto

define_trs (no_sim) msort.psimps

termination merge
apply (trs_reduce merge)
apply (simp only: all_trs_defs set_list_conv)
export_termination_problem "Examples/merge.xml"
apply (ceta_import_proof "Examples/proofs/merge.proof.xml")
apply ceta_eval
(*apply ceta_simp*)
done

(* sim rule for merge not needed below... *)

termination msort
apply (trs_reduce msort)
apply (simp only: all_trs_defs set_list_conv)
export_termination_problem "Examples/msort.xml"
oops (* no proof yet *)


end

