theory TermFun_Setup
imports "../Ceta_Checker"
uses
  ("ML/term.ML")
  ("ML/basics.ML") 
  ("ML/call_tree.ML") 
  ("ML/trs_data.ML")
  ("ML/tactics.ML")
  ("ML/embed.ML")
  ("ML/term_fun.ML")
  ("ML/xml.ML")
begin

section {* Preliminaries *}

(*<*)
consts
  emb_const :: "'a \<Rightarrow> (string, string) term list \<Rightarrow> (string, string) term"
  (infix "<>" 100)

setup {*
let
  fun quote (Const (@{const_name emb_const}, _) $ Const (n, _)) =
      SOME (@{term "Fun :: string \<Rightarrow> (string,string) term list \<Rightarrow> (string,string) term"} $ HOLogic.mk_string n)
    | quote _ = NONE

  fun unquote (Const (@{const_name Term.Fun}, _) $ n) =
      ((SOME (Const (@{const_name emb_const}, dummyT)
        $ Const(HOLogic.dest_string n, dummyT)))
      handle TERM _ => NONE)
    | unquote _ = NONE

  fun rew f ts ctxt =
    let 
      val thy = ProofContext.theory_of ctxt
      val ts' = map (Pattern.rewrite_term thy [] [f]) ts
    in
      if forall (op aconv) (ts ~~ ts') then NONE
      else SOME (ts', ctxt)
    end
in
  Context.theory_map 
    (Syntax.add_term_check ~1 "quote_funs" (rew quote)
    #> Syntax.add_term_uncheck 1 "quote_funs" (rew unquote))
end
*}
(*>*)

lemma length_eqI: "length xs = length ys \<Longrightarrow> length (x#xs) = length (y#ys)"
by auto

lemma pairwiseI1: "\<forall>i<length []. P i"
  and pairwiseI2: "(x, y) \<in> R
  \<Longrightarrow> \<forall>i<length xs. (xs ! i, ys ! i) \<in> R
  \<Longrightarrow> \<forall>i<length (x#xs). ((x#xs) ! i, (y#ys) ! i) \<in> R"
by auto (case_tac i, auto)

lemma list_memberI1: "x \<in> set (x # xs)"
  and list_memberI2: "y \<in> set xs \<Longrightarrow> y \<in> set (x # xs)"
by auto

lemma fixes f 
shows inv_imageI: "(f a, f b) \<in> R \<Longrightarrow> (a, b) \<in> inv_image R f" by auto

lemmas into_CC =  subsetD[OF subset_ctxt_closure]

lemma CS_into_rstep: "p \<in> CS R \<Longrightarrow> p \<in> rstep R"
unfolding rstep_CC_CS by (rule into_CC)

abbreviation 
  "sub_steps R \<equiv> (supt \<union> rstep R)^*"

lemma term_relI_initial:
  "(x, y) \<in> rstep R \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> (supt \<union> rstep R)^+"
by (rule rtrancl_into_trancl2) (rule UnI2)

lemma term_relI_step:
  "(x, y) \<in> rstep R \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> sub_steps R"
by (rule converse_rtrancl_into_rtrancl) (rule UnI2)

lemma term_relI_steps:
  "(x, y) \<in> (rstep R)^* \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> sub_steps R"
apply (rule rtrancl_trans)
apply (rule subsetD[OF rtrancl_mono])
by (rule Un_upper2)

lemma term_relI_sub:
  "(x, y) \<in> supt \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> sub_steps R"
by (rule converse_rtrancl_into_rtrancl) (rule UnI1)

class trs_rep =
  fixes embed :: "'a \<Rightarrow> (string, string) term" ("\<lceil>_\<rceil>") 

abbreviation dummy_term :: "(string, string) term"
where
  "dummy_term == Fun ''dummy_term'' []"
notation (output) dummy_term ("\<bullet>")

use "ML/term.ML"
use "ML/basics.ML"
use "ML/call_tree.ML"
use "ML/trs_data.ML"
use "ML/tactics.ML"
use "ML/embed.ML"
use "ML/xml.ML"


lemma catch_weak_cong: "x \<equiv> y \<Longrightarrow> try x catch z \<equiv> try y catch z" by auto
lemma check_weak_cong: "x \<equiv> y \<Longrightarrow> check x z \<equiv> check y z" by auto
lemma bind_weak_cong: "x \<equiv> y \<Longrightarrow> bind x z \<equiv> bind y z" by auto
lemma implies_weak_cong: "x \<equiv> y \<Longrightarrow> x \<longrightarrow> z \<equiv> y \<longrightarrow> z" by auto
lemma disj_weak_cong: "x \<equiv> y \<Longrightarrow> x \<or> z \<equiv> y \<or> z" by auto
lemma conj_weak_cong: "x \<equiv> y \<Longrightarrow> x \<and> z \<equiv> y \<and> z" by auto
lemma debug_cong: "debug x y z \<equiv> debug x y z" by auto

lemmas ceta_congs = 
  (* if, boolean operators *)
  if_weak_cong 
  implies_weak_cong
  disj_weak_cong
  conj_weak_cong
  (* cases *)
  sum.weak_case_cong  
  list.weak_case_cong 
  nat.weak_case_cong 
  bool.weak_case_cong 
  prod.weak_case_cong
  (* try .. catch, bind, check *)
  catch_weak_cong
  bind_weak_cong
  check_weak_cong
  (* debug *)
  debug_cong

(* if the following is enabled, then ceta_simp and ceta_checker_simp do not work any longer *)
(*
code_reflect Ceta_Checker
  datatypes TRSProof = "_" 
  and DPProof = "_" and redpair = "_"
  and projL = "_" and sl_inter = "_" and arithFun = "_" and redpair = "_"
  and af_entry = "_" and arctic_rat = "_" and arctic = "_" and tpoly = "_"
  and lab = "_" and nibble = "_" and sum = "_" and pos = "_"
  and subst = "_" and hterm = "_" and "term" = "_" 
  and scg = "_" and rat = "_" and nat = "_"
   (*FIXME:other types must also be mentioned here *)
  functions ceta_checker

(* remove code equation (already installed anyway), to avoid expensive collection of all other code equations *)
declare ceta_checker_def[code del]
*)

use "ML/term_fun.ML"

setup Term_Fun.setup

(* Example proof to test startup-time of ceta_eval.
Florian promised to have a look at how to speed this up *)
lemma "SN (rstep (set ([] :: (string,string)rules)))"
apply (rule ceta_checker[of _ "RisEmpty :: (string,label_type,string)TRSProof"])
apply ceta_eval
done

end