(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SemanticLabeling
imports SLFramework LexicographicExtension Substitution DpTransImpl Util2
begin

type_synonym  ('v, 'c) assign = "'v \<Rightarrow> 'c"
type_synonym  ('f, 'c) inter = "'f \<Rightarrow> 'c list \<Rightarrow> 'c"
type_synonym  ('f, 'l)labels = "'f \<Rightarrow> nat \<Rightarrow> 'l set" 
type_synonym  ('f, 'c, 'l)label = "'f \<Rightarrow> 'c list \<Rightarrow> 'l"
type_synonym  ('f,'l,'lf)lcompose = "'f \<Rightarrow> nat \<Rightarrow> 'l \<Rightarrow> 'lf"
type_synonym  ('lf,'f,'l)ldecompose = "'lf \<Rightarrow> ('f \<times> 'l)"

fun eval_lab :: "('f,'c)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('f,'l,'lf)lcompose \<Rightarrow> ('v,'c)assign \<Rightarrow> ('f,'v)term \<Rightarrow> ('c \<times> ('lf,'v)term)" 
where "eval_lab I L LC \<alpha> (Var x) = (\<alpha> x, Var x)"
  |   "eval_lab I L LC \<alpha> (Fun f ts) = 
     (let clts = map (eval_lab I L LC \<alpha>) ts;
          cs = map fst clts;
          c = I f cs;
          lts = map snd clts in     
     (c, Fun (LC f (length ts) (L f cs)) lts))"

fun lab_root :: "('f,'c)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('f,'l,'lf)lcompose \<Rightarrow> ('v,'c)assign \<Rightarrow> ('f,'v)term \<Rightarrow> ('lf,'v)term" 
where "lab_root I L L' LC \<alpha> (Fun f ts) = 
     (let clts = map (eval_lab I L LC \<alpha>) ts;
          cs = map fst clts;
          lts = map snd clts in     
     (Fun (LC f (length ts) (L' f cs)) lts))"
  | "lab_root _ _ _ _ _ (Var x) = (Var x)"

fun lab_root_all :: "('f,'c)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('f,'l,'lf)lcompose \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> 'l \<Rightarrow> 'l \<Rightarrow> bool) \<Rightarrow> ('v,'c)assign \<Rightarrow> ('f,'v)term \<Rightarrow> ('lf,'v)term set" 
where "lab_root_all I L L' LC lge \<alpha> (Fun f ts) = 
     (let clts = map (eval_lab I L LC \<alpha>) ts;
          cs = map fst clts;
          lts = map snd clts;
          l   = L' f cs in     
     {(Fun (LC f (length ts) ls) lts) | ls. lge f (length ts) l ls})"
  | "lab_root_all _ _ _ _ _ _ (Var x) = {(Var x)}"

abbreviation eval where "eval I L LC \<alpha> t \<equiv> fst (eval_lab I L LC \<alpha> t)"
abbreviation lab where "lab I L LC \<alpha> t \<equiv> snd (eval_lab I L LC \<alpha> t)"
definition wf_assign where "wf_assign C \<alpha> \<equiv> \<forall> x. \<alpha> x \<in> C"
definition lab_rule where "lab_rule I L LC C lr \<equiv> {(lab I L LC \<alpha> (fst lr), lab I L LC \<alpha> (snd lr)) | \<alpha> . wf_assign C \<alpha>}"
definition lab_root_rule where "lab_root_rule I L L' LC C lr \<equiv> {(lab_root I L L' LC \<alpha> (fst lr), lab_root I L L' LC \<alpha> (snd lr)) | \<alpha> . wf_assign C \<alpha>}"
definition lab_root_all_rule where "lab_root_all_rule I L L' LC lge C rule \<equiv> {(lab_root I L L' LC \<alpha> (fst rule), lr) | \<alpha> lr. wf_assign C \<alpha> \<and> lr \<in> lab_root_all I L L' LC lge \<alpha> (snd rule)}"
abbreviation lab_trs where "lab_trs I L LC C R \<equiv> \<Union>(lab_rule I L LC C ` R)"
abbreviation lab_root_trs where "lab_root_trs I L L' LC C R \<equiv> \<Union>(lab_root_rule I L L' LC C ` R)"
abbreviation lab_root_all_trs where "lab_root_all_trs I L L' LC lge C R \<equiv> \<Union>(lab_root_all_rule I L L' LC lge C ` R)"
definition wf_inter where "wf_inter I C \<equiv> \<forall> f cs. set cs \<subseteq> C \<longrightarrow> I f cs \<in> C"
definition wf_label where "wf_label L LS C \<equiv> \<forall> f cs. set cs \<subseteq> C \<longrightarrow> L f cs \<in> LS f (length cs)"
abbreviation qmodel_rule where "qmodel_rule I L LC C cge l r \<equiv> \<forall> \<alpha>. wf_assign C \<alpha> \<longrightarrow> cge (eval I L LC \<alpha> l) (eval I L LC \<alpha> r)"
definition qmodel where "qmodel I L LC C cge R \<equiv> \<forall> (l,r) \<in> R. qmodel_rule I L LC C cge l r"
definition cge_wm: "cge_wm I C cge \<equiv> \<forall> f bef c d aft. (set ([c,d] @ bef @ aft) \<subseteq> C \<and> cge c d \<longrightarrow> cge (I f (bef @ c # aft)) (I f (bef @ d # aft)))"
definition lge_wm: "lge_wm I L C cge lge \<equiv> \<forall> f bef c d aft. (set ([c,d] @ bef @ aft) \<subseteq> C \<and> cge c d \<longrightarrow> lge f (Suc (length bef + length aft)) (L f (bef @ c # aft)) (L f (bef @ d # aft)))"



definition
  decr_of_ord :: "('f \<Rightarrow> nat \<Rightarrow> 'l ars) \<Rightarrow> ('f,'l,'lf)lcompose \<Rightarrow> ('f,'l)labels \<Rightarrow> ('lf, 'v) trs"
where
  "decr_of_ord gr LC LS \<equiv> {(Fun (LC f (length ts) l) ts, Fun (LC f (length ts) l') ts) | f l l' ts. l \<in> LS f (length ts) \<and> l' \<in> LS f (length ts) \<and> (l,l') \<in> gr f (length ts)}"

definition lge_to_lgr :: "('f \<Rightarrow> nat \<Rightarrow> 'l \<Rightarrow> 'l \<Rightarrow> bool) \<Rightarrow> ('f,'l)labels \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> 'l \<Rightarrow> 'l \<Rightarrow> bool" 
  where "lge_to_lgr lge LS f n \<equiv> let LSfn = LS f n; lgefn = lge f n in (\<lambda> l l'. l \<noteq> l' \<and> l \<in> LSfn \<and> l' \<in> LSfn \<and> lgefn l l')"
definition lge_to_lgr_rel :: "('f \<Rightarrow> nat \<Rightarrow> 'l \<Rightarrow> 'l \<Rightarrow> bool) \<Rightarrow> ('f,'l)labels \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> 'l ars"
  where "lge_to_lgr_rel lge LS f n \<equiv> {(l,l')| l l'. lge_to_lgr lge LS f n l l'}"

locale sl_interpr = 
  fixes C :: "'c set"
  and  c :: "'c"
  and  I :: "('f,'c)inter"
  and  cge :: "'c \<Rightarrow> 'c \<Rightarrow> bool"
  and  lge :: "'f \<Rightarrow> nat \<Rightarrow> 'l \<Rightarrow> 'l \<Rightarrow> bool"
  and  L :: "('f,'c,'l)label"
  and  LC :: "('f,'l,'lf)lcompose"
  and  LD :: "('lf,'f,'l)ldecompose"
  and  LS :: "('f,'l)labels" 
  assumes c: "c \<in> C"
  and wf_I: "wf_inter I C" 
  and wf_L: "wf_label L LS C" 
  and wm_cge: "cge_wm I C cge"
  and wm_lge: "lge_wm I L C cge lge"
  and gr_SN: "\<And> f n. SN (lge_to_lgr_rel lge LS f n)"
  and LD_LC: "\<And> f n l. LD (LC f n l) = (f,l)"

locale sl_interpr_root = sl_interpr C c I cge lge L LC LD LS
  for C c I cge lge L and LC :: "('f,'l,'lf)lcompose" and LD LS +
  fixes 
       L' :: "('f,'c,'l)label"
  and  LS' :: "('f,'l)labels" 
  assumes
      wf_L': "wf_label L' LS' C" 
  and wm_lge': "lge_wm I L' C cge lge"
  and lge_trans: "\<And> f n x y z. \<lbrakk>lge f n x y; lge f n y z\<rbrakk> \<Longrightarrow> lge f n x z"
  and lge_refl: "\<And> f n x. x \<in> LS' f n \<Longrightarrow> lge f n x x"
  and gr'_SN: "\<And> f n. SN (lge_to_lgr_rel lge (\<lambda> f n. LS f n \<union> LS' f n) f n)"


context sl_interpr
begin
abbreviation gr where "gr \<equiv> lge_to_lgr_rel lge LS"
abbreviation Eval where "Eval \<equiv> eval I L LC"
abbreviation Lab where "Lab \<equiv> lab I L LC"
abbreviation Lab_trs :: "('f,'v)trs \<Rightarrow> ('lf,'v)trs" where "Lab_trs \<equiv> lab_trs I L LC C"
abbreviation wf_ass where "wf_ass \<equiv> wf_assign C"
definition subst_ass where "subst_ass \<alpha> \<sigma> \<equiv> \<lambda> x. Eval \<alpha> (get_subst \<sigma> x)"
definition lab_subst where "lab_subst \<alpha> \<sigma> \<equiv> Subst (\<lambda> x. Lab \<alpha> (get_subst \<sigma> x))"
abbreviation Decr :: "('lf,'v)trs" where "Decr \<equiv> decr_of_ord gr LC LS"
definition default_ass where "default_ass x = c"
abbreviation LAB where "LAB t \<equiv> Lab default_ass t"
end

context sl_interpr_root
begin
abbreviation LS_both where "LS_both \<equiv> (\<lambda> f n. LS f n \<union> LS' f n)"
abbreviation gr_root where "gr_root \<equiv> lge_to_lgr_rel lge LS'"
abbreviation gr_both where "gr_both \<equiv> lge_to_lgr_rel lge LS_both"
abbreviation Lab_root where "Lab_root \<equiv> lab_root I L L' LC"
abbreviation Lab_root_all where "Lab_root_all \<equiv> lab_root_all I L L' LC lge"
abbreviation Lab_root_trs :: "('f,'v)trs \<Rightarrow> ('lf,'v)trs" where "Lab_root_trs \<equiv> lab_root_trs I L L' LC C"
abbreviation Lab_root_all_trs :: "('f,'v)trs \<Rightarrow> ('lf,'v)trs" where "Lab_root_all_trs \<equiv> lab_root_all_trs I L L' LC lge C"
abbreviation Decr_root :: "('lf,'v)trs" where "Decr_root \<equiv> decr_of_ord gr_root LC LS'"
abbreviation Decr_both :: "('lf,'v)trs" where "Decr_both \<equiv> decr_of_ord gr_both LC LS_both"
abbreviation LAB_root where "LAB_root t \<equiv> Lab_root default_ass t"
end

context sl_interpr
begin

lemma Decr_SN_generic: assumes grSN: "\<And> f n. SN (lge_to_lgr_rel lge LSg f n)"
  shows "SN (rstep (decr_of_ord (lge_to_lgr_rel lge LSg) LC LSg))" (is "SN (rstep ?Decr)")
proof -
  let ?gr = "lge_to_lgr_rel lge LSg"
  let ?id = "id :: 'lf \<Rightarrow> 'lf"
  show ?thesis 
  proof(rule ccontr)
    assume nSN: "\<not> ?thesis"
    have wf_trs: "wf_trs ?Decr" unfolding decr_of_ord_def wf_trs_def by auto
    obtain s t \<sigma> where ichain: "ichain (DP ?id ?Decr, ?Decr) s t \<sigma>"
      using not_SN_imp_ichain[OF wf_trs nSN] by blast
    have DP: "DP ?id ?Decr \<subseteq> {(Fun ((LC f (length ts) l)) ts, Fun ((LC f (length ts) l')) ts) | f l l' (ts :: ('lf,'a)term list). (l,l') \<in> ?gr f (length ts)}" (is "_ \<subseteq> ?DP")
    proof -
      {
        fix f l l' ts h and us :: "('lf,'a) term list"
        assume gr: "(l,l') \<in> ?gr f (length ts)" and right: "Fun (LC f (length ts) l') ts \<unrhd> Fun h us" and left: "\<not> (Fun (LC f (length ts) l) ts \<rhd> Fun h us)"
        let ?n = "length ts"
        have "(sharp_term id (Fun (LC f ?n l) ts), sharp_term id (Fun h us)) \<in> ?DP" 
        proof (cases "Fun h us = Fun (LC f ?n l') ts")
          case True
          with gr show ?thesis by auto
        next
          case False
          with right have "Fun (LC f ?n l') ts \<rhd> Fun h us" by auto
          hence "Fun (LC f ?n l) ts \<rhd> Fun h us"
          proof (cases, (auto)[1])          
            case (subt s)
            thus ?thesis ..
          qed
          with left show ?thesis by blast
        qed
      }
      thus ?thesis unfolding DP_def decr_of_ord_def by blast
    qed
    from ichain DP have DP: "\<And> i. (s i, t i) \<in> ?DP" by simp blast
    obtain l where l: "l = (\<lambda> t :: ('lf,'a)term. case t of Fun lf _ \<Rightarrow> snd (LD lf))" by auto
    obtain n where n: "n = (\<lambda> i. num_args (s i))" by auto
    obtain f where f: "f = (\<lambda> i. case (s i) of Fun lf _ \<Rightarrow> fst (LD lf))" by auto
    {
      fix i :: nat
      from DP[of i] obtain fi li li' usi where si: "s i = Fun ((LC fi (length usi) li')) usi" and ti: "t i = Fun ((LC fi (length usi) li)) usi" by blast
      let ?n = "length usi"
      from DP[of i] obtain f' ll l' where "LC fi (length usi) li' = LC f' (length usi) l' \<and> LC fi (length usi) li = LC f' (length usi) ll \<and> (l', ll) \<in> ?gr f' (length usi)" unfolding si ti by auto
      hence id: "LD (LC fi (length usi) li') = LD (LC f' (length usi) l') \<and> LD (LC fi (length usi) li) = LD (LC f' (length usi) ll)" and gr: "(l', ll) \<in> ?gr f' (length usi)" by auto
      from id[unfolded LD_LC] gr have labels: "(li',li) \<in> ?gr fi ?n"  by simp
      have lti: "l (t i) = li" unfolding l ti using LD_LC by simp
      have lsi: "l (s i) = li'" unfolding l si using LD_LC by simp
      from ichain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> (\<sigma> (Suc i))) \<in> (rstep ?Decr)^*" by simp
      hence steps: "(Fun ((LC fi ?n li)) (map (\<lambda> u. u \<cdot> \<sigma> i) usi), s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep ?Decr)^*" unfolding ti by simp
      have nvar: "\<forall> lr \<in> ?Decr. \<forall> x. fst lr \<noteq> Var x" unfolding decr_of_ord_def map_trs.simps by auto
      obtain t where t: "s (Suc i) \<cdot> \<sigma> (Suc i) = t" by auto
      obtain ts where ts: "map (\<lambda> u. u \<cdot> \<sigma> i) usi = ts" and lts: "length usi = length ts" by auto
      from steps have "\<exists> ssi li'. s (Suc i) \<cdot> \<sigma> (Suc i) = Fun ((LC fi ?n li')) ssi \<and> length ssi = ?n \<and> (li,li') \<in> (?gr fi ?n)^*" 
        unfolding t ts lts 
      proof (induct)
        case base
        show ?case by auto
      next
        case (step u s)
        from step(3) obtain vs li' where u: "u = Fun (LC fi (length ts) li') vs" and len: "length vs = length ts" and li': "(li,li') \<in> (?gr fi (length ts))^*" by auto
        from step(2) obtain C l r \<sigma> where ul: "u = C\<langle>l \<cdot> \<sigma>\<rangle>" and sr: "s = C\<langle>r \<cdot> \<sigma>\<rangle>" and lr: "(l,r) \<in> ?Decr" by auto
        show ?case
        proof (cases C)
          case (More g bef D aft)
          let ?ts = "(bef @ D\<langle>r\<cdot>\<sigma>\<rangle> # aft)"
          from u[unfolded ul] len have "s = Fun (LC fi (length ts) li') ?ts" and "length ?ts = length ts" unfolding sr More by auto
          with li' show ?thesis by auto
        next
          case Hole
          hence ul: "u = l \<cdot> \<sigma>" and sr: "s = r \<cdot> \<sigma>" using ul sr by auto
          from lr[unfolded decr_of_ord_def] obtain f l1 l2 ls where l: "l = Fun (LC f (length ls) l1) ls" and r: "r = Fun (LC f (length ls) l2) ls" and l1l2: "(l1,l2) \<in> ?gr f (length ls)"
            by auto
          from ul[unfolded u l] len have len: "length ls = length ts" and f: "LC fi (length ts) li' = LC f (length ts) l1" (is "?l = ?l1") by auto
          hence "LD ?l = LD ?l1" by auto
          from this[unfolded LD_LC] have f: "f = fi" and l1: "l1 = li'" by auto           
          from li' l1l2[unfolded len f l1] have "(li,l2) \<in> (?gr fi (length ts))^*" by auto
          with sr[unfolded r f len] len show ?thesis by auto
        qed
      qed
      then obtain ssi li'' where "s (Suc i) \<cdot> \<sigma> (Suc i) = Fun ((LC fi ?n li'')) ssi" and "length ssi = ?n" and li'': "(li,li'') \<in> (?gr fi ?n)^*" by auto
      with DP[of "Suc i"] obtain ssi where ssi: "s (Suc i) = Fun ((LC fi ?n li'')) ssi" and len: "length ssi = ?n" by auto
      have lssi: "l (s (Suc i)) = li''" unfolding l ssi using LD_LC by simp
      from labels li'' have "(l (s i), l (s (Suc i))) \<in> (?gr (f i) (n i))^+" and "n (Suc i) = n i" and "f (Suc i) = f i" unfolding lsi lssi lti n f
        unfolding si ssi by (auto simp: len LD_LC)
    } note main = this
    {
      fix i
      have "n i = n 0 \<and> f i = f 0"
        by (induct i, auto simp: main)
      with main(1)[of i] have "(l (s i), l (s (Suc i))) \<in> (?gr (f 0) (n 0))^+" by simp
    }
    with SN_imp_SN_trancl[OF grSN[of "f 0" "n 0"]] show False unfolding SN_defs by force
  qed
qed

lemma Decr_SN: "SN (rstep Decr)"
  by (rule Decr_SN_generic, rule gr_SN)

lemma wf_default_ass: "wf_ass default_ass" 
  unfolding default_ass_def wf_assign_def using c by auto

lemma eval_subst[simp] :
  "Eval \<alpha> (t \<cdot> \<sigma>) = Eval (subst_ass \<alpha> \<sigma>) t"
proof (induct t, cases \<sigma>, simp add: subst_ass_def)
  case (Fun f ts)
  let ?beta = "subst_ass \<alpha> \<sigma>"
  have evalInd: "map (\<lambda> x. Eval \<alpha> (x \<cdot> \<sigma>)) ts = map (Eval ?beta) ts" using Fun  by auto
  have "Eval \<alpha> (Fun f ts \<cdot> \<sigma>) = Eval \<alpha> (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" by simp
  also have "\<dots> = Eval ?beta (Fun f ts)" by (simp add: Let_def o_def evalInd)
  finally show ?case .
qed

lemma lab_subst[simp] :
  "Lab \<alpha> (t \<cdot> \<sigma>) = Lab (subst_ass \<alpha> \<sigma>) t \<cdot> lab_subst \<alpha> \<sigma>"
  by (induct t, cases \<sigma>, simp add: lab_subst_def, simp add: Let_def o_def)
end

context sl_interpr_root
begin

lemma Decr_both: "Decr \<union> Decr_root \<subseteq> Decr_both"
proof -
  {
    fix l r f n
    assume "(l,r) \<in> gr f n"
    hence "(l,r) \<in> gr_both f n" unfolding lge_to_lgr_rel_def lge_to_lgr_def
      by (auto simp: Let_def)
  } note subset = this
  {
    fix l r f n
    assume "(l,r) \<in> gr_root f n"
    hence "(l,r) \<in> gr_both f n" unfolding lge_to_lgr_rel_def lge_to_lgr_def
      by (auto simp: Let_def)
  } note subset_root = this
  show "Decr \<union> Decr_root \<subseteq> Decr_both" (is "?DD \<subseteq> _")   
  proof 
    fix s t
    assume "(s,t) \<in> ?DD"    
    thus "(s,t) \<in> Decr_both"
    proof
      assume "(s,t) \<in> Decr"
      thus ?thesis unfolding decr_of_ord_def using subset by blast
    next
      assume "(s,t) \<in> Decr_root"
      thus ?thesis unfolding decr_of_ord_def using subset_root by blast
    qed
  qed
qed

lemma Decr_SN_both: "SN (rstep Decr_both)"
  by (rule Decr_SN_generic[OF gr'_SN] )

lemma Decr_SN_both_2: "SN (rstep (Decr \<union> Decr_root))"
  by (rule SN_subset[OF Decr_SN_both rstep_mono[OF Decr_both]])

lemma lab_root_subst[simp]:    
  "\<not> is_var t \<Longrightarrow> Lab_root \<alpha> (t \<cdot> \<sigma>) = Lab_root (subst_ass \<alpha> \<sigma>) t \<cdot> lab_subst \<alpha> \<sigma>"
  by (cases t, auto simp add: o_def)
end

context sl_interpr
begin
lemma wf_term[simp]: assumes "wf_ass \<alpha>"
  shows "Eval \<alpha> t \<in> C"
proof (induct t)
  case (Var x)
  with assms show ?case by (auto simp: wf_assign_def)
next
  case (Fun f ts)
  let ?cs = "map (Eval \<alpha>) ts"
  let ?c = "I f ?cs"
  from Fun have "set ?cs \<subseteq> C" by auto
  with wf_I have "?c \<in> C" unfolding wf_inter_def by (best)
  thus ?case by (simp add: Let_def o_def)
qed

lemma wf_ass_subst_ass[simp]: assumes "wf_ass \<alpha>"
  shows "wf_ass (subst_ass \<alpha> \<sigma>)"
using assms unfolding subst_ass_def by (simp add: wf_assign_def)


lemma [simp]: "Eval \<alpha> (t\<cdot>\<sigma>) = Eval (subst_ass \<alpha> \<sigma>) t" by (cases \<sigma>) (auto simp: subst_ass_def)

lemma quasi_sem_rewrite: assumes step: "(s,t) \<in> rstep R"
  and qmodel: "qmodel I L LC C cge R"
  and wfass: "wf_ass \<alpha>"
  shows "cge (Eval \<alpha> s) (Eval \<alpha> t)"
using step
proof
  fix l r C' \<sigma> assume in_R: "(l, r) \<in> R" and s: "s = C'\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C'\<langle>r\<cdot>\<sigma>\<rangle>"
  let ?beta = "subst_ass \<alpha> \<sigma>"
  have wfbeta: "wf_ass ?beta" using wfass by auto
  have "Eval \<alpha> (l\<cdot>\<sigma>) = Eval ?beta l" by simp
  also have "cge \<dots> (Eval ?beta r)" using in_R wfbeta qmodel by (auto simp: qmodel_def)
  also have "\<dots> = Eval \<alpha> (r\<cdot>\<sigma>)" by simp
  finally have root: "cge (Eval \<alpha> (l\<cdot>\<sigma>)) (Eval \<alpha> (r\<cdot>\<sigma>))" .
  have "Eval \<alpha> s = Eval \<alpha> (C'\<langle>l\<cdot>\<sigma>\<rangle>)" by (simp add: s)
  also have "cge \<dots> (Eval \<alpha> (C'\<langle>r\<cdot>\<sigma>\<rangle>))" 
  proof (induct C', simp add: root[simplified])
    case (More f bef D aft)
    show ?case
    proof (simp add: Let_def, rule wm_cge[unfolded cge_wm, THEN spec, THEN spec, THEN spec, THEN spec, THEN spec, THEN mp], 
      rule conjI[OF _ More], rule subsetI)
      fix x 
      assume "x \<in> set ([Eval \<alpha> D\<langle>l \<cdot> \<sigma>\<rangle>, Eval \<alpha> D\<langle>r \<cdot> \<sigma>\<rangle>] @ map (fst \<circ> eval_lab I L LC \<alpha>) bef @ map (fst \<circ> eval_lab I L LC \<alpha>) aft)"
      hence "x \<in> { Eval \<alpha> u | u. True}" by auto            
      with wf_term[OF wfass] show "x \<in> C" by blast
    qed
  qed
  also have "\<dots> = Eval \<alpha> t" by (simp add: t)
  finally show ?thesis .
qed

lemma lab_rrewrite: assumes step: "(s,t) \<in> CS R"
  and wfass: "wf_ass \<alpha>"
  shows "(Lab \<alpha> s, Lab \<alpha> t) \<in> CS (Lab_trs R)"
proof -
  from subst_closed_decompose[OF step] obtain l r \<sigma>
    where rule: "(l,r) \<in> R" and s: "s = l \<cdot> \<sigma>" and t: "t = r \<cdot> \<sigma>"    
    by auto
  let ?beta = "subst_ass \<alpha> \<sigma>"
  let ?delt = "lab_subst \<alpha> \<sigma>"
  have wfbeta: "wf_ass ?beta" using wfass by auto
  from rule wfbeta
  have "(Lab ?beta l, Lab ?beta r) \<in> (Lab_trs R)" by (simp add: lab_rule_def, force)
  hence "(Lab ?beta l \<cdot> ?delt, Lab ?beta r \<cdot> ?delt) \<in> CS (Lab_trs R)" .. 
  thus ?thesis unfolding s t by simp
qed

lemma quasi_lab_rewrite: assumes step: "(s,t) \<in> rstep R"
  and qmodel: "qmodel I L LC C cge R"
  and wfass: "wf_ass \<alpha>"
  shows "(Lab \<alpha> s, Lab \<alpha> t) \<in> (rstep Decr)^* O rstep (Lab_trs R)"
using step proof
  fix l r C' \<sigma>
  assume rule: "(l, r) \<in> R" and s: "s = C'\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C'\<langle>r \<cdot> \<sigma>\<rangle>"
  have "(Lab \<alpha> C'\<langle>l \<cdot> \<sigma>\<rangle>, Lab \<alpha> C'\<langle>r \<cdot> \<sigma>\<rangle>) \<in> (rstep Decr)^* O rstep (Lab_trs R)" 
  proof (induct C')
    case Hole
    from rule have step: "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> CS R" ..
    show ?case using lab_rrewrite[OF step wfass]  using subset_ctxt_closure[of "CS (Lab_trs R)"] unfolding rstep_CC_CS  by auto
  next
    case (More f ss D ts)
    let ?n = "Suc (length ss + length ts)"
    let ?D = "More (LC f ?n (L f (map (Eval \<alpha>) (ss @ D\<langle>l \<cdot> \<sigma>\<rangle> # ts)))) (map (Lab \<alpha>) ss) \<box> (map (Lab \<alpha>) ts)"
    let ?E = "More (LC f ?n (L f (map (Eval \<alpha>) (ss @ D\<langle>r \<cdot> \<sigma>\<rangle> # ts)))) (map (Lab \<alpha>) ss) \<box> (map (Lab \<alpha>) ts)"
    from rule have "(D\<langle>l \<cdot> \<sigma>\<rangle>, D\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R" using ctxt_closureI by auto
    hence cge: "cge (Eval \<alpha> D\<langle>l \<cdot> \<sigma>\<rangle>) (Eval \<alpha> D\<langle>r \<cdot> \<sigma>\<rangle>)" using qmodel wfass by (rule quasi_sem_rewrite)
    have id: "(Lab \<alpha> (More f ss D ts)\<langle>l \<cdot> \<sigma>\<rangle>,Lab \<alpha> (More f ss D ts)\<langle>r \<cdot> \<sigma>\<rangle>) =
      (?D\<langle>Lab \<alpha> D\<langle>l \<cdot> \<sigma>\<rangle>\<rangle>, ?E\<langle>Lab \<alpha> D\<langle>r \<cdot> \<sigma>\<rangle>\<rangle>)" (is "_ = (?l,?r)") 
      by (auto simp: Let_def o_def)
    {
      fix c
      assume "c \<in> set ([Eval \<alpha> D\<langle>l \<cdot> \<sigma>\<rangle>, Eval \<alpha> D\<langle>r \<cdot> \<sigma>\<rangle>] @ map (Eval \<alpha>) ss @ map (Eval \<alpha>) ts)"
      hence "c \<in> { Eval \<alpha> u | u. True}" by auto            
      with wf_term[OF wfass] have "c \<in> C" by blast
    } note inC = this
    have lge: "lge f ?n (L f (map (Eval \<alpha>) (ss @ D\<langle>l \<cdot> \<sigma>\<rangle> # ts))) (L f (map (Eval \<alpha>) (ss @ D\<langle>r \<cdot> \<sigma>\<rangle> # ts)))" (is "lge _ _ ?ll ?lr")
    by (simp, rule wm_lge[unfolded lge_wm, THEN spec, THEN spec[of _ "map (Eval \<alpha>) ss"], THEN spec, THEN spec, 
        THEN spec[of _ "map (Eval \<alpha>) ts"], THEN mp, simplified length_map], rule conjI[OF _ cge], rule subsetI, rule inC, simp)
    from More have "(?E\<langle>Lab \<alpha> D\<langle>l \<cdot> \<sigma>\<rangle>\<rangle>, ?E\<langle>Lab \<alpha> D\<langle>r \<cdot> \<sigma>\<rangle>\<rangle>) \<in> CC ((rstep Decr)^* O rstep (Lab_trs R))" (is "(?l',_) \<in> _")
      using ctxt_closureI[where C = ?E] by auto
    hence "(?l',?r) \<in> ((CC (rstep Decr))^* O CC (rstep (Lab_trs R)))" using CC_O_distrib CC_rtrancl_distrib by blast
    hence step: "(?l',?r) \<in> (rstep Decr)^* O rstep (Lab_trs R)" using ctxt_closed_rstep unfolding ctxt_closed_def by auto
    have steps: "(?l,?l') \<in> (rstep Decr)^*"
    proof (cases "?ll = ?lr")
      case True
      thus ?thesis by simp
    next
      case False
      have ll: "?ll \<in> LS f ?n" 
        by (simp, rule wf_L[unfolded wf_label_def, THEN spec, THEN spec[of _ "map (Eval \<alpha>) (ss @ D\<langle>l \<cdot> \<sigma>\<rangle> # ts)"], THEN mp, simplified], auto simp: inC)
      have lr: "?lr \<in> LS f ?n" 
        by (simp, rule wf_L[unfolded wf_label_def, THEN spec, THEN spec[of _ "map (Eval \<alpha>) (ss @ D\<langle>r \<cdot> \<sigma>\<rangle> # ts)"], THEN mp, simplified], auto simp: inC)
      from False lge have "(?ll,?lr) \<in> gr f ?n" unfolding lge_to_lgr_rel_def lge_to_lgr_def Let_def using ll lr by auto
      hence "(?l,?l') \<in> Decr" unfolding decr_of_ord_def using ll lr by auto
      hence "(?l,?l') \<in> rstep Decr" ..
      thus ?thesis by simp
    qed
    from step steps have "(?l,?r) \<in> ((rstep Decr)^* O (rstep Decr)^*) O rstep (Lab_trs R)" by blast
    hence "(?l,?r) \<in> (rstep Decr)^* O rstep (Lab_trs R)" unfolding rtrancl_idemp_self_comp .
    with id show ?case by auto
  qed
  with s t show ?thesis by auto
qed
end


context sl_interpr_root
begin

lemma quasi_lab_nrrewrite: assumes step: "(s,t) \<in> nrrstep R"
  and qmodel: "qmodel I L LC C cge R"
  and wfass: "wf_ass \<alpha>"
  shows "(Lab_root \<alpha> s, Lab_root \<alpha> t) \<in> Decr_root^= O (nrrstep Decr)^* O nrrstep (Lab_trs R)"
proof -
  from nrrstep_args[OF step]
  obtain f ss ts i where s: "s = Fun f ss" and t: "t = Fun f ts" and len: "length ss = length ts" and i: "i < length ss"
    and step: "(ss ! i, ts ! i) \<in> rstep R" and id: "\<And> j. \<lbrakk>j < length ss; j \<noteq> i\<rbrakk> \<Longrightarrow> ss ! j = ts ! j" by force+
  obtain bef aft si ti where bef: "bef = take i ss" and aft: "aft = drop (Suc i) ss" and si: "si = ss ! i" and ti: "ti = ts ! i" by simp
  have ss: "ss = bef @ si # aft" unfolding bef aft si by (rule id_take_nth_drop[OF i])
  from quasi_lab_rewrite[OF step qmodel wfass] 
  have lstep: "(Lab \<alpha> si, Lab \<alpha> ti) \<in> (rstep Decr)^* O rstep (Lab_trs R)" unfolding si ti .
  from quasi_sem_rewrite[OF step qmodel wfass]
  have cge: "cge (Eval \<alpha> si) (Eval \<alpha> ti)" unfolding si ti .
  let ?n = "Suc (length bef + length aft)"
  let ?C = "\<lambda> t. More (LC f ?n (L' f (map (Eval \<alpha>) (bef @ t # aft)))) (map (Lab \<alpha>) bef) \<box> (map (Lab \<alpha>) aft)"
  let ?D = "?C si"
  let ?E = "?C ti"
  {
    fix c
    assume "c \<in> set ([Eval \<alpha> si, Eval \<alpha> ti] @ map (Eval \<alpha>) bef @ map (Eval \<alpha>) aft)"
    hence "c \<in> { Eval \<alpha> u | u. True}" by auto            
    with wf_term[OF wfass] have "c \<in> C" by blast
  } note inC = this  
  have lge: "lge f ?n (L' f (map (Eval \<alpha>) (bef @ si # aft))) (L' f (map (Eval \<alpha>) (bef @ ti # aft)))" (is "lge _ _ ?ll ?lr")
    by (simp, rule wm_lge'[unfolded lge_wm, THEN spec, THEN spec[of _ "map (Eval \<alpha>) bef"], THEN spec, THEN spec, 
        THEN spec[of _ "map (Eval \<alpha>) aft"], THEN mp, simplified length_map], rule conjI[OF _ cge], rule subsetI, rule inC, simp)
  from lstep obtain m where steps: "(Lab \<alpha> si,m) \<in> (rstep Decr)^*" and step: "(m,Lab \<alpha> ti) \<in> rstep (Lab_trs R)" by auto
  from rsteps_ctxt_imp_nrrsteps[OF steps, of ?E] rstep_ctxt_imp_nrrstep[OF step, of ?E] 
  have step: "(?E\<langle>Lab \<alpha> si\<rangle>, ?E\<langle>Lab \<alpha> ti\<rangle>) \<in> (nrrstep Decr)^* O nrrstep (Lab_trs R)" (is "(?l',?r) \<in> ?R")
    by auto
  have "ts = take i ts @ ti # drop (Suc i) ts" unfolding ti using id_take_nth_drop[OF i[unfolded len]] .
  also have "\<dots> = bef @ ti # aft" unfolding bef aft 
  proof (intro nth_equalityI, simp add: len, intro allI impI)
    fix j
    assume j: "j < length (take i ts @ ti # drop (Suc i) ts)" 
    show "(take i ts @ ti # drop (Suc i) ts) ! j = 
          (take i ss @ ti # drop (Suc i) ss) ! j"
    proof (cases "j = i")
      case True
      with i len have "j \<le> length ss" and "j \<le> length ts" by auto
      with True show ?thesis by (auto simp: nth_append_take )
    next
      case False
      from j i len have i: "i \<le> length ss" and j: "j \<le> length ss" and jj: "j < length ss" by auto
      from nth_append_take_drop_is_nth_conv[OF j i False]
           nth_append_take_drop_is_nth_conv[OF j[unfolded len] i[unfolded len] False]
           id[OF jj False] 
      show ?thesis by (simp)
    qed
  qed
  finally have ts: "ts = bef @ ti # aft" .
  have t: "Lab_root \<alpha> t = ?r" unfolding t ts by (simp add: o_def)
  from step have steps: "(?l',Lab_root \<alpha> t) \<in> ?R" unfolding t .
  let ?l = "?D\<langle>Lab \<alpha> si\<rangle>"
  have s: "?l = Lab_root \<alpha> s" unfolding s ss by (simp add: o_def)
  have steps2: "(?l,?l') \<in> Decr_root^="
  proof (cases "?ll = ?lr")
    case True
    thus ?thesis by simp
  next
    case False
    have ll: "?ll \<in> LS' f ?n" 
      by (simp, rule wf_L'[unfolded wf_label_def, THEN spec, THEN spec[of _ "map (Eval \<alpha>) (bef @ si # aft)"], THEN mp, simplified], auto simp: inC)
    have lr: "?lr \<in> LS' f ?n" 
      by (simp, rule wf_L'[unfolded wf_label_def, THEN spec, THEN spec[of _ "map (Eval \<alpha>) (bef @ ti # aft)"], THEN mp, simplified], auto simp: inC)
    from False lge have "(?ll,?lr) \<in> gr_root f ?n" unfolding lge_to_lgr_rel_def  lge_to_lgr_def Let_def using ll lr by auto
    hence "(?l,?l') \<in> Decr_root" unfolding decr_of_ord_def using ll lr by auto
    thus ?thesis by simp
  qed
  obtain mid where mid: "?l' = mid" by auto
  from steps2[unfolded s] steps
  show ?thesis unfolding mid by (force simp: O_assoc)
qed
end


context sl_interpr
begin

lemma quasi_rewrites_imp_lab_rewrites: 
  assumes steps: "(s,t) \<in> (rstep R)^*"
  and qmodel: "qmodel I L LC C cge R"
  shows "(LAB s, LAB t) \<in> (rstep (Lab_trs R) \<union> rstep Decr)^*"
using steps unfolding rstep_union[symmetric] unfolding rstep_CC_CS
proof (induct, simp)
  case (step y z)
  let ?R = "Lab_trs R"
  let ?D = "Decr"
  let ?RD = "?R \<union> ?D"
  from quasi_lab_rewrite[unfolded rstep_CC_CS, OF step(2) qmodel wf_default_ass, unfolded rstep_CC_CS[symmetric]]
  have "(LAB y, LAB z) \<in> (rstep ?D)^* O (rstep ?R)" by simp 
  hence "(LAB y, LAB z) \<in> (rstep ?RD)^* O (rstep ?RD)^*" using rtrancl_mono[OF rstep_mono, of ?D ?RD] rstep_mono[of ?R ?RD]
    by blast
  with step(3)
    show ?case unfolding rstep_CC_CS[symmetric] by force
qed
end

lemma nrrstep_decr_commute: assumes R: "(x,y) \<in> nrrstep R" and D: "(y,z) \<in> decr_of_ord gr LC LS"
  shows "(x,z) \<in> decr_of_ord gr LC LS O nrrstep R"
proof -
  let ?D = "decr_of_ord gr LC LS"
  let ?R = "nrrstep R"
  from D[unfolded decr_of_ord_def] obtain f l l' ts where
    y: "y = Fun (LC f (length ts) l) ts" and z: "z = Fun (LC f (length ts) l') ts" and l: "l \<in> LS f (length ts)" and l': "l' \<in> LS f (length ts)" and  gr: "(l, l') \<in> gr f (length ts)" by auto
  let ?f = "LC f (length ts) l"
  let ?g = "LC f (length ts) l'"
  from R[unfolded nrrstep_def'] obtain ll r C \<sigma> where lr: "(ll, r) \<in> R" and C: "C \<noteq> \<box>" and x: "x = C\<langle>ll \<cdot> \<sigma>\<rangle>" and y': "y = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from C y' y obtain D bef aft where C: "C = (More ?f bef D aft)" by (cases C, auto)
  let ?C = "\<lambda> l. bef @ D\<langle>l \<cdot> \<sigma>\<rangle> # aft"
  from y C y' have ts: "ts = ?C r" by auto  
  let ?x = "Fun ?f (?C ll)"
  let ?y = "Fun ?g (?C ll)"
  let ?z = "Fun ?g (?C r)"
  have x: "?x = x" unfolding x C by simp
  have z: "?z = z" unfolding z ts by simp
  have len: "Suc (length bef + length aft) = length ts" unfolding ts by simp
  have "(?x,?y) \<in> ?D" unfolding decr_of_ord_def
    by (rule, rule exI[of _ f], rule exI[of _ l], rule exI[of _ l'], rule exI[of _ "?C ll"], simp add: len l l' gr)
  moreover have "(?y,?z) \<in> ?R"
    by (unfold nrrstep_def', rule, clarify, rule exI[of _ ll], rule exI[of _ r], rule exI[of _ "More ?g bef D aft"], rule exI[of _ \<sigma>],
      simp add: lr)
  ultimately show ?thesis unfolding x z by auto
qed

lemma nrrsteps_decr_commute: assumes R: "(x,y) \<in> (nrrstep R)^*" and D: "(y,z) \<in> decr_of_ord gr LC LS"
  shows "(x,z) \<in> decr_of_ord gr LC LS O (nrrstep R)^*"
using R
proof (induct rule: converse_rtrancl_induct)
  case base from D show ?case by auto
next
  case (step x u)
  from step(3) obtain v where uv: "(u,v) \<in> decr_of_ord gr LC LS" and vz: "(v,z) \<in> (nrrstep R)^*" by auto
  from nrrstep_decr_commute[OF step(1) uv] vz  show ?case by auto
qed

lemma nrrsteps_decrs_commute: assumes R: "(x,y) \<in> (nrrstep R)^*" and D: "(y,z) \<in> (decr_of_ord gr LC LS)^*"
  shows "(x,z) \<in> (decr_of_ord gr LC LS)^* O (nrrstep R)^*"
using D
proof (induct rule: rtrancl_induct)
  case base from R show ?case by auto
next
  case (step z u) 
  let ?D = "decr_of_ord gr LC LS"
  from step(3) obtain v where xv: "(x,v) \<in> ?D^*" and vz: "(v,z) \<in> (nrrstep R)^*" by auto
  from xv nrrsteps_decr_commute[OF vz step(2)] have "(x,u) \<in> (?D^* O ?D) O (nrrstep R)^*" by auto 
  moreover have "?D^* O ?D \<subseteq> ?D^*" by auto
  ultimately show ?case by auto
qed
    

context sl_interpr_root
begin


lemma quasi_nrrewrites_imp_lab_rewrites: 
  assumes steps: "(s,t) \<in> (nrrstep R)^*"
  and qmodel: "qmodel I L LC C cge R"
  shows "(LAB_root s, LAB_root t) \<in> Decr_root^* O (nrrstep (Lab_trs R \<union> Decr))^*"
  (is "_ \<in> _ O ?rel^*")
using steps
proof (induct rule: converse_rtrancl_induct)
  case base show ?case by auto
next
  case (step y z)
  let ?R = "Lab_trs R"
  let ?D = "Decr"
  let ?RD = "?R \<union> ?D"
  let ?Dr = "Decr_root"
  from quasi_lab_nrrewrite[OF step(1) qmodel wf_default_ass] 
  have "(LAB_root y, LAB_root z) \<in> ?Dr^= O (nrrstep ?D)^* O (nrrstep ?R)" (is "?pair \<in> _") by simp 
  hence "?pair \<in> ?Dr^= O (nrrstep ?RD)^* O (nrrstep ?RD)^*" using rtrancl_mono[OF nrrstep_mono, of ?D ?RD] nrrstep_mono[of ?R ?RD]
    by blast
  hence "?pair \<in> ?Dr^= O (nrrstep ?RD)^*" by auto
  with step(3) obtain m1 m2 where ym1: "(LAB_root y,m1) \<in> ?Dr^=" and m1z: "(m1,LAB_root z) \<in> (nrrstep ?RD)^*" and zm2: "(LAB_root z,m2) \<in> ?Dr^*" and m2t: "(m2,LAB_root t) \<in> (nrrstep ?RD)^*" by auto
  from ym1 nrrsteps_decrs_commute[OF m1z zm2] have ym2: "(LAB_root y, m2) \<in> (?Dr^= O ?Dr^*) O (nrrstep ?RD)^*" by auto
  have "?Dr^= O ?Dr^* \<subseteq> ?Dr^*" by auto
  with ym2 have ym2: "(LAB_root y, m2) \<in> ?Dr^* O (nrrstep ?RD)^*" by auto
  from ym2 m2t show ?case by auto
qed
end

context sl_interpr
begin

lemma quasi_rel_rstep_imp_lab_rel_rstep: 
  assumes step: "(s,t) \<in> rel_rstep (R,S)"
  and qmodel: "qmodel I L LC C cge (R \<union> S)"
  shows "(LAB s, LAB t) \<in> rel_rstep (Lab_trs R, Lab_trs S \<union> Decr)"
proof -
  from qmodel have R: "qmodel I L LC C cge R" and S: "qmodel I L LC C cge S" unfolding qmodel_def by auto
  from step obtain u v where su: "(s,u) \<in> (rstep S)^*"
    and uv: "(u,v) \<in> rstep R" and vt: "(v,t) \<in> (rstep S)^*" unfolding rel_rstep.simps rel_step.simps by auto
  let ?R = "Lab_trs R"
  let ?S = "Lab_trs S \<union> Decr"
  from quasi_rewrites_imp_lab_rewrites[OF su S]
       quasi_lab_rewrite[OF uv R wf_default_ass]
       quasi_rewrites_imp_lab_rewrites[OF vt S]
  have "(LAB s, LAB u) \<in> (rstep ?S)^*"
   and "(LAB u, LAB v) \<in> (rstep ?S)^* O rstep ?R"
   and "(LAB v, LAB t) \<in> (rstep ?S)^*"
    unfolding rstep_union[symmetric] using rtrancl_mono[OF rstep_mono, of "Decr" ?S] by blast+
  hence "(LAB s, LAB t) \<in> ((rstep ?S)^* O (rstep ?S)^*) O rstep ?R O (rstep ?S)^*" by blast
  thus ?thesis unfolding rel_step.simps rel_rstep.simps rtrancl_idemp_self_comp  .
qed

  
lemma quasi_SN_lab_imp_SN_term: 
  assumes SN: "SN_elt (rstep LR) (LAB t)" 
  and qmodel: "qmodel I L LC C cge R"
  and decr: "Decr \<subseteq> (rstep LR)^+"
  and lab: "Lab_trs R \<subseteq> LR"
  shows "SN_elt (rstep R) t"
proof (rule ccontr)
  assume "\<not> SN_elt (rstep R) t"
  from this obtain f where inf: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" unfolding SN_elt_def  by auto
  obtain g where g: "\<And> i. g i = LAB (f i)" by auto
  let ?R = "Lab_trs R"
  let ?D = "Decr"
  let ?RD = "LR"
  from rstep_mono[OF decr] have "rstep ?D \<subseteq> (rstep LR)^+" using rstep_trancl_distrib[of "rstep LR", simplified rstep_rstep] by auto
  hence subset: "rstep ?D \<subseteq> (rstep ?RD)^+" unfolding rstep_union using trancl_mono by blast
  have subset: "(rstep ?D)^* O  rstep ?R \<subseteq> (rstep (?RD))^+" 
    using rtrancl_mono[OF subset, simplified trancl_rtrancl_absorb] rstep_mono[OF lab] unfolding trancl_unfold_right by blast
  have one: "g 0 = LAB t" using inf g by auto
  have "\<forall> i. (g i, g (Suc i)) \<in> (rstep ?D)^* O  rstep ?R" 
    by (rule allI, simp only: g, rule quasi_lab_rewrite, auto simp: inf wf_default_ass qmodel g)
  hence two: "\<forall> i. (g i, g (Suc i)) \<in> (rstep (?RD))^+"
    using subset by auto
  from SN_elt_imp_SN_elt_trancl[OF SN] one two
  show False unfolding SN_elt_def by auto
qed    

lemma quasi_rel_SN_lab_imp_rel_SN: 
  assumes SN: "rel_SN (rstep L_R, rstep L_S)" 
  and qmodel: "qmodel I L LC C cge (R \<union> S)"
  and decr: "Decr \<subseteq> (rstep L_S)^+"
  and labS: "Lab_trs S \<subseteq> L_S"
  and labR: "Lab_trs R \<subseteq> L_R"
  shows "rel_SN (rstep R,rstep S)"
unfolding rel_SN_def
proof
  fix f
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> rel_step (rstep R, rstep S)"
  obtain g where g: "\<And> i. g i = LAB (f i)" by auto
  let ?R = "Lab_trs R"
  let ?D = "Decr"
  let ?SD = "L_S"
  from rstep_mono[OF decr] have "rstep ?D \<subseteq> (rstep L_S)^+" using rstep_trancl_distrib[of "rstep L_S", simplified rstep_rstep] by auto
  hence "rstep ?D \<subseteq> (rstep L_S)^*" by auto
  with rstep_mono[OF labS] have subset: "rstep (Lab_trs S \<union> ?D) \<subseteq> (rstep ?SD)^*" unfolding rstep_union using rtrancl_mono by blast
  from quasi_rel_rstep_imp_lab_rel_rstep[OF _ qmodel] steps have
    steps: "\<And> i. (g i, g (Suc i)) \<in> rel_rstep (?R, Lab_trs S \<union> Decr)" unfolding g by auto
  with rel_step_mono[OF rstep_mono[OF labR] subset] have
    steps: "\<And> i. (g i, g (Suc i)) \<in> rel_rstep (L_R, ?SD)" unfolding rel_rstep.simps rel_step_rtrancl by auto
  with SN show False unfolding rel_SN_def rel_rstep.simps SN_defs by force
qed

lemma gr_irreflexive: "(l,l) \<notin> gr f n"
  using SN_imp_irreflexive[OF gr_SN[of f n]] .
end



subsection {* integrate these results in SL-Framework *}

fun
  label :: "('f, 'l) lab \<Rightarrow> nat \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> ('f, 'l) lab"
where
  "label f n (Inl l) = Lab f l" |
  "label f n (Inr l) = FunLab f l"

fun label_decomp :: "('f,'l)lab \<Rightarrow> ('f,'l)lab \<times> ('l + ('f,'l) lab list)"
where "label_decomp (Lab f l) = (f, Inl l)"
    | "label_decomp (FunLab f l) = (f, Inr l)"

lemma label_decomp_label: "label_decomp (label f n l) = (f,l)" by (cases l, auto)

lemma [simp]: "unlab (label f n l) = f" by (cases l, auto)

lemma unlab_term: "unlab_term (lab I L label \<alpha> t) = t"
proof (induct t, simp)
  case (Fun f ts)
  thus ?case by (auto simp: Let_def, induct ts, auto)
qed

lemma unlab_term_root: "unlab_term (lab_root I L L' label \<alpha> t) = t"
  by (cases t, simp, auto simp: unlab_term o_def)

lemma [simp]: "remove_lab (label f n l) = remove_lab f" by (cases l, auto)

lemma decr_of_lab_is_empty: "decr_of_trs (lab_trs I L label C R) = {}"
unfolding decr_of_trs_def by (auto simp: lab_rule_def unlab_term)


lemma unlab_rule_I: 
   "unlab_trs (lab_rule I L label C rule) \<subseteq> {rule}" (is "?rule \<subseteq> {rule}")
proof
  fix ulrule
  assume ulrule: "ulrule \<in> ?rule"
  show "ulrule \<in> {rule}"
  proof (cases ulrule)
    case (Pair ull ulr)
    with ulrule have "(ull,ulr) \<in> ?rule" by auto
    from this obtain ll lr where lrule: "ull = unlab_term ll \<and> ulr = unlab_term lr \<and> (ll,lr) \<in> lab_rule I L label C rule" 
      by (force simp: map_trs.simps)
    from this obtain l r \<alpha> where "rule = (l,r) \<and> ll = lab I L label \<alpha> l \<and> lr = lab I L label \<alpha> r" unfolding lab_rule_def by (cases rule, auto)
    with Pair ulrule lrule show ?thesis by (simp add: unlab_term)
  qed
qed

definition
  R_subset_lab_decr :: "(('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('l + ('f, 'l) lab list) ars) \<Rightarrow> (('f,'l)lab,('l + ('f, 'l) lab list))labels \<Rightarrow> (('f, 'l) lab, 'v) trs \<Rightarrow> (('f, 'l) lab, 'v) trs \<Rightarrow> bool"
where
  "R_subset_lab_decr gr LS R rR \<equiv> \<forall> (l,r) \<in> rR. (unlab_term l, unlab_term r) \<in> R \<and> (l = r \<or> unlab_term l \<noteq> unlab_term r) \<or> (l,r) \<in> decr_of_ord gr label LS"

lemma label_inject: "(label f n l = label f n l') = (l = l')"
  by (cases l, cases l', auto, cases l', auto)


locale sl_interpr_lab = sl_interpr C c I cge lge L label label_decomp LS
  for C :: "'c set"
  and  c :: "'c"
  and  I :: "(('f,'l)lab,'c)inter"
  and  cge :: "'c \<Rightarrow> 'c \<Rightarrow> bool"
  and  lge :: "('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> bool"
  and  L :: "(('f,'l)lab,'c,('l + ('f, 'l) lab list))label"
  and  LS :: "(('f,'l)lab,('l + ('f, 'l) lab list))labels" 
begin

abbreviation R_subset_lab_Decr :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs \<Rightarrow> bool" where "R_subset_lab_Decr \<equiv> R_subset_lab_decr gr LS"

lemma unlab_rule_II: 
   "{rule} \<subseteq> unlab_trs (lab_rule I L label C rule)" (is "?rule \<supseteq> {rule}")
proof (cases rule)
  case (Pair l r)
  with wf_default_ass
  have "(LAB l, LAB r) \<in> lab_rule I L label C rule" (is "(?l,?r) \<in> _") 
    unfolding lab_rule_def using Pair by auto
  hence "(unlab_term ?l, unlab_term ?r) \<in> ?rule"  by (force simp: map_trs.simps)
  hence "rule \<in> ?rule" using Pair by (simp add: unlab_term)
  thus ?thesis by auto
qed

lemma unlab_rule[simp]: 
   "unlab_trs (lab_rule I L label C rule) = {rule}"
using unlab_rule_I unlab_rule_II by blast


lemma decr_of_decr_is_decr: assumes gr_irreflexive: "\<And> f n l. (l,l) \<notin> grg f n " 
  shows "decr_of_trs (decr_of_ord grg label LSg) = (decr_of_ord grg label LSg :: (('f,'l)lab,'v)trs)"
proof -
  {
    fix f l l' ts
    assume gr: "(l,l') \<in> grg f (length ts)" and id: "label f (length ts) l = label f (length ts) l'" (is "?l = ?r")
    from id have "label_decomp ?l = label_decomp ?r" by simp
    hence "l = l'" unfolding LD_LC by simp
    with gr gr_irreflexive have False by simp
  }
  thus ?thesis unfolding decr_of_trs_def decr_of_ord_def by auto
qed

lemma unlab_trs: "unlab_trs (Lab_trs R) = R"
proof (rule set_eqI)
  fix rule
  let ?ULR = "unlab_trs (Lab_trs R)"
  show "(rule \<in> ?ULR) = (rule \<in> R)"
  proof
    assume rule: "rule \<in> R"
    from unlab_rule[where ?rule = rule] have "rule \<in> unlab_trs (lab_rule I L label C rule)" by auto  
    with map_trs_mono rule show "rule \<in> ?ULR" by (auto simp: map_trs.simps)
  next
    assume rule: "rule \<in> ?ULR"
    from this obtain rrule where "rrule \<in> R \<and> rule \<in> unlab_trs (lab_rule I L label C rrule)" by (auto simp: map_trs.simps)
    thus "rule \<in> R" using rule unlab_rule[where ?rule = rrule] by auto
  qed
qed

lemma SN_imp_SN_lab_term: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trs"
  assumes SN: "SN_elt (rstep R) t"
  shows "SN_elt (rstep (Lab_trs R)) (LAB t)"
proof -
  from SN have "SN_elt (rstep (unlab_trs (Lab_trs R))) (unlab_term (LAB t))" 
    by (simp only: unlab_trs unlab_term)
  thus ?thesis by (rule SN_map_imp_SN)
qed    


lemma SN_imp_SN_lab_term_2: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trs"
  assumes SN: "SN_elt (rstep R) t"
  and subset: "unlab_trs LR \<subseteq> R"
  shows "SN_elt (rstep LR) (LAB t)"
proof -
  from SN have "SN_elt (rstep R) (unlab_term (LAB t))" 
    by (simp only: unlab_trs unlab_term)  
  show ?thesis 
    by (rule SN_map_imp_SN[where fg = unlab], simp only: unlab_term, rule SN_elt_subset, rule SN, rule rstep_mono, rule subset)
qed

lemma decr_irreflexive: assumes gr_irreflexive: "\<And> f n l. (l,l) \<notin> grg f n " 
  shows "(l,l) \<notin> (decr_of_ord grg label LSg :: (('f,'l)lab,'v)trs)"
proof
  assume "(l,l) \<in> decr_of_ord grg label LSg"
  then obtain f l' l'' ts 
    where l: "l = Fun (label f (length ts) l') ts" and l': "l = Fun (label f (length ts) l'') ts"
    and gr: "(l',l'') \<in> grg f (length ts)" unfolding decr_of_ord_def by blast
  hence "label f (length ts) l' = label f (length ts) l''" unfolding l l' by auto
  hence "l' = l''" unfolding label_inject  .
  with gr gr_irreflexive show False by simp
qed
  
  
lemma lab_trs_decr: assumes gr_irreflexive: "\<And> f n l. (l,l) \<notin> grg f n " 
  shows "Lab_trs R \<inter> (decr_of_ord grg label LSg) = {}"
proof -
  {
    fix ll rr
    assume llrr: "(ll,rr) \<in> decr_of_ord grg label LSg" and R: "(ll,rr) \<in> Lab_trs R"
    then obtain f l' l'' ts 
      where ll: "ll = Fun (label f (length ts) l') ts" and rr: "rr = Fun (label f (length ts) l'') ts"
      and gr: "(l',l'') \<in> grg f (length ts)" unfolding decr_of_ord_def by auto
    from R obtain l r \<alpha> where llR: "ll = lab I L label \<alpha> l" and rrR: "rr = lab I L label \<alpha> r" 
      by (force simp: lab_rule_def)
    from ll have "unlab_term ll = unlab_term (Fun (label f (length ts) l') ts)" by simp
    hence l: "l = unlab_term (Fun (label f (length ts) l') ts)" unfolding llR unlab_term .
    from rr have "unlab_term rr = unlab_term (Fun (label f (length ts) l'') ts)" by simp
    hence r: "r = unlab_term (Fun (label f (length ts) l'') ts)" unfolding rrR unlab_term .
    have "l = r" unfolding l r by auto
    hence "ll = rr" unfolding llR rrR by auto
    with llrr have False using decr_irreflexive[OF gr_irreflexive] by force
  }
  thus ?thesis by auto
qed
  

lemma lab_sound_tt: 
  shows "sound_lab_tt (\<lambda> R m rR rm. rm = Suc m \<and> Lab_trs R \<subseteq> rR \<and> Decr \<subseteq> (rstep rR)^+ \<and> R_subset_lab_Decr R rR \<and> qmodel I L label C cge R)"
unfolding sound_lab_tt_def 
proof (intro allI impI, (erule conjE)+)
  fix R :: "(('f,'l)lab,'v)trs" and m and rR :: "(('f,'l)lab,'v)trs" and rm
  assume m: "rm = Suc m" and subset: "Lab_trs R \<subseteq> rR" and decrSN: "decrSN rR rm" and subset2: "R_subset_lab_Decr R rR" 
    and decrR: "Decr \<subseteq> (rstep rR)^+" and qmodel: "qmodel I L label C cge R"
  from decrSN show "decrSN R m" 
    unfolding decrSN_def
  proof (intro impI)
    assume SN: "\<forall> n<m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))" (is "\<forall> n<m. ?invR n")
    assume imp: "(\<forall>n<rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))) \<longrightarrow> SN (rstep rR)" (is "(\<forall> n<rm. ?invLR n) \<longrightarrow> ?SNL")
    have "\<forall> n < rm. ?invLR n"
    proof (intro allI impI)
      fix n
      assume le: "n < rm"
      from subset2 have decr_subset: "decr_of_trs rR \<subseteq> Decr" unfolding R_subset_lab_decr_def decr_of_trs_def
	by auto
      show "?invLR n"
      proof (cases n)
	case (Suc mm)
	have "unlab_of_trs rR = R" unfolding unlab_of_trs_def 
	proof (rule set_eqI, simp only: split_paired_all)
	  fix l r 
	  show "((l,r) \<in> unlab_trs (rR - decr_of_trs rR)) = ((l,r) \<in> R)"
	  proof
	    assume "(l,r) \<in> R"
	    hence lab: "(LAB l, LAB r) \<in> Lab_trs R" by (force simp: lab_rule_def wf_default_ass)
	    with subset have one: "(LAB l, LAB r) \<in> rR" by auto
            from lab have "(LAB l, LAB r) \<notin> Decr" using lab_trs_decr[of gr, OF gr_irreflexive] by blast
	    with decr_subset one have "(LAB l, LAB r) \<in> rR - decr_of_trs rR" by force
	    hence "(unlab_term (LAB l), unlab_term (LAB r)) \<in> unlab_trs (rR - decr_of_trs rR)" by (force simp: map_trs.simps)
	    thus "(l,r) \<in> unlab_trs (rR - decr_of_trs rR)" by (simp only: unlab_term)
	  next
	    assume "(l,r) \<in> unlab_trs (rR - decr_of_trs rR)"
	    from this obtain ll lr where mem: "(ll,lr) \<in> rR - decr_of_trs rR" and id: "unlab_term ll = l \<and> unlab_term lr = r" by (force simp: map_trs.simps)
	    with subset2 have disj: "(l, r) \<in> R \<and> (ll = lr \<or> l \<noteq> r) \<or> (ll, lr) \<in> Decr" unfolding R_subset_lab_decr_def by auto
	    {
	      assume "\<not> ((l,r) \<in> R \<and> (ll = lr \<or> l \<noteq> r)) \<and> (ll,lr) \<in> Decr"
	      hence inDecr: "(ll,lr) \<in> Decr" and disj2: "(l,r) \<notin> R \<or> ll \<noteq> lr \<and> l = r" by blast+
	      from inDecr have "unlab_term ll = unlab_term lr" unfolding decr_of_ord_def by auto
	      with mem inDecr have "(ll,ll) \<in> Decr" unfolding decr_of_trs_def by force
	      with decr_irreflexive[of gr, OF gr_irreflexive] have False by auto
	    }
	    with disj show "(l,r) \<in> R" by blast
	  qed
	qed
	hence id: "(unlab_of_trs ^^ (Suc mm)) rR = (unlab_of_trs ^^ mm) R" by (simp add: funpow_swap1)
	have "?invLR (Suc mm)" using le m SN Suc by (simp only: id)
	thus ?thesis using Suc by auto
      next
	case 0
	show ?thesis by (simp add: 0, rule SN_subset, rule Decr_SN, rule rstep_mono, rule decr_subset)
      qed
    qed
    with imp have ?SNL by auto
    show "SN (rstep R)"
      using quasi_SN_lab_imp_SN_term[OF _ qmodel decrR subset] `?SNL`
      unfolding SN_defs by auto 
  qed
qed


definition sem_lab_proc :: "
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  bool"
  where "sem_lab_proc P R m lP lR rm \<equiv> rm = Suc m \<and> Lab_trs P \<subseteq> lP \<and> Lab_trs R \<subseteq> lR \<and> Decr \<subseteq> (rstep lR)^+ \<and> R_subset_lab_Decr R lR \<and> qmodel I L label C cge R"

abbreviation lab_terms where "lab_terms s \<sigma> i \<equiv> Lab (subst_ass default_ass (\<sigma> i)) (s i)"
abbreviation lab_substs  where "lab_substs  \<sigma> i \<equiv> lab_subst default_ass (\<sigma> i)"

lemma lab_minimal_termI:
  assumes  gr_SN: "\<And> f n. SN (lge_to_lgr_rel lge LSg f n)" 
  and subset: "R_subset_lab_decr (lge_to_lgr_rel lge LSg) LSg R lR"
  and minimal: "lab_minimal_term R m (unlab_term t)"
  shows "lab_minimal_term lR (Suc m) t"
proof (unfold lab_minimal_term_def, intro allI impI)
  fix n
  let ?gr = "lge_to_lgr_rel lge LSg"
  let ?Decr = "decr_of_ord ?gr label LSg"
  assume le: "n \<le> Suc m"
  from SN_imp_irreflexive[OF gr_SN] have gr_irreflexive: "\<And> f n l. (l,l) \<notin> ?gr f n"  .
  from subset have lR: "unlab_of_trs lR \<subseteq> R" unfolding R_subset_lab_decr_def unlab_of_trs_def decr_of_ord_def decr_of_trs_def 
    by (auto simp: gr_irreflexive label_inject map_trs.simps)
  show "SN_elt (rstep ((unlab_of_trs ^^ n) lR)) ((unlab_term ^^ n) t)" (is "?goal n")
  proof (cases n)
    case (Suc mm)
    with le have "mm \<le> m" by auto
    with minimal have cond: "SN_elt (rstep ((unlab_of_trs ^^ mm) R)) ((unlab_term ^^ mm) (unlab_term t))" by (simp only: lab_minimal_defs)
    have cond2: "SN_elt (rstep ((unlab_of_trs ^^ mm) (unlab_of_trs lR))) ((unlab_term ^^ mm) (unlab_term t))" 
      by (rule SN_elt_subset[OF cond rstep_mono[OF funpow_mono[OF unlab_of_trs_mono]]], auto simp: lR)
    have "?goal (Suc mm)" 
      by (simp add: funpow_swap1, simp only: unlab_term cond2)	  
    with Suc show ?thesis by auto
  next
    case 0
    from subset have decr_subset: "decr_of_trs lR \<subseteq> ?Decr" unfolding R_subset_lab_decr_def decr_of_trs_def
      by auto
    from minimal have "SN_elt (rstep ((unlab_of_trs ^^ 0) R)) ((unlab_term ^^ 0) (unlab_term t))" by (simp only: lab_minimal_defs)
    hence SN: "SN_elt (rstep R) (unlab_term t)" by simp
    have "SN_elt (rstep lR) t"  
      by (rule SN_unlab, rule SN_subset, rule Decr_SN_generic[OF gr_SN], rule rstep_mono, rule decr_subset,
	rule SN_elt_subset, rule SN, rule rstep_mono, rule lR)	  
    with 0 show ?thesis by auto
  qed
qed

lemma lab_minimal_termI_2:
  assumes subset: "R_subset_lab_Decr R lR"
  and minimal: "lab_minimal_term R m t"
  shows "lab_minimal_term lR (Suc m) (Lab default_ass t)"
  by (rule lab_minimal_termI[OF gr_SN subset], unfold unlab_term, rule minimal)

  
lemma sem_lab_proc_sound: 
  shows "sound_lab_proc sem_lab_proc"
unfolding sound_lab_proc_def
proof (intro allI impI, unfold sem_lab_proc_def, (erule conjE)+ )
  fix P R :: "(('f,'l)lab,'v)trs" and m and lP lR :: "(('f,'l)lab,'v)trs" and rm
  assume finite: "finite_lab_dpp (lP, lR) rm" and lP: "Lab_trs P \<subseteq> lP" and lR1: "Lab_trs R \<subseteq> lR" and subset2: "R_subset_lab_Decr R lR" 
      and decr: "Decr \<subseteq> (rstep lR)^+" and qmodel: "qmodel I L label C cge R" and m: "rm = Suc m"
  let ?lr = "Lab_trs R"
  let ?D = "Decr"
  let ?lR = "Lab_trs R \<union> Decr"
  from rstep_mono[OF decr] have "rstep ?D \<subseteq> (rstep lR)^+" using rstep_trancl_distrib[of "rstep lR", simplified rstep_rstep] by auto
  hence "rstep ?D \<subseteq> (rstep lR)^+" unfolding rstep_union using trancl_mono by blast
  hence one: "rstep ?D \<subseteq> (rstep lR)^*" by auto
  from rstep_mono[OF lR1] have "rstep ?lr \<subseteq> (rstep (lR))^*" by auto
  with one have "rstep ?lR \<subseteq> (rstep lR)^*" unfolding rstep_union by auto
  from rtrancl_mono[OF this] have rstepSub: "(rstep ?lR)^* \<subseteq> (rstep lR)^*" unfolding rtrancl_idemp .
  from subset2 have lR2: "unlab_of_trs lR \<subseteq> R" unfolding R_subset_lab_decr_def unlab_of_trs_def decr_of_ord_def decr_of_trs_def 
    by (auto simp: gr_irreflexive label_inject map_trs.simps)
  let ?lP = "Lab_trs P"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    let ?ls = "lab_terms s \<sigma>"
    let ?lt = "lab_terms t \<sigma>"
    let ?lsig = "lab_substs \<sigma>"
    let ?lss = "\<lambda> i. Lab default_ass (s i \<cdot> \<sigma> i)"
    let ?lts = "\<lambda> i. Lab default_ass (t i \<cdot> \<sigma> i)"
    have s: "\<forall> i. ?lss i = ?ls i \<cdot> ?lsig i" by simp
    have t: "\<forall> i. ?lts i = ?lt i \<cdot> ?lsig i" by simp
    from chain have "\<forall> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    hence "\<forall> i. (?lts i, ?lss (Suc i)) \<in> (rstep ?lR)^*" using quasi_rewrites_imp_lab_rewrites[OF _ qmodel] unfolding rstep_union by best
    with s t have "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*" by auto
    with rstepSub have steps: "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep lR)^*" by auto    
    from chain have mem: "\<forall> i. (s i, t i) \<in> P" by simp
    have "\<forall> i. (?ls i, ?lt i) \<in> ?lP"
    proof
      fix i
      have memI: "(s i, t i) \<in> P" using mem by simp
      have "wf_ass (subst_ass default_ass (\<sigma> i))" 
	by (rule wf_ass_subst_ass, rule wf_default_ass)
      hence "(?ls i, ?lt i) \<in> lab_rule I L label C (s i, t i)"
	unfolding lab_rule_def by (auto)
      with memI show "(?ls i, ?lt i) \<in> ?lP" by auto
    qed
    hence memLab: "\<forall> i. (?ls i, ?lt i) \<in> lP" using lP by auto    
    have minimal: "lab_minimal_cond lR rm ?ls ?lt ?lsig"
      unfolding lab_minimal_cond_def
    proof (intro allI)
      fix i
      from chain have "lab_minimal_term R m (t i \<cdot> \<sigma> i)" by (auto simp: lab_minimal_cond_def)
      from lab_minimal_termI_2[OF subset2 this] m 
      show "lab_minimal_term lR rm (?lt i \<cdot> ?lsig i)" by simp
    qed
    from steps minimal memLab have "min_lab_ichain (lP,lR) rm ?ls ?lt ?lsig" by auto
    with finite show False unfolding finite_lab_dpp_def by blast
  qed
qed

end

      
 

locale sl_interpr_root_lab = sl_interpr_root C c I cge lge L label label_decomp LS L' LS' + sl_interpr_lab C c I cge lge L LS 
  for C :: "'c set"
  and  c :: "'c"
  and  I :: "(('f,'l)lab,'c)inter"
  and  cge :: "'c \<Rightarrow> 'c \<Rightarrow> bool"
  and  lge :: "('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> bool"
  and  L :: "(('f,'l)lab,'c,('l + ('f, 'l) lab list))label"
  and  LS :: "(('f,'l)lab,('l + ('f, 'l) lab list))labels" 
  and  L' :: "(('f,'l)lab,'c,('l + ('f, 'l) lab list))label"
  and  LS' :: "(('f,'l)lab,('l + ('f, 'l) lab list))labels" 
begin

lemma decr_root_unlab_trs: "unlab_trs (Decr_root) \<subseteq> Id"
  unfolding map_trs.simps decr_of_ord_def
  by auto


lemma quasi_min_nrrewrites_imp_lab_rewrites: 
  assumes steps: "(s,t) \<in> (nrrstep R)^*"
  and min: "lab_minimal_term R m s"
  and qmodel: "qmodel I L label C cge R"
  shows "(LAB_root s, LAB_root t) \<in> (rstep (Lab_trs R \<union> Decr) \<union> (Decr_root \<inter> {(s,t) | s t. lab_minimal_term R m (unlab_term s)}))^*"
  (is "_ \<in> ?rel^*")
using steps 
proof (induct)
  case base thus ?case by simp 
next
  case (step y z)  
  let ?R = "Lab_trs R"
  let ?D = "Decr"
  let ?RD = "?R \<union> ?D"
  let ?Dr = "Decr_root"
  from step(1) have "(s,y) \<in> (rstep R)^*" by (rule nrrsteps_imp_rsteps)
  from lab_minimal_term_rsteps[OF min this] have miny: "lab_minimal_term R m y" .
  from quasi_lab_nrrewrite[OF step(2) qmodel wf_default_ass]
  have "(LAB_root y, LAB_root z) \<in> ?Dr^= O (nrrstep ?D)^* O (nrrstep ?R)" (is "?pair \<in> _") by simp 
  hence "?pair \<in> ?Dr^* O (nrrstep ?RD)^* O (nrrstep ?RD)^*" using rtrancl_mono[OF nrrstep_mono, of ?D ?RD] nrrstep_mono[of ?R ?RD]
    by blast
  hence "?pair \<in> ?Dr^* O (nrrstep ?RD)^*" by auto
  hence "?pair \<in> ?Dr^* O ?rel^*"  using rtrancl_mono[of "nrrstep ?RD" ?rel] unfolding rstep_iff_rrstep_or_nrrstep by blast
  then obtain t where first: "(LAB_root y, t) \<in> ?Dr^*" and second: "(t,LAB_root  z) \<in> ?rel^*" by blast+
  from first have "(LAB_root y, t) \<in> ?rel^* \<and> lab_minimal_term R m (unlab_term t)"
  proof (induct)
    case base show ?case using miny by ( simp add: unlab_term_root)
  next
    case (step t tt)
    from step(3) have min: "lab_minimal_term R m (unlab_term t)" ..
    from step(2) have id: "unlab_term t = unlab_term tt" unfolding decr_of_ord_def by auto
    from min min[unfolded id]
      step(2)  have "(t,tt) \<in> ?rel" by auto 
    with step(3) have "(LAB_root y, tt) \<in> ?rel^+" using rtrancl_into_trancl1[where r = ?rel] by auto
    hence "(LAB_root y, tt) \<in> ?rel^*" by auto
    with min[unfolded id] show ?case by auto
  qed
  hence "?pair \<in> ?rel^*" using second by auto
  with step(3) show ?case by auto
qed


definition sem_lab_root_proc :: "
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  bool"
  where "sem_lab_root_proc P R m lP lR rm \<equiv> rm = Suc m \<and> (\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> \<not> defined R (the (root t), num_args t)) 
     \<and> Lab_root_trs P \<subseteq> lP \<and> Lab_trs R \<subseteq> lR 
     \<and> Decr_root  \<subseteq> (CS (lP \<union> lR) \<inter> Decr_root)^+ 
     \<and> Decr \<subseteq> (rstep lR)^+ 
     \<and> R_subset_lab_decr gr_both LS_both R lR 
     \<and> qmodel I L label C cge R"

abbreviation lab_root_terms where "lab_root_terms s \<sigma> i \<equiv> Lab_root (subst_ass default_ass (\<sigma> i)) (s i)"

lemma sem_lab_root_proc_sound: 
  shows "sound_lab_proc sem_lab_root_proc"
unfolding sound_lab_proc_def
proof (intro allI impI, unfold sem_lab_root_proc_def, (erule conjE)+ )
  fix P R :: "(('f,'l)lab,'v)trs" and m and lP lR :: "(('f,'l)lab,'v)trs" and rm
  let ?M = "{(s,t) | s t. lab_minimal_term lR (Suc m) t} :: (('f,'l)lab,'v)trs"
  let ?Ml = "{(s,t) | s t. lab_minimal_term R  m (unlab_term s)} :: (('f,'l)lab,'v)trs"
  assume finite: "finite_lab_dpp (lP, lR) rm" and lP: "Lab_root_trs P \<subseteq> lP" and lR1: "Lab_trs R \<subseteq> lR" and subset2: "R_subset_lab_decr gr_both LS_both R lR" 
      and decr: "Decr \<subseteq> (rstep lR)^+" and decr_root: "Decr_root \<subseteq> (CS (lP \<union> lR) \<inter> Decr_root)^+" and qmodel: "qmodel I L label C cge R" and m: "rm = Suc m"
      and nonvar_def: "\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> \<not> defined R (the (root t), num_args t)"
  hence nonvar_s: "\<And> s t. (s,t) \<in> P \<Longrightarrow> \<not> is_var s"
    and nonvar_t: "\<And> s t. (s,t) \<in> P \<Longrightarrow> \<not> is_var t" by auto
  note min_intro = lab_minimal_termI[OF gr'_SN subset2]
  have decr_root: "Decr_root \<inter> ?Ml \<subseteq> (CS (lP \<union> lR) \<inter> ?M)^+" (is "_ \<subseteq> ?rel^+")
  proof 
    fix s t
    assume "(s,t) \<in> Decr_root \<inter> ?Ml"    
    hence step: "(s,t) \<in> Decr_root" and min: "lab_minimal_term R m (unlab_term s)" 
      by auto
    from decr_root step have "(s,t) \<in> (CS (lP \<union> lR) \<inter> Decr_root)^+" by auto
    from this min have "(s,t) \<in> ?rel^+ \<and> lab_minimal_term R m (unlab_term t)"
    proof (induct)
      case (base y)
      hence "lab_minimal_term R m (unlab_term y)" 
        unfolding decr_of_ord_def lge_to_lgr_rel_def lge_to_lgr_def Let_def by auto
      with min_intro[OF this] base(1) show ?case by auto
    next
      case (step y z)
      from step(3)[OF step(4)] have "lab_minimal_term R m (unlab_term y)" by simp
      with step(2) have min: "lab_minimal_term R m (unlab_term z)" 
        unfolding decr_of_ord_def lge_to_lgr_rel_def lge_to_lgr_def Let_def by auto
      from min_intro[OF this] step(2) have second: "(y,z) \<in> ?rel" by simp
      from step(3)[OF step(4)] have "(s,y) \<in> ?rel^+" by auto
      from this second have "(s,z) \<in> ?rel^+" ..
      with min show ?case by auto
    qed
    thus "(s,t) \<in> (CS (lP \<union> lR) \<inter> ?M)^+" by auto
  qed
  let ?lr = "Lab_trs R"
  let ?D = "Decr"
  let ?lR = "Lab_trs R \<union> Decr"
  let ?Dr = "Decr_root"
  let ?Drr = "?Dr \<inter> ?M"
  let ?Drl = "?Dr \<inter> ?Ml"
  from rstep_mono[OF decr] have "rstep ?D \<subseteq> (rstep lR)^+" using rstep_trancl_distrib[of "rstep lR", simplified rstep_rstep] by auto
  hence "rstep ?D \<subseteq> (rstep lR)^+" unfolding rstep_union using trancl_mono by blast
  hence one: "rstep ?D \<subseteq> (rstep lR)^*" by auto
  from rstep_mono[OF lR1] have "rstep ?lr \<subseteq> (rstep (lR))^*" by auto
  with one have "rstep ?lR \<subseteq> (rstep lR)^*" unfolding rstep_union by auto
  hence rstepSub: "rstep ?lR \<union> ?Drl \<subseteq> (rstep lR)^* \<union> ?Drl" by auto
  from subset2 have lR2: "unlab_of_trs lR \<subseteq> R" unfolding R_subset_lab_decr_def unlab_of_trs_def decr_of_ord_def decr_of_trs_def 
    by (auto simp: SN_imp_irreflexive[OF gr'_SN] label_inject map_trs.simps)
  let ?lP = "Lab_root_trs P"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    let ?ls = "lab_root_terms s \<sigma>"
    let ?lt = "lab_root_terms t \<sigma>"
    let ?lsig = "lab_substs \<sigma>"
    let ?lss = "\<lambda> i. Lab_root default_ass (s i \<cdot> \<sigma> i)"
    let ?lts = "\<lambda> i. Lab_root default_ass (t i \<cdot> \<sigma> i)"
    from chain have mem: "\<And> i. (s i, t i) \<in> P" by simp
    have s: "\<And> i. ?lss i = ?ls i \<cdot> ?lsig i" using nonvar_def mem lab_root_subst by blast
    have t: "\<And> i. ?lts i = ?lt i \<cdot> ?lsig i" using nonvar_def mem lab_root_subst by blast
    from chain have "\<And> i. SN_elt (rstep ((unlab_of_trs ^^ 0) R)) ((unlab_term ^^ 0) (t i \<cdot> \<sigma> i))" by (simp only: min_lab_ichain.simps lab_minimal_defs)
    hence SN: "\<And> i. SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by simp
    from chain have min: "\<And> i. lab_minimal_term R m (t i \<cdot> \<sigma> i)" by (simp add: lab_minimal_cond_def)
    from left_var_imp_not_SN[where R = R] SN have "\<And> x r. (Var x, r) \<notin> R" by blast
    hence no_left_var: "\<And> l r. (l,r) \<in> R \<Longrightarrow> \<not> is_var l" by force
    have "\<forall> i. (?ls i, ?lt i) \<in> ?lP"
    proof
      fix i
      have memI: "(s i, t i) \<in> P" using mem by simp
      have "wf_ass (subst_ass default_ass (\<sigma> i))" 
	by (rule wf_ass_subst_ass, rule wf_default_ass)
      hence "(?ls i, ?lt i) \<in> lab_root_rule I L L' label C (s i, t i)"
	unfolding lab_root_rule_def by (auto)
      with memI show "(?ls i, ?lt i) \<in> ?lP" by auto
    qed
    hence "\<And> i. (?ls i, ?lt i) \<in> lP" using lP by auto
    hence Psteps: "\<And> i. (?lss i, ?lts i) \<in> CS lP" unfolding s t by auto
    have min2: "\<And> i. lab_minimal_term R m (unlab_term (?lts i))" 
      by (simp only: unlab_term_root, rule min)
    have minimal: "\<And> i. lab_minimal_term lR (Suc m) (?lts i)"
      by (rule min_intro[OF min2])
    from chain have steps: "\<And> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    {
      fix i
      from bspec[OF nonvar_def mem[of i]] have "\<not> is_var (t i \<cdot> \<sigma> i)" and "\<not> defined R (the (root (t i \<cdot> \<sigma> i)), num_args (t i \<cdot> \<sigma> i))" 
        by auto
      from rsteps_imp_nrrsteps[OF this _ steps[of i]] no_left_var have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrrstep R)^*" by auto
    } 
    from  quasi_min_nrrewrites_imp_lab_rewrites[OF this min qmodel] have "\<And> i. (?lts i, ?lss (Suc i)) \<in> (rstep ?lR \<union> ?Drl)^*"  .
    with rtrancl_mono[OF rstepSub] have "\<And> i. (?lts i, ?lss (Suc i)) \<in> ((rstep lR)^* \<union> ?Drl)^*" by blast
    with decr_root rtrancl_mono[of "(rstep lR)^* \<union> ?Drl" "(rstep lR)^* \<union> (CS (lP \<union> lR) \<inter> ?M)^+"]
    have steps: "\<And> i. (?lts i, ?lss (Suc i)) \<in> ((rstep lR)^* \<union> (CS (lP \<union> lR) \<inter> ?M)^+)^*" by auto
    let ?A = "rstep lR"
    let ?B = "CS (lP \<union> lR) \<inter> ?M"
    have small: "?B^+ \<subseteq> ?B^*" by auto
    have "(?A^* \<union> ?B^+)^* \<subseteq> (?A^* \<union> ?B^*)^*" by (rule rtrancl_mono, rule Un_mono[OF _ small], simp)
    also have "\<dots> \<subseteq> ((?A \<union> ?B)^* \<union> (?A \<union> ?B)^*)^*" by (rule rtrancl_mono, rule Un_mono[OF rtrancl_mono rtrancl_mono], auto)
    also have "\<dots> = (?A \<union> ?B)^*" using rtrancl_idemp by simp
    also have "\<dots> = ((?A \<union> CS lR \<inter> ?M) \<union> CS lP \<inter> ?M)^*" unfolding cs_union_distr Un_assoc Int_Un_distrib2 using Un_commute[of "CS lP \<inter> ?M" "CS lR \<inter> ?M"] by auto
    also have "\<dots> \<subseteq> ((?A \<union> CS lR) \<union> CS lP \<inter> ?M)^*" by (rule rtrancl_mono, auto)
    also have "\<dots> = (?A \<union> CS lP \<inter> ?M)^*" unfolding rstep_CC_CS using Un_absorb2[OF subset_ctxt_closure[of "CS lR"]] by simp
    finally have steps: "\<And> i. (?lts i, ?lss (Suc i)) \<in> (rstep lR \<union> CS lP \<inter> ?M)^*" using steps by auto
    let ?rel = "CS lP \<inter> ?M"
    from Psteps minimal have minPsteps: "\<And> i. (?lss i, ?lts i) \<in> ?rel" by auto
    let ?rel' = "?rel \<union> rstep lR"
    from steps have steps: "\<And> i. (?lts i, ?lss (Suc i)) \<in> ?rel'^*" unfolding Un_commute[of ?rel "rstep lR"] .
    have SN: "SN (?rel'^* O ?rel O ?rel'^*)" unfolding m[symmetric] rel_step_trancl_conv[symmetric] SN_trancl_SN_conv rel_SN_def[symmetric] finite_dpp_iff_rel_SN[symmetric]
      by (rule finite)
    from steps minPsteps have "\<And> i. (?lss i, ?lss (Suc i)) \<in> ?rel'^* O ?rel O ?rel'^*" by auto
    hence "\<not> SN_elt (?rel'^* O ?rel O ?rel'^*) (?lss 0)" unfolding SN_elt_def by auto
    with SN show False unfolding SN_def by best
  qed
qed


definition sem_lab_root_all_proc :: "
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  (('f,'l)lab,'v)trs \<Rightarrow> 
  nat \<Rightarrow> 
  bool"
  where "sem_lab_root_all_proc P R m lP lR rm \<equiv> rm = Suc m \<and> (\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> \<not> defined R (the (root t), num_args t)) 
     \<and> Lab_root_all_trs P \<subseteq> lP \<and> Lab_trs R \<subseteq> lR 
     \<and> Decr \<subseteq> (rstep lR)^+ 
     \<and> R_subset_lab_decr gr LS R lR 
     \<and> qmodel I L label C cge R"

lemma label_inject2: "label f n l = label g n l' \<Longrightarrow> f = g \<and> l = l'"
  by (cases l, (cases l', auto)+)

lemma sem_lab_root_all_proc_sound: 
  shows "sound_lab_proc sem_lab_root_all_proc"
unfolding sound_lab_proc_def
proof (intro allI impI, unfold sem_lab_root_all_proc_def, (erule conjE)+ )
  fix P R :: "(('f,'l)lab,'v)trs" and m and lP lR :: "(('f,'l)lab,'v)trs" and rm
  let ?M = "{(s,t) | s t. lab_minimal_term lR (Suc m) t} :: (('f,'l)lab,'v)trs"
  let ?Ml = "{(s,t) | s t. lab_minimal_term R  m (unlab_term s)} :: (('f,'l)lab,'v)trs"
  assume finite: "finite_lab_dpp (lP, lR) rm" and lP: "Lab_root_all_trs P \<subseteq> lP" and lR1: "Lab_trs R \<subseteq> lR" and subset2: "R_subset_lab_decr gr LS R lR" 
      and decr: "Decr \<subseteq> (rstep lR)^+" and qmodel: "qmodel I L label C cge R" and m: "rm = Suc m"
      and nonvar_def: "\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> \<not> defined R (the (root t), num_args t)"
  hence nonvar_s: "\<And> s t. (s,t) \<in> P \<Longrightarrow> \<not> is_var s"
    and nonvar_t: "\<And> s t. (s,t) \<in> P \<Longrightarrow> \<not> is_var t" by auto
  note min_intro = lab_minimal_termI[OF gr_SN subset2]
  let ?lr = "Lab_trs R"
  let ?D = "Decr"
  let ?lR = "Lab_trs R \<union> Decr"
  let ?Dr = "Decr_root"
  from rstep_mono[OF decr] have "rstep ?D \<subseteq> (rstep lR)^+" using rstep_trancl_distrib[of "rstep lR", simplified rstep_rstep] by auto
  hence "rstep ?D \<subseteq> (rstep lR)^+" unfolding rstep_union using trancl_mono by blast
  hence one: "rstep ?D \<subseteq> (rstep lR)^*" by auto
  from rstep_mono[OF lR1] have "rstep ?lr \<subseteq> (rstep (lR))^*" by auto
  with one have "rstep ?lR \<subseteq> (rstep lR)^*" unfolding rstep_union by auto
  hence "nrrstep ?lR \<subseteq> (rstep lR)^*" unfolding rstep_iff_rrstep_or_nrrstep by auto
  from rtrancl_mono[OF this] have nrrstepSub: "(nrrstep ?lR)^* \<subseteq> (rstep lR)^*" by simp
  from subset2 have lR2: "unlab_of_trs lR \<subseteq> R" unfolding R_subset_lab_decr_def unlab_of_trs_def decr_of_ord_def decr_of_trs_def 
    by (auto simp: SN_imp_irreflexive[OF gr_SN] label_inject map_trs.simps)
  let ?lP = "Lab_root_all_trs P"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    let ?ls = "lab_root_terms s \<sigma>"
    let ?lt = "lab_root_terms t \<sigma>"
    let ?lsig = "lab_substs \<sigma>"
    let ?lss = "\<lambda> i. Lab_root default_ass (s i \<cdot> \<sigma> i)"
    let ?lts = "\<lambda> i. Lab_root default_ass (t i \<cdot> \<sigma> i)"
    from chain have mem: "\<And> i. (s i, t i) \<in> P" by simp
    have s: "\<And> i. ?lss i = ?ls i \<cdot> ?lsig i" using nonvar_def mem lab_root_subst by blast
    have t: "\<And> i. ?lts i = ?lt i \<cdot> ?lsig i" using nonvar_def mem lab_root_subst by blast
    from chain have "\<And> i. SN_elt (rstep ((unlab_of_trs ^^ 0) R)) ((unlab_term ^^ 0) (t i \<cdot> \<sigma> i))" by (simp only: min_lab_ichain.simps lab_minimal_defs)
    hence SN: "\<And> i. SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by simp
    from chain have min: "\<And> i. lab_minimal_term R m (t i \<cdot> \<sigma> i)" by (simp add: lab_minimal_cond_def)
    from left_var_imp_not_SN[where R = R] SN have "\<And> x r. (Var x, r) \<notin> R" by blast
    hence no_left_var: "\<And> l r. (l,r) \<in> R \<Longrightarrow> \<not> is_var l" by force
    from chain have steps: "\<And> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    {
      fix i
      from bspec[OF nonvar_def mem[of i]] have "\<not> is_var (t i \<cdot> \<sigma> i)" and "\<not> defined R (the (root (t i \<cdot> \<sigma> i)), num_args (t i \<cdot> \<sigma> i))" 
        by auto
      from rsteps_imp_nrrsteps[OF this _ steps[of i]] no_left_var have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrrstep R)^*" by auto
    } note nrrsteps = this
    from quasi_nrrewrites_imp_lab_rewrites[OF this qmodel] have lsteps: "\<And> i. (?lts i, ?lss (Suc i)) \<in> ?Dr^* O (nrrstep ?lR)^*" .
    {
      fix i
      from bspec[OF nonvar_def mem[of i]] obtain f ts where ti: "t i = Fun f ts" by (cases "t i", auto)
      from bspec[OF nonvar_def mem[of "Suc i"]] obtain g ss where ssi: "s (Suc i) = Fun g ss" by (cases "s (Suc i)", auto)
      note steps = nrrsteps[of i, unfolded ti]
      from nrrsteps_preserve_root[OF steps[simplified]] ssi have ssi: "s (Suc i) = Fun f ss" by simp
      from nrrsteps_equiv_num_args[OF steps] ssi have len: "length ts = length ss" by simp
      from lsteps[of i] obtain u where tu: "(?lts i, u) \<in> ?Dr^*" and us: "(u, ?lss (Suc i)) \<in> (nrrstep ?lR)^*" by auto
      let ?cs = "(map (\<lambda> x. eval I L label default_ass (x \<cdot> \<sigma> i)) ts)"
      let ?l = "L' f ?cs"
      obtain l where l: "l = ?l" by simp
      let ?ts = "map (\<lambda> x. lab I L label default_ass (x \<cdot> \<sigma> i)) ts"
      obtain tts where tts: "tts = ?ts" by simp
      let ?f = "label f (length tts) l"
      have lti: "?lts i = Fun ?f tts" unfolding ti l tts  by (simp add: o_def)
      from tu[unfolded lti] have "\<exists> l'. u = Fun (label f (length tts) l') tts \<and> lge f (length tts) l l'"
      proof (induct)
        case base 
        have len: "length ?ts = length ?cs" by simp
        show ?case 
        proof (rule exI[of _ l], intro conjI[OF refl], rule lge_refl)
          show "l \<in> LS' f (length tts)"
            unfolding l tts len
          proof (rule wf_L'[unfolded wf_label_def, THEN spec, THEN spec, THEN mp])
            show "set ?cs \<subseteq> C"              
              using wf_term[OF wf_default_ass] by (auto simp del: eval_subst)
          qed
        qed
      next
        case (step v w)
        from step(3) obtain l' where v: "v = Fun (label f (length tts) l') tts" and lge: "lge f (length tts) l l'" by auto
        from step(2)[unfolded v decr_of_ord_def] obtain l1 l2 g where left: "Fun (label f (length tts) l') tts = Fun (label g (length tts) l1) tts" and
           right: "w = Fun (label g (length tts) l2) tts" and gr: "(l1,l2) \<in> gr_root g (length tts)" by auto        
        from left have gf: "g = f" and l1: "l1 = l'" using label_inject2[of f "length tts" l' g] by auto
        note w =  right[unfolded gf]
        note gr = gr[unfolded l1 gf]
        from gr[unfolded lge_to_lgr_rel_def lge_to_lgr_def Let_def] 
        have lge2: "lge f (length tts) l' l2" by auto
        show ?case
          by (intro exI conjI, rule w, rule lge_trans[OF lge lge2])
      qed
      then obtain l' where u: "u = Fun (label f (length tts) l') tts" and lge: "lge f (length tts) l l'" by auto
      let ?g = "label f (length ts) l'"
      let ?lt = "Fun ?g (map (Lab (subst_ass default_ass (\<sigma> i))) ts)"
      from us nrrstepSub have us: "(u, ?lss (Suc i)) \<in> (rstep lR)^*" by auto 
      have memI: "(s i, t i) \<in> P" using mem by simp
      let ?ass = "subst_ass default_ass (\<sigma> i)"
      have wf_ass: "wf_ass ?ass"
	by (rule wf_ass_subst_ass, rule wf_default_ass)
      have l2: "L' f (map (eval I L label ?ass) ts) = l"
        unfolding l by (simp)
      have "(?ls i, ?lt) \<in> lab_root_all_rule I L L' label lge C (s i, t i)"        
	unfolding lab_root_all_rule_def
        by (rule, rule exI[of _ ?ass], intro exI conjI, simp, rule wf_ass, simp add: ti o_def, rule exI[of _ l'], simp add: l2, insert lge, unfold tts, simp)
      with set_mp[OF lP] memI have pair: "(?ls i, ?lt) \<in> lP" by blast
      have u: "u = ?lt \<cdot> ?lsig i"
        unfolding u tts by simp
      have min2: "lab_minimal_term R m (unlab_term (?lts i))" 
        by (simp only: unlab_term_root, rule min)
      from min2 have "lab_minimal_term R m (unlab_term u)" unfolding u ti
        by (simp add: o_def)
      note minimal = min_intro[OF this] 
      from pair us[unfolded u s[of "Suc i"]] minimal[unfolded u] have "\<exists> t. (?ls i,t) \<in> lP \<and> (t \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep lR)^* \<and> lab_minimal_term lR (Suc m) (t \<cdot> ?lsig i)" by blast
    }
    hence "\<forall> i. \<exists> t. (?ls i,t) \<in> lP \<and> (t \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep lR)^* \<and> lab_minimal_term lR (Suc m) (t \<cdot> ?lsig i)" by blast
    from choice[OF this] obtain lt where "\<And> i. (?ls i,lt i) \<in> lP \<and> (lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep lR)^* \<and> lab_minimal_term lR (Suc m) (lt i \<cdot> ?lsig i)" by blast
    hence "min_lab_ichain (lP,lR) (Suc m) ?ls lt ?lsig" by (auto simp: lab_minimal_cond_def)
    with finite show False unfolding finite_lab_dpp_def m by blast
  qed
qed


end


subsection {* head for concrete carriers *}


datatype arithFun = Arg nat | Const nat | Sum "arithFun list" | Max "arithFun list" | Min "arithFun list" | Prod "arithFun list"

text {* An alternative induction scheme for arithFuns. *}
lemma
  fixes P :: "arithFun \<Rightarrow> bool"
  assumes "\<And>n. P(Arg n)" and "\<And>n. P(Const n)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Sum fs)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Max fs)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Min fs)"
    and "\<And>fs. (\<And>f. f \<in> set fs \<Longrightarrow> P f) \<Longrightarrow> P(Prod fs)"
  shows arithFun_induct[case_names Arg Const Sum Max Min Prod,induct type: arithFun]: "P f"
    and "\<And>f. f \<in> set fs \<Longrightarrow> P f"
proof (induct f and fs)
  fix n::nat show "P(Arg n)" by (simp add: assms)
next
  fix n::nat show "P(Const n)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<forall>f\<in>set fs. P f" thus "P(Sum fs)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<forall>f\<in>set fs. P f" thus "P(Max fs)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<forall>f\<in>set fs. P f" thus "P(Min fs)" by (simp add: assms)
next
  fix fs::"arithFun list" assume "\<And>f. f \<in> set fs \<Longrightarrow> P f" thus "P(Prod fs)" by (simp add: assms)
next
  show "\<forall>f\<in>set []. P f" by simp
next
  fix f fs assume "P f" and "\<forall>f\<in>set fs. P f" thus "\<forall>g\<in>set(f#fs). P g" by simp
next
  show "\<forall>f\<in>set []. P f" by simp
next
  fix f fs assume "P f" and "\<forall>f\<in>set fs. P f" thus "\<forall>g\<in>set(f#fs). P g" by simp
next
  show "\<forall>f\<in>set []. P f" by simp
next
  fix f fs assume "P f" and "\<forall>f\<in>set fs. P f" thus "\<forall>g\<in>set(f#fs). P g" by simp
next
  fix f::"arithFun" assume "f \<in> set []" thus "P f" by simp
next
  fix f fs g assume "P f" and "\<And>f. f \<in> set fs \<Longrightarrow> P f" and "g \<in> set(f#fs)" thus "P g" by auto
qed

instantiation arithFun :: "show" begin
  fun shows_arithFun :: "arithFun \<Rightarrow> shows"
  where
    "shows_arithFun (Arg i) = (''x_'' +#+ shows (Suc i))" |
    "shows_arithFun (Const n) = shows n" |
    "shows_arithFun (Sum fs) = shows_list_gen shows_arithFun [] [] '' + '' [] fs" |
    "shows_arithFun (Prod fs) = shows_list_gen shows_arithFun [] [] '' * '' [] fs" |
    "shows_arithFun (Min fs) = shows_list_gen shows_arithFun [] ''min('' '','' '')'' fs" |
    "shows_arithFun (Max fs) = shows_list_gen shows_arithFun [] ''max('' '','' '')'' fs"
  definition "shows_prec (d::nat) f = shows_arithFun f"
  definition "shows_list (as::arithFun list) \<equiv> shows_list_aux shows as"

  lemma assoc_arithFun: "shows_prec d (f::arithFun) r @ s = shows_prec d f (r @ s)"
  proof (induct f arbitrary: r s)
    case (Arg n) show ?case unfolding shows_arithFun.simps shows_string_def o_def assoc[symmetric]
      unfolding shows_prec_arithFun_def by simp
  next
    case (Const n) show ?case unfolding shows_arithFun.simps shows_string_def o_def assoc[symmetric]
      unfolding shows_prec_arithFun_def by simp
  next
    case (Sum fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  next
    case (Max fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  next
    case (Min fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  next
    case (Prod fs) with shows_list_gen_assoc show ?case unfolding shows_arithFun.simps
      unfolding shows_prec_arithFun_def by force
  qed
  lemma assoc_arithFun_elt: "\<forall>(f::arithFun)\<in>set fs. shows_prec d f r @ s = shows_prec d f (r @ s)"
    using assoc_arithFun by auto
  instance by
    (intro_classes,rule assoc_arithFun,unfold shows_list_arithFun_def)
    (rule shows_list_aux_assoc[OF assoc_arithFun_elt])
end

fun take_default where "take_default def [] _ = def" | "take_default _ (x # xs) 0 = x" | "take_default def (_ # xs) (Suc i) = take_default def xs i"

fun eval_arithFun_unbound :: "nat list \<Rightarrow> arithFun \<Rightarrow> nat"
where "eval_arithFun_unbound nats (Arg i) = take_default 0 nats i"
    | "eval_arithFun_unbound nats (Const n) = n"
    | "eval_arithFun_unbound nats (Sum []) = 0"
    | "eval_arithFun_unbound nats (Sum (f # fs)) =  (eval_arithFun_unbound nats f + eval_arithFun_unbound nats (Sum fs))"
    | "eval_arithFun_unbound nats (Prod []) = 1"
    | "eval_arithFun_unbound nats (Prod (f # fs)) =  (eval_arithFun_unbound nats f * eval_arithFun_unbound nats (Prod fs))"
    | "eval_arithFun_unbound nats (Max [f]) = eval_arithFun_unbound nats f"
    | "eval_arithFun_unbound nats (Max (f # fs)) =  (max (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Max fs)))"
    | "eval_arithFun_unbound nats (Min [f]) = eval_arithFun_unbound nats f"
    | "eval_arithFun_unbound nats (Min (f # fs)) =  (min (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Min fs)))"

declare eval_arithFun_unbound.simps[simp del]

fun eval_arithFun :: "nat \<Rightarrow> arithFun \<Rightarrow> nat list \<Rightarrow> nat"
where "eval_arithFun c f nats = (eval_arithFun_unbound nats f) mod c"


datatype ('f,'l)sl_inter = SL_Inter nat "((('f,'l)lab  \<times> nat) \<times> arithFun)list"

instantiation sl_inter :: ("show", "show") "show" begin
  fun
    shows_sl_inter :: "('f::show, 'l::show) sl_inter \<Rightarrow> shows"
  where
    "shows_sl_inter(SL_Inter n fnas) = (
       ''carrier {0,...,'' +#+ shows n +@+ ''}'' +#+ shows_nl +@+
       shows_map (\<lambda>((f,n),a). ''['' +#+ shows f +@+ ''/'' +#+ shows n +@+ ''] = '' +#+ shows a +@+ shows_nl) fnas
     )"
  definition "shows_prec (d::nat) sli = shows_sl_inter sli"
  definition "shows_list_sl_inter (ss::('f::show,'l::show)sl_inter list) \<equiv> shows_list_aux shows ss"

  lemma assoc_sl_inter: "shows_prec d (sli::('f::show,'l::show)sl_inter) r @ s = shows_prec d sli (r @ s)"
  proof (cases sli)
    fix n::nat and fnas::"((('f::show,'l::show)lab\<times>nat)\<times>arithFun)list" assume assm: "sli = SL_Inter n fnas"
    let ?elt = "\<lambda>((f::('f::show,'l::show)lab,n::nat),a::arithFun). ''['' +#+ shows f +@+ ''/'' +#+ shows n +@+ ''] = '' +#+ shows a +@+ shows_nl"
    have "\<And>r s x. ?elt x r @ s = ?elt x (r@s)" unfolding show_defs o_def unfolding assoc[symmetric] by auto
    hence elt: "\<And>r s.\<forall>x\<in>set fnas. ?elt x r @ s = ?elt x (r@s)" by blast
    show ?thesis unfolding assm unfolding shows_prec_sl_inter_def shows_sl_inter.simps split_def show_defs o_def unfolding assoc[symmetric]
      unfolding shows_map_assoc[OF elt,of r s,unfolded show_defs o_def split_def,symmetric,unfolded assoc[symmetric]] by simp
  qed
  lemma assoc_sl_inter_elt: "\<forall>(sli::('f::show,'l::show)sl_inter)\<in>set slis. shows sli r @ s = shows sli (r@s)"
    using assoc_sl_inter by auto
  instance by
    (intro_classes,rule assoc_sl_inter,unfold shows_list_sl_inter_def)
    (rule shows_list_aux_assoc[OF assoc_sl_inter_elt])
end

fun get_largest_element where "get_largest_element (SL_Inter n _) = n"

fun sl_inter_to_inter :: "('f,'l)sl_inter \<Rightarrow> (('f,'l)lab,nat)inter"
where "sl_inter_to_inter (SL_Inter c ls) fl cs = (case lookup (fl,length cs) ls of 
       None \<Rightarrow> 0
     | Some f \<Rightarrow> eval_arithFun (Suc c) f cs)"

lemma wf_sl_inter: 
  shows "wf_inter (sl_inter_to_inter sli) {x | x.  x < Suc (get_largest_element sli)}"
proof (cases sli)
  case (SL_Inter c ls)
  thus ?thesis unfolding wf_inter_def
  proof (intro allI, intro impI)
    fix f cs
    show "sl_inter_to_inter sli f cs \<in> {x | x. x < Suc (get_largest_element sli)}" using SL_Inter
      by (cases "lookup (f, length cs) ls", auto)
  qed
qed

definition model_cge where "model_cge \<equiv> op ="
definition model_lge where "model_lge \<equiv> \<lambda> _ _. op ="
definition model_LS where "model_LS \<equiv> \<lambda> _ _. UNIV"

interpretation sl_finite: sl_interpr_lab "{x | x . x < Suc (get_largest_element sli)}" "get_largest_element sli" "sl_inter_to_inter sli" 
  model_cge model_lge L model_LS for L
  by (unfold_locales, 
      simp,
      rule wf_sl_inter,
      unfold wf_label_def, simp add: model_LS_def,
      unfold cge_wm, simp add: model_cge_def,
      unfold lge_wm, simp add: model_lge_def model_cge_def,
      simp add: model_cge_def model_lge_def lge_to_lgr_def lge_to_lgr_rel_def Let_def,
      rule label_decomp_label)


(* now let us head for checking the model condition *)


lemma eval_lab_independent:
  fixes t::"('f,'v)term" and I::"('f,'c)inter"
  assumes "\<forall>x\<in>vars_term t. \<alpha> x = \<beta> x"
  shows "eval_lab I L LC \<alpha> t = eval_lab I L LC \<beta> t"
using assms
proof (induct t, simp)
  case (Fun f ts)
  hence map: "map (eval_lab I L LC \<alpha>) ts = map (eval_lab I L LC \<beta>) ts" by auto
  thus ?case by (auto simp: Let_def map)
qed

lemma lab_root_independent:
  fixes t::"('f,'v)term" and I::"('f,'c)inter"
  assumes vars: "\<forall>x\<in>vars_term t. \<alpha> x = \<beta> x"
  shows "lab_root I L L' LC \<alpha> t = lab_root I L L' LC \<beta> t"
proof (cases t, simp)
  case (Fun f ts)
  from vars[unfolded Fun] have vars: "\<And> t . t \<in> set ts \<Longrightarrow> (\<forall> x \<in> vars_term t. \<alpha> x = \<beta> x)" by auto
  from eval_lab_independent[OF vars]
  have map: "map (eval_lab I L LC \<alpha>) ts = map (eval_lab I L LC \<beta>) ts" by auto
  show ?thesis unfolding Fun by (auto simp: Let_def map)
qed

fun
  check_sl_rule_ass ::
    "bool \<Rightarrow> ('f::show,'c::show)inter \<Rightarrow> ('f,'c,'l)label
     \<Rightarrow> ('f,'l,'lf :: show)lcompose 
     \<Rightarrow> ('c \<Rightarrow> 'c \<Rightarrow> bool)
     \<Rightarrow> ('lf,'v)rules
     \<Rightarrow> ('v::show,'c)assign \<Rightarrow> ('f,'v)rule \<Rightarrow> shows check"
where
  "check_sl_rule_ass mc I L LC cge lR \<alpha> (l, r) = do {
     let cl_ll = eval_lab I L LC \<alpha> l;
     let cr_lr = eval_lab I L LC \<alpha> r;
     check (mc \<longrightarrow> (cge (fst cl_ll) (fst cr_lr)))
       (''rule '' +#+ shows_rule (l, r) +@+ '' violates the model condition, [lhs] = ''
         +#+ shows (fst cl_ll) +@+ '', [rhs] = '' +#+ shows (fst cr_lr));
     check ((snd cl_ll, snd cr_lr) \<in> set lR)
       (''labeled rule '' +#+ shows_rule (snd cl_ll, snd cr_lr) +@+ shows_string '' missing'')
   }"

fun
  check_sl_rule_all_ass ::
    "('f::show,'c::show)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('f,'c,'l)label
     \<Rightarrow> ('f,'l,'lf :: show)lcompose 
     \<Rightarrow> ('l \<Rightarrow> 'l list)
     \<Rightarrow> ('lf,'v)rules
     \<Rightarrow> ('v::show,'c)assign \<Rightarrow> ('f,'v)rule \<Rightarrow> shows check"
where
  "check_sl_rule_all_ass I L L' LC gen_smaller lR \<alpha> (l, Fun f ts) = do {
     let ll = lab_root I L L' LC \<alpha> l;
         clts = map (eval_lab I L LC \<alpha>) ts;
         lts = map snd clts;
         l'   = L' f (map fst clts);
         n = length ts;
         small = gen_smaller l' in
     check_allm (\<lambda> l'. check ((ll, Fun (LC f n l') lts) \<in> set lR) 
       (''labeled rule '' +#+ shows_rule (ll, Fun (LC f n l') lts) +@+ shows_string '' missing'')) small
   }" |
  "check_sl_rule_all_ass I L L' LC gen_smaller lR \<alpha> (l, Var x) = do {
     let ll = lab_root I L L' LC \<alpha> l;
     let lr = lab_root I L L' LC \<alpha> (Var x) in
   check ((ll, lr) \<in> set lR) 
       (''labeled rule '' +#+ shows_rule (ll, lr) +@+ shows_string '' missing'')
   }"


fun
  check_sl_rule ::
    "('f::show,'c::show)inter \<Rightarrow> ('f,'c,'l)label 
     \<Rightarrow> ('f,'l,'lf :: show)lcompose 
     \<Rightarrow> 'c list 
     \<Rightarrow> ('c \<Rightarrow> 'c \<Rightarrow> bool)
     \<Rightarrow> bool
     \<Rightarrow> ('lf :: show,'v)rules \<Rightarrow> ('f,'v::show)rule \<Rightarrow> shows check"
where
  "check_sl_rule I L LC C cge mc lR (l,r) = check_allm (\<lambda>\<alpha>. check_sl_rule_ass mc I L LC cge lR \<alpha> (l,r)) 
     (map assign_from_vector (enum_vectors C (list_union (vars_term_list l) (vars_term_list r))))"

fun
  check_sl_rule_all ::
    "('f::show,'c::show)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('f,'c,'l)label 
     \<Rightarrow> ('f,'l,'lf :: show)lcompose 
     \<Rightarrow> 'c list 
     \<Rightarrow> ('l \<Rightarrow> 'l list)
     \<Rightarrow> ('lf :: show,'v)rules \<Rightarrow> ('f,'v::show)rule \<Rightarrow> shows check"
where
  "check_sl_rule_all I L L' LC C gen_smaller lR (l,r) = check_allm (\<lambda>\<alpha>. check_sl_rule_all_ass I L L' LC gen_smaller lR \<alpha> (l,r)) 
     (map assign_from_vector (enum_vectors C (list_union (vars_term_list l) (vars_term_list r))))"

lemma check_sl_rule_sound[simp]:
  fixes lR :: "('f::show,'v::show)rules" and I :: "('f,'c::show)inter" 
  assumes "isOK (check_sl_rule I L LC C cge mc lR (l,r))"
  shows "(mc \<longrightarrow> qmodel_rule I L LC (set C) cge l r) \<and> lab_rule I L LC (set C) (l,r) \<subseteq> set lR"
proof (cases C)
  case Nil
  thus ?thesis unfolding lab_rule_def wf_assign_def by auto
next
  case (Cons c D)
  let ?C = "Cons c D"
  let ?V = "list_union (vars_term_list l) (vars_term_list r) :: 'v list"
  from assms Cons have check: "\<forall> \<alpha> \<in> set (map assign_from_vector (enum_vectors ?C ?V)). isOK(check_sl_rule_ass mc I L LC cge lR \<alpha> (l,r))"
    by auto
  have "\<forall> \<alpha>. wf_assign (set ?C) \<alpha> \<longrightarrow> isOK(check_sl_rule_ass mc I L LC cge lR \<alpha> (l,r))"
  proof (intro allI, intro impI)
    fix \<alpha> :: "'v \<Rightarrow> 'c"
    assume wf_ass: "wf_assign (set ?C) \<alpha>"
    from enum_vectors_complete obtain vec where vec: "vec \<in> set (enum_vectors ?C ?V) \<and> (\<forall> x \<in> set ?V. \<forall> c \<in> set ?C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c)"
      by best
    let ?\<beta> = "assign_from_vector vec"
    from wf_ass vec have "\<forall> x \<in> set ?V. \<alpha> x = ?\<beta> x" unfolding wf_assign_def by auto
    hence l: "(\<forall> x \<in> vars_term l. ?\<beta> x = \<alpha> x)" and r: "(\<forall> x \<in> vars_term r. ?\<beta> x = \<alpha> x) " by auto
    have equal: "eval_lab I L LC \<alpha> l = eval_lab I L LC ?\<beta> l \<and> eval_lab I L LC \<alpha> r = eval_lab I L LC ?\<beta> r" 
      unfolding eval_lab_independent[OF l] eval_lab_independent[OF r] by auto
    from vec check have "isOK(check_sl_rule_ass mc I L LC cge lR ?\<beta> (l,r))" by auto
    with equal show "isOK(check_sl_rule_ass mc I L LC cge lR \<alpha> (l,r))" by auto
  qed
  thus ?thesis using Cons unfolding lab_rule_def by (auto simp: Let_def)
qed

declare check_sl_rule.simps[simp del]

lemma check_sl_rule_all_sound:
  fixes lR :: "('f::show,'v::show)rules" and I :: "('f,'c::show)inter" 
  assumes ok: "isOK (check_sl_rule_all I L L' LC C gen_smaller lR (l,r))"
  and gen: "\<And> f n l l'. lge f n l l' \<Longrightarrow> l' \<in> set (gen_smaller l)"
  shows "lab_root_all_rule I L L' LC lge (set C) (l,r) \<subseteq> set lR"
proof (cases C)
  case Nil
  thus ?thesis unfolding lab_root_all_rule_def wf_assign_def by auto
next
  case (Cons c D)
  let ?C = "Cons c D"
  let ?V = "list_union (vars_term_list l) (vars_term_list r) :: 'v list"
  from assms Cons have check: "\<forall> \<alpha> \<in> set (map assign_from_vector (enum_vectors ?C ?V)). isOK(check_sl_rule_all_ass I L L' LC gen_smaller lR \<alpha> (l,r))"
    by auto
  have check: "\<forall> \<alpha>. wf_assign (set ?C) \<alpha> \<longrightarrow> isOK(check_sl_rule_all_ass I L L' LC gen_smaller lR \<alpha> (l,r))"
  proof (intro allI, intro impI)
    fix \<alpha> :: "'v \<Rightarrow> 'c"
    assume wf_ass: "wf_assign (set ?C) \<alpha>"
    from enum_vectors_complete obtain vec where vec: "vec \<in> set (enum_vectors ?C ?V) \<and> (\<forall> x \<in> set ?V. \<forall> c \<in> set ?C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c)"
      by best
    let ?\<beta> = "assign_from_vector vec"
    from wf_ass vec have "\<forall> x \<in> set ?V. \<alpha> x = ?\<beta> x" unfolding wf_assign_def by auto
    hence l: "(\<forall> x \<in> vars_term l. ?\<beta> x = \<alpha> x)" and r: "(\<forall> x \<in> vars_term r. ?\<beta> x = \<alpha> x) " by auto
    have equal: "lab_root I L L' LC \<alpha> l = lab_root I L L' LC ?\<beta> l \<and> lab_root I L L' LC \<alpha> r = lab_root I L L' LC ?\<beta> r" 
      unfolding lab_root_independent[OF l] lab_root_independent[OF r] by auto
    from vec check have check: "isOK(check_sl_rule_all_ass I L L' LC gen_smaller lR ?\<beta> (l,r))" by auto
    show "isOK(check_sl_rule_all_ass I L L' LC gen_smaller lR \<alpha> (l,r))" 
    proof (cases r)
      case (Var x)
      with equal check show ?thesis by simp
    next
      case (Fun f ts)
      with r have "\<And> t. t \<in> set ts \<Longrightarrow> \<forall> x \<in> vars_term t. ?\<beta> x = \<alpha> x" by auto
      from eval_lab_independent[OF this, of _ I L LC] 
      have map: "map (eval_lab I L LC \<alpha>) ts = map (eval_lab I L LC ?\<beta>) ts" by auto
      from equal[THEN conjunct1] check
      show ?thesis unfolding Fun
        by (simp add: Let_def map)
    qed
  qed
  note check = check[unfolded Cons[symmetric], THEN spec, THEN mp]
  { 
    fix \<alpha> lr
    assume wf: "wf_assign (set C) \<alpha>" and lr: "lr \<in> lab_root_all I L L' LC lge \<alpha> r"
    note ok = check[OF wf]
    have "(lab_root I L L' LC \<alpha> l, lr) \<in> set lR" 
    proof (cases r)
      case (Var x)
      with ok lr show ?thesis by (simp add: Let_def)
    next
      case (Fun f ts)
      let ?n = "length ts"
      from lr[unfolded Fun]
      obtain l' where lr: "lr = Fun (LC f ?n l') (map (lab I L LC \<alpha>) ts)" and lge: "lge f ?n (L' f (map (eval I L LC \<alpha>) ts)) l'" 
        by (auto simp: o_def)
      from gen[OF lge] have l': "l' \<in> set (gen_smaller (L' f (map (eval I L LC \<alpha>) ts)))" by simp
      from ok[unfolded Fun] l' show ?thesis
        unfolding lr by (auto simp: Let_def o_def)
    qed
  }
  thus ?thesis unfolding lab_root_all_rule_def by auto 
qed

declare check_sl_rule_all.simps[simp del]

type_synonym ('f,'c,'l,'v)sl_check1 =
  "(('f,'l)lab,'c)inter \<Rightarrow> (('f,'l)lab,'c,'l + ('f,'l)lab list)label \<Rightarrow> 'c list \<Rightarrow> ('c \<Rightarrow> 'c \<Rightarrow> bool) \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules
   \<Rightarrow> shows check"

definition
  check_sl_model_lab_trs  :: "('f::show,'c::show,'l::show,'v::show)sl_check1" 
where
  "check_sl_model_lab_trs I L C cge lR R \<equiv> check_allm (check_sl_rule I L label C cge True lR) R"

definition 
  check_sl_lab_trs :: "('f::show,'c::show,'l::show,'v::show)sl_check1" 
where
  "check_sl_lab_trs I L C cge lP P \<equiv> check_allm (check_sl_rule I L label C cge False lP) P"

definition 
  check_sl_lab_all_trs 
where
  "check_sl_lab_all_trs I L L' C gen lP P \<equiv> check_allm (check_sl_rule_all I L L' label C gen lP) P"

fun is_labeled :: "('f,'l)lab \<Rightarrow> bool" 
  where "is_labeled (UnLab _) = False"
      | "is_labeled (Sharp _) = False"
      | "is_labeled _ = True"

lemma label_is_labeled: "is_labeled (label f n l)" 
  by (cases l, auto)

lemma is_labeled_imp_label: "is_labeled lf \<Longrightarrow>  label (unlab lf) n (snd (label_decomp lf)) = lf" 
  by  (cases lf, auto)

definition
  check_sl_only_lab_and_decr
where
  "check_sl_only_lab_and_decr lge LS R \<equiv>
     check_allm (\<lambda>(ll,lr).
       let l = unlab_term ll; r = unlab_term lr; n = num_args l in
         check
           (if l = r
             then (\<not> is_var ll \<and> \<not> is_var lr 
                   \<and> args ll = args lr
                   \<and> is_labeled (the (root ll)) \<and> is_labeled (the (root lr)) 
                   \<and> LS (the (root l)) n (snd (label_decomp (the (root ll))))
                   \<and> LS (the (root l)) n (snd (label_decomp (the (root lr))))
                   \<and> lge_to_lgr lge LS (the (root l)) n (snd (label_decomp (the (root ll))))
               (snd (label_decomp (the (root lr)))) \<or> (ll = lr \<and> (l,r) \<in> set R))
             else (l,r) \<in> set R)
           (''labeled rule '' +#+ shows_rule(l,r) +@+
             shows_string '' is neither decreasing nor a labeled version of an original rule''))"

lemma split_pairs_all: "(\<forall> x \<in> r :: 'a ars. P x) = (\<forall> a b . (a,b) \<in> r \<longrightarrow> P (a,b))"
  by (auto)

lemma check_sl_only_lab_and_decr_sound[simp]:
  "isOK(check_sl_only_lab_and_decr lge LS R lR) = R_subset_lab_decr (lge_to_lgr_rel lge LS) LS (set R) (set lR)"
proof -   
  {
    fix a b
    let ?gr = "lge_to_lgr lge LS"
    have "((unlab_term a = unlab_term b \<longrightarrow>
            (a, b) \<in> set lR  \<longrightarrow>
            \<not> is_var a \<and>
            \<not> is_var b \<and> 
            args a = args b \<and> is_labeled (the (root a)) \<and> is_labeled (the (root b)) \<and>
            LS (the (root (unlab_term b))) (num_args (unlab_term b)) (snd (label_decomp (the (root a)))) \<and>
            LS (the (root (unlab_term b))) (num_args (unlab_term b)) (snd (label_decomp (the (root b)))) \<and>
            ?gr (the (root (unlab_term b))) (num_args (unlab_term b)) (snd (label_decomp (the (root a)))) (snd (label_decomp (the (root b))))  \<or>
            a = b \<and> (unlab_term b, unlab_term b) \<in> set R ) \<and>
           (unlab_term a \<noteq> unlab_term b \<longrightarrow> set lR (a, b) \<longrightarrow> set R (unlab_term a, unlab_term b))) = 
           ( (a,b) \<in> set lR \<longrightarrow>
            (unlab_term a, unlab_term b) \<in> set R  \<and> (a = b \<or> unlab_term a \<noteq> unlab_term b) \<or>
           (\<exists>f l l' ts.
               a = Fun (label f (length ts) l) ts \<and>
               b = Fun (label f (length ts) l') ts \<and> LS f (length ts) l \<and> LS f (length ts) l' \<and> ?gr f (length ts) l l'))" (is "?one = ?two")
      proof (intro iffI)
        assume ?two
        thus ?one 
          by (auto simp: label_decomp_label label_is_labeled mem_def)
      next
        assume ?one
        show ?two
        proof (cases "(a,b) \<in> set lR")
          case False thus ?two by auto
        next
          case True 
          hence mem: "set lR (a,b)" unfolding mem_def .
          note one = `?one`[simplified mem True, simplified]
          show ?two
          proof (cases "unlab_term a = unlab_term b")
            case False
            from one[simplified False, simplified]
            have "set R (unlab_term a, unlab_term b)" .
            thus ?two 
              by (simp add: True False mem_def)
          next
            case True
            note one = one[simplified True, simplified] 
            thus ?two
            proof
              assume "a = b \<and> (unlab_term b, unlab_term b) \<in> set R"
              thus ?two by simp
            next
              assume ass: "\<not> is_var a \<and>
                \<not> is_var b \<and>
                args a = args b \<and> is_labeled (the (root a)) \<and> is_labeled (the (root b)) \<and>
                LS (the (root (unlab_term b))) (num_args (unlab_term b)) (snd (label_decomp (the (root a)))) \<and>
                LS (the (root (unlab_term b))) (num_args (unlab_term b)) (snd (label_decomp (the (root b)))) \<and>
                ?gr (the (root (unlab_term b))) (num_args (unlab_term b)) (snd (label_decomp (the (root a)))) (snd (label_decomp (the (root b))))"
              from ass obtain lf ts where a: "a = Fun lf ts" by (cases a, auto)
              with ass obtain lf' where b: "b = Fun lf' ts" by (cases b, auto)
              from True have eq: "unlab lf = unlab lf'" unfolding a b by auto
              note one = ass[simplified a b, simplified]
              from one have lf: "is_labeled lf" and lf': "is_labeled lf'" by auto
              from one is_labeled_imp_label[OF lf, of "length ts"] is_labeled_imp_label[OF lf', of "length ts"]
              have "\<exists>f l l' ts. a = Fun (label f (length ts) l) ts \<and> b = Fun (label f (length ts) l') ts \<and> LS f (length ts) l \<and> LS f (length ts) l' \<and> ?gr f (length ts) l l'"
                unfolding a b unfolding eq by (force)
              thus ?two by auto
            qed
          qed
        qed
      qed    
  }
  thus ?thesis unfolding check_sl_only_lab_and_decr_def R_subset_lab_decr_def decr_of_ord_def Let_def
    by (simp, simp only: split_pairs_all, simp add: mem_def lge_to_lgr_rel_def Collect_def)
qed

lemma check_sl_lab_trs_sound:
  assumes "isOK(check_sl_lab_trs I L C cge lP P)"
  shows "lab_trs I L label (set C) (set P) \<subseteq> set lP"
proof -
  from assms have check: "\<forall> (l,r) \<in> set P. isOK(check_sl_rule I L label C cge False lP (l,r))"
    unfolding check_sl_lab_trs_def by auto
  have "\<forall> (l,r) \<in> set P. lab_rule I L label (set C) (l,r) \<subseteq> set lP" 
  proof (clarify)
    fix l r ll lr
    assume lr: "(l,r) \<in> set P" and llr: "(ll,lr) \<in> lab_rule I L label (set C) (l,r)"
    from lr check have "isOK(check_sl_rule I L label C cge False lP (l,r))" by auto 
    hence "lab_rule I L label (set C) (l,r) \<subseteq> set lP" by auto
    with lr llr show "(ll,lr) \<in> set lP" by auto
  qed
  thus ?thesis by auto
qed

lemma check_sl_lab_all_trs_sound:
  assumes ok: "isOK(check_sl_lab_all_trs I L L' C gen lP P)"
  and gen: "\<And> f n l l'. lge f n l l' \<Longrightarrow> l' \<in> set (gen l)"
  shows "lab_root_all_trs I L L' label lge (set C) (set P) \<subseteq> set lP"
proof -
  from ok have check: "\<forall> (l,r) \<in> set P. isOK(check_sl_rule_all I L L' label C gen lP (l,r))"
    unfolding check_sl_lab_all_trs_def by auto
  have "\<forall> (l,r) \<in> set P. lab_root_all_rule I L L' label lge (set C) (l,r) \<subseteq> set lP" 
  proof (clarify)
    fix l r ll lr
    assume lr: "(l,r) \<in> set P" and llr: "(ll,lr) \<in> lab_root_all_rule I L L' label lge (set C) (l,r)"
    from lr check have "isOK(check_sl_rule_all I L L' label C gen lP (l,r))" by auto 
    from check_sl_rule_all_sound[OF this gen]
    have "lab_root_all_rule I L L' label lge (set C) (l,r) \<subseteq> set lP" by simp
    with lr llr show "(ll,lr) \<in> set lP" by auto
  qed
  thus ?thesis by auto
qed

lemma check_sl_model_lab_trs_sound:
  assumes "isOK(check_sl_model_lab_trs I L C cge lR R)"
  shows "lab_trs I L label (set C) (set R) \<subseteq> set lR \<and> qmodel I L label (set C) cge (set R)"
proof -
  from assms have check: "\<forall> (l,r) \<in> set R. isOK(check_sl_rule I L label C cge True lR (l,r))"
    unfolding check_sl_model_lab_trs_def by auto
  have "\<forall> (l,r) \<in> set R. qmodel_rule I L label (set C) cge l r \<and> lab_rule I L label (set C) (l,r) \<subseteq> set lR" 
  proof (clarify)
    fix l r
    assume lr: "(l,r) \<in> set R" 
    from lr check have "isOK(check_sl_rule I L label C cge True lR (l,r))" by auto 
    thus "qmodel_rule I L label (set C) cge l r \<and> lab_rule I L label (set C) (l,r) \<subseteq> set lR" by auto
  qed
  thus ?thesis unfolding qmodel_def by auto
qed


type_synonym ('f,'l,'v)sl_check2 =
  "(('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> shows check"

definition
  check_sem_lab_proc ::
    "shows check \<Rightarrow> ((('f,'l)lab,'v)rules \<Rightarrow> shows check) \<Rightarrow> 
     ('f::show,'l::show,'v::show)sl_check2
     \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> (('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> bool) \<Rightarrow> 
     (('f,'l)lab \<Rightarrow> nat \<Rightarrow> 'l + ('f,'l)lab list \<Rightarrow> bool) \<Rightarrow>
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check"
where
  "check_sem_lab_proc valid check_decr check_lab check_model_lab lge LS P R lP lR \<equiv> do {
     valid;
     check_decr lR;
     check_lab lP P
       <+? (\<lambda>s. ''not all labeled DPs present:'' +#+ shows_nl +@+ s);
     check_model_lab lR R
       <+? (\<lambda>s. ''problem with labeled TRS:'' +#+ shows_nl +@+ s);
     check_sl_only_lab_and_decr lge LS R lR 
       <+? (\<lambda>s. s +@+ shows_nl)
   }"

definition
  check_sem_lab_root_proc 
where
  "check_sem_lab_root_proc valid check_decr check_lab check_model_lab lge LS rm P R lP lR \<equiv> case rm of None \<Rightarrow> OK | Some irm \<Rightarrow> (do {
     valid;
     check_allm (\<lambda>(s, t). check_no_var s \<guillemotright> check_no_var t \<guillemotright> check_no_defined_root irm t) P;
     check_decr lR;
     check_lab lP P
       <+? (\<lambda>s. ''not all labeled DPs present:'' +#+ shows_nl +@+ s);
     check_model_lab lR R
       <+? (\<lambda>s. ''problem with labeled TRS:'' +#+ shows_nl +@+ s);
     check_sl_only_lab_and_decr lge LS R lR 
       <+? (\<lambda>s. s +@+ shows_nl)
   })"

definition
  check_sem_lab_tt ::
    "shows check \<Rightarrow> ((('f,'l)lab,'v)rules \<Rightarrow> shows check) \<Rightarrow> 
     ('f::show,'l::show,'v::show)sl_check2
     \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> (('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> ('l + ('f, 'l) lab list) \<Rightarrow> bool) \<Rightarrow> 
     (('f,'l)lab \<Rightarrow> nat \<Rightarrow> 'l + ('f,'l)lab list \<Rightarrow> bool) \<Rightarrow>
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check"
where
  "check_sem_lab_tt valid check_decr check_lab check_model_lab lge LS R lR \<equiv> do {
     valid;
     do {
     check_decr lR;
     check_model_lab lR R;
     check_sl_only_lab_and_decr lge LS R lR
     } <+? (\<lambda>s. ''problem with labeled TRS:'' +#+ shows_nl +@+ s)
   }"


definition
  check_sem_lab_rel_tt ::
    "shows check \<Rightarrow> ((('f,'l)lab,'v)rules \<Rightarrow> shows check) \<Rightarrow> 
     ('f::show,'l::show,'v::show)sl_check2
     \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> 
     (('f,'l)lab \<Rightarrow> nat \<Rightarrow> 'l + ('f,'l)lab list \<Rightarrow> bool) \<Rightarrow>
     (('f,'l)lab,'v)reltrsL \<Rightarrow> 
     (('f,'l)lab,'v)reltrsL \<Rightarrow> 
     shows check"
where
  "check_sem_lab_rel_tt valid check_decr check_lab check_model_lab LS RS lRS \<equiv> do {
     valid; 
     do {
       check_decr (snd lRS);
       check_model_lab (fst lRS) (fst RS);
       check_model_lab (snd lRS) (snd RS)
     } <+? (\<lambda>s. ''problem with labeled TRS:'' +#+ shows_nl +@+ s)
   }"


locale sl_interpr_impl = 
  fixes check_lab :: "('f :: show,'l :: show,'v::show)sl_check2"
   and  check_model_lab :: "('f,'l,'v::show)sl_check2"
   and  check_decr :: "(('f,'l)lab,'v)rules \<Rightarrow> shows check"
   and  check_valid :: "shows check"
   and  C :: "'c set"
   and  c :: "'c"
   and  I :: "(('f,'l)lab,'c)inter"
   and  cge :: "'c \<Rightarrow> 'c \<Rightarrow> bool"
   and  lge :: "('f,'l)lab \<Rightarrow> nat \<Rightarrow> 'l + ('f,'l)lab list \<Rightarrow> 'l + ('f,'l)lab list \<Rightarrow> bool"
   and  L :: "(('f,'l)lab,'c,'l + ('f,'l)lab list)label"
   and  LS :: "(('f,'l)lab,'l + ('f,'l)lab list)labels" 
 assumes check_lab: "isOK(check_lab lR R) \<Longrightarrow> lab_trs I L label C (set R) \<subseteq> set lR"
   and   check_model_lab: "isOK(check_model_lab lR R) \<Longrightarrow> lab_trs I L label C (set R) \<subseteq> set lR \<and> qmodel I L label C cge (set R)"
   and   check_decr:  "isOK(check_decr lR) \<Longrightarrow> decr_of_ord (lge_to_lgr_rel lge LS) label LS \<subseteq> (CS (set lR) \<inter> decr_of_ord (lge_to_lgr_rel lge LS) label LS)^+"
   and   check_valid: "isOK(check_valid) \<Longrightarrow> sl_interpr_lab C c I cge lge L LS"

locale sl_interpr_root_impl = sl_interpr_impl check_lab check_model_lab check_decr check_valid C c I cge lge L LS
  for check_lab :: "('f :: show,'l :: show,'v::show)sl_check2"
   and  check_model_lab :: "('f,'l,'v::show)sl_check2"
   and  check_decr :: "(('f,'l)lab,'v)rules \<Rightarrow> shows check"
   and  check_valid :: "shows check"
   and  C :: "'c set"
   and  c :: "'c"
   and  I :: "(('f,'l)lab,'c)inter"
   and  cge :: "'c \<Rightarrow> 'c \<Rightarrow> bool"
   and  lge :: "('f,'l)lab \<Rightarrow> nat \<Rightarrow> 'l + ('f,'l)lab list \<Rightarrow> 'l + ('f,'l)lab list \<Rightarrow> bool"
   and  L :: "(('f,'l)lab,'c,'l + ('f,'l)lab list)label"
   and  LS :: "(('f,'l)lab,'l + ('f,'l)lab list)labels" +
 fixes  L' :: "(('f,'l)lab,'c,'l + ('f,'l)lab list)label"
   and  LS' :: "(('f,'l)lab,'l + ('f,'l)lab list)labels" 
   and  gen :: "'l \<Rightarrow> 'l list"
   and check_lab_all :: "('f :: show,'l :: show,'v::show)sl_check2"
  assumes gen: "\<And> f n l l'. lge f n (Inl l) (Inl l') \<Longrightarrow> l' \<in> set (gen l)"
   and check_lab_all: "isOK(check_lab_all lP P) \<Longrightarrow> lab_root_all_trs I L L' label lge C (set P) \<subseteq> set lP"
   and   check_valid2: "isOK(check_valid) \<Longrightarrow> sl_interpr_root_lab C c I cge lge L LS L' LS'" 

context sl_interpr_impl
begin 

lemma check_decr_rstep: assumes check: "isOK(check_decr lR)" shows "decr_of_ord (lge_to_lgr_rel lge LS) label LS \<subseteq> (rstep (set lR))^+"
proof -
  show ?thesis unfolding rstep_CC_CS 
    by (rule subset_trans[OF check_decr[OF check]], rule trancl_mono_set,
    rule subset_trans[OF _ subset_ctxt_closure], auto)
qed


lemma check_sem_lab_tt_sound: 
  "sound_lab_tt_impl (\<lambda> R m lR rm. rm = Suc m \<and> isOK(check_sem_lab_tt check_valid check_decr check_lab check_model_lab lge LS R lR))"
unfolding sound_lab_tt_impl_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume "isOK(check_sem_lab_tt check_valid check_decr check_lab check_model_lab lge LS R rR)" and SN: "decrSN (set rR) rm" and m: "rm = Suc m" 
  hence cond: "R_subset_lab_decr (lge_to_lgr_rel lge LS) LS (set R) (set rR) \<and> qmodel I L label C cge (set R) \<and> lab_trs I L label C (set R) \<subseteq> set rR \<and> 
    decr_of_ord (lge_to_lgr_rel lge LS) label LS \<subseteq> (rstep (set rR))^+" 
    and valid: "isOK(check_valid)" unfolding check_sem_lab_tt_def 
    using check_model_lab check_decr_rstep by (auto simp: check_model_lab)
  from check_valid[OF valid] interpret sl_interpr_lab C c I cge lge L LS .
  from cond lab_sound_tt SN show "decrSN (set R) m" unfolding sound_lab_tt_def by (auto simp: m)
qed

lemma check_sem_lab_rel_tt_sound: 
  "sound_rel_tt_impl (\<lambda> RS lRS. isOK(check_sem_lab_rel_tt check_valid check_decr check_lab check_model_lab LS RS lRS))"
unfolding sound_rel_tt_impl_def
proof (intro allI impI)
  fix R S rR rS
  assume "isOK(check_sem_lab_rel_tt check_valid check_decr check_lab check_model_lab LS (R,S) (rR,rS))" and SN: "rel_SN_alt (rstep (set rR), rstep (set rS))"
  hence mR: "qmodel I L label C cge (set R)" and R: "lab_trs I L label C (set R) \<subseteq> set rR" and
        mS: "qmodel I L label C cge (set S)" and S: "lab_trs I L label C (set S) \<subseteq> set rS" 
        and decr: "decr_of_ord (lge_to_lgr_rel lge LS) label LS \<subseteq> (rstep (set rS))^+" 
        and valid: "isOK(check_valid)" unfolding check_sem_lab_rel_tt_def 
    using check_model_lab check_decr_rstep by (auto simp: check_model_lab)
  from mR mS have model: "qmodel I L label C cge (set R \<union> set S)" unfolding qmodel_def by auto
  from check_valid[OF valid] interpret sl_interpr_lab C c I cge lge L LS .
  from quasi_rel_SN_lab_imp_rel_SN[OF _ model _ S R] decr SN check_valid[OF valid]
  show "rel_SN_alt (rstep (set R), rstep (set S))" unfolding rel_SN_conv by simp
qed

lemma check_sem_lab_proc_sound: 
  "sound_lab_proc_impl (\<lambda> P R m rmap lP lR rm. rm = Suc m \<and> isOK(check_sem_lab_proc check_valid check_decr check_lab check_model_lab lge LS P R lP lR))"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m lP lR rm
  assume check: "isOK(check_sem_lab_proc check_valid check_decr check_lab check_model_lab lge LS P R lP lR)"
    and  finite: "finite_lab_dpp (set lP, set lR) rm"
    and  m: "rm = Suc m"
  hence cond: "lab_trs I L label C (set P) \<subseteq> set lP \<and> lab_trs I L label C (set R) \<subseteq> set lR \<and> qmodel I L label C cge (set R) \<and> 
    R_subset_lab_decr (lge_to_lgr_rel lge LS) LS (set R) (set lR) \<and> 
          decr_of_ord (lge_to_lgr_rel lge LS) label LS \<subseteq> (rstep (set lR))^+" 
        and valid: "isOK(check_valid)" 
    using check_lab check_model_lab check_decr_rstep
    unfolding check_sem_lab_proc_def by auto
  from check_valid[OF valid] interpret sl_interpr_lab C c I cge lge L LS .
  from cond have "sem_lab_proc (set P) (set R) m (set lP) (set lR) rm" unfolding sem_lab_proc_def by (auto simp: m)
  with finite sem_lab_proc_sound show "finite_lab_dpp (set P, set R) m" unfolding sound_lab_proc_def by blast
qed
end

context sl_interpr_root_impl
begin
lemma check_sem_lab_root_proc_sound: 
  "sound_lab_proc_impl (\<lambda> P R m rmap lP lR rm. rm = Suc m \<and> isOK(check_sem_lab_root_proc check_valid check_decr check_lab_all check_model_lab lge LS rmap P R lP lR))"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m lP lR rm
  assume check: "isOK(check_sem_lab_root_proc check_valid check_decr check_lab_all check_model_lab lge LS (computeRuleMap R) P R lP lR)"
    and  finite: "finite_lab_dpp (set lP, set lR) rm"
    and  m: "rm = Suc m"
  show "finite_lab_dpp (set P, set R) m"
  proof (rule ccontr)
    assume nfinite: "\<not> ?thesis"
    then obtain s t \<sigma> where "min_lab_ichain (set P, set R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "lab_minimal_cond (set R) m s t \<sigma>" by auto
    from this[unfolded lab_minimal_cond_def, THEN spec[of _ 0], unfolded lab_minimal_term_def, THEN spec[of _ 0]]
    have "SN_elt (rstep (set R)) (t 0 \<cdot> \<sigma> 0)" by auto
    then obtain irm where irm: "computeRuleMap R = Some irm" using computeRuleMap_None_not_SN by (cases "computeRuleMap R", auto)
    from check[unfolded irm]
    have cond: "lab_root_all_trs I L L' label lge C (set P) \<subseteq> set lP \<and> lab_trs I L label C (set R) \<subseteq> set lR \<and> qmodel I L label C cge (set R) \<and> 
    R_subset_lab_decr (lge_to_lgr_rel lge LS) LS (set R) (set lR) \<and> 
          decr_of_ord (lge_to_lgr_rel lge LS) label LS \<subseteq> (rstep (set lR))^+" 
        and valid: "isOK(check_valid)" 
    using check_lab_all check_model_lab check_decr_rstep
    unfolding check_sem_lab_root_proc_def by auto
    from check[unfolded irm check_sem_lab_root_proc_def] 
    have vardef: "\<forall>(l, r)\<in>set P. \<not> is_var l \<and> \<not> is_var r \<and> \<not> defined (set R) (the (root r), num_args r)" 
      by (simp add: split_def check_no_defined_root_sound[OF irm])
    from check_valid2[OF valid] interpret sl_interpr_root_lab C c I cge lge L LS L' LS' .
    from cond vardef have "sem_lab_root_all_proc (set P) (set R) m (set lP) (set lR) rm" unfolding sem_lab_root_all_proc_def by (auto simp: m)
    with finite sem_lab_root_all_proc_sound
    have "finite_lab_dpp (set P, set R) m" unfolding sound_lab_proc_def by blast
    with nfinite  show False ..
  qed
qed
end



lemma sl_helper:  "{x | x. x < Suc n} = set [0..<Suc n]" by auto

definition model_check_decr where "model_check_decr \<equiv> \<lambda> _. OK"
definition model_check_valid where "model_check_valid \<equiv> OK"


(* the following implements the model-condition by choosing \<ge> as =, then everything is monotone and we do
   not have to worry about label-sets as there won't be decreasing rules at all *)
interpretation sl_finite_impl: sl_interpr_impl 
  "check_sl_lab_trs (sl_inter_to_inter sli) L [0 ..< Suc (get_largest_element sli)] model_cge"
  "check_sl_model_lab_trs (sl_inter_to_inter sli) L [0 ..< Suc (get_largest_element sli)] model_cge" 
  model_check_decr
  model_check_valid
  "{x | x . x < Suc (get_largest_element sli)}" 
  "get_largest_element sli" 
  "sl_inter_to_inter sli" 
  model_cge
  model_lge
  L  
  model_LS
  by (unfold_locales,
      simp only: sl_helper check_sl_lab_trs_sound,
      simp only: sl_helper check_sl_model_lab_trs_sound, simp,
      auto simp: decr_of_ord_def mem_def lge_to_lgr_rel_def lge_to_lgr_def Let_def Collect_def model_lge_def)


fun enum_vectors_nat :: "'c list \<Rightarrow> nat \<Rightarrow> 'c list list"
where "enum_vectors_nat C 0 = [[]]"
    | "enum_vectors_nat C (Suc n) = (let rec = enum_vectors_nat C n in concat (map (\<lambda> vec. map (\<lambda> c. c # vec) C) rec))" 


lemma enum_vectors_nat[simp]: 
  shows "set (enum_vectors_nat C n) = { xs. set xs \<subseteq> set C \<and> length xs = n}"
proof (induct n)
  case 0 
  show ?case by auto
next
  case (Suc n)
  have "set (enum_vectors_nat C (Suc n)) = 
    (\<Union>a\<in>{xs. set xs \<subseteq> set C \<and> length xs = n}. (\<lambda>c. c # a) ` set C)"
    unfolding enum_vectors_nat.simps Let_def
    by (simp add: Suc)
  also have "\<dots> = {xs. set xs \<subseteq> set C \<and> length xs = Suc n}" (is "?l = ?r")
  proof (intro set_eqI iffI)
    fix xs
    assume "xs \<in> ?l"
    thus "xs \<in> ?r" by auto
  next
    fix xs
    assume "xs \<in> ?r"
    thus "xs \<in> ?l" by (cases xs, auto)
  qed
  finally show ?case .
qed

declare enum_vectors_nat.simps[simp del]

definition qmodel_LS :: "('f \<times> nat) list \<Rightarrow> 'l list \<Rightarrow> ('f,'l list + 'b)labels" 
  where [code del]: "qmodel_LS sig LS \<equiv> \<lambda> f n. if (f,n) \<in> set sig then {Inl xs | xs. set xs \<subseteq> set LS \<and> length xs = n} else {Inl []}"
lemma [code]: "qmodel_LS sig LS = (\<lambda> f n. if (f,n) \<in> set sig then set (map Inl (enum_vectors_nat LS n)) else {Inl []})"
  unfolding qmodel_LS_def set_map enum_vectors_nat image_Collect by simp
definition qmodel_L :: "(('f,'l)lab \<times> nat) list \<Rightarrow> (('f,'l)lab,'c,'c list + 'a)label" 
  where "qmodel_L sig \<equiv> \<lambda> f cs. if (f,length cs) \<in> set sig then Inl cs else Inl []"
definition qmodel_cge where "qmodel_cge \<equiv> greater_eq :: nat \<Rightarrow> nat \<Rightarrow> bool"
definition qmodel_lge where "qmodel_lge f n \<equiv> \<lambda> l r. case (l,r) of 
  (Inl cs1, Inl cs2) \<Rightarrow> snd (pointwise_ext (\<lambda> x y :: nat. (y < x, y \<le> x)) cs1 cs2)
  | _ \<Rightarrow> False"

definition qmodel_check_interpretation :: "arithFun \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> shows check"
  where "qmodel_check_interpretation f n c \<equiv>
            let C = [0 ..< Suc c];
                css = enum_vectors_nat C n
             in check_allm 
                   (\<lambda> cs. check_allm 
                             (\<lambda> i. check_allm 
                                     (\<lambda> l. check (eval_arithFun (Suc c) f cs \<le> eval_arithFun (Suc c) f (cs [i := l])) (shows ''not monotone in '' +@+ shows (Suc i) +@+ shows ''. argument''))
                                     [cs ! i ..< Suc c])
                             [0 ..< n])
                css                  
                 "

fun qmodel_check_valid :: "('f :: show,'l :: show)sl_inter \<Rightarrow> shows check" 
  where "qmodel_check_valid (SL_Inter c ls) = check_allm (\<lambda> ((f,n),g). qmodel_check_interpretation g n c <+? 
            (\<lambda> e. shows ''problem in weak-monotonicity of interpretation of '' +@+ shows f +@+ shows_nl +@+ e )) ls"

lemma qmodel_check_valid: assumes valid: "isOK(qmodel_check_valid sli)"
  shows "cge_wm (sl_inter_to_inter sli) {x | x. x < Suc (get_largest_element sli)} qmodel_cge"
proof (cases sli)
  case (SL_Inter n fas)
  thus ?thesis
  proof (simp, unfold cge_wm qmodel_cge_def, intro allI impI, clarify)
    fix f bef c d aft
    assume mem: "set ([c,d] @ bef @ aft) \<subseteq> {u. u < Suc n}" and dc: "d \<le> c"
    let ?n = "Suc (length bef + length aft)"
    let ?i = "length bef"
    show "sl_inter_to_inter (SL_Inter n fas) f (bef @ d # aft) \<le> sl_inter_to_inter (SL_Inter n fas) f (bef @ c # aft)"
    proof (cases "lookup (f, ?n) fas")
      case None
      thus ?thesis by simp 
    next
      case (Some g)        
      let ?cs = "bef @ d # aft"
      let ?ccs = "bef @ c # aft"
      have elem: "((f, ?n), g) \<in> set fas" by (rule lookup_sound[OF Some])
      from elem valid[unfolded SL_Inter]
      have ok: "isOK(qmodel_check_interpretation g ?n n)" by auto
      from mem have mem: "bef @ d # aft \<in> set (enum_vectors_nat [0 ..< Suc n] ?n)" and c: "c < Suc n" by auto
      have i: "?i \<in> set [0 ..< ?n]" by auto
      from dc c have c: "c \<in> set [?cs ! ?i ..< Suc n]" by auto
      from ok[unfolded qmodel_check_interpretation_def Let_def isOK_simps, 
        THEN bspec[OF _ mem], THEN bspec[OF _ i], THEN bspec[OF _ c]]
      have "eval_arithFun (Suc n) g ?cs \<le> eval_arithFun (Suc n) g ?ccs" by auto
      thus ?thesis by (simp add: Some)
    qed
  qed
qed


definition check_decr_present_aux_1 :: "('f,'v \<times> nat)rules \<Rightarrow> 'v \<Rightarrow> 'f \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> ('f,'v \<times> nat) rule check"
where "check_decr_present_aux_1 R v f1 f2 n \<equiv> 
           let vs = map (\<lambda> n. Var (v,n)) [0 ..< n];
               rule = (Fun f1 vs, Fun f2 vs)
             in check (find (instance_rule rule) R \<noteq> None) rule"
        
lemma check_decr_present_aux_1: fixes R :: "('f,'v \<times> nat)rules"
  assumes check: "isOK(check_decr_present_aux_1 R v f1 f2 n)"
  shows "{(Fun f1 ts,Fun f2 ts) | ts. length ts = n} \<subseteq> CS (set R)"
proof -
  {
    fix ts :: "('f,'v \<times> nat) term list"
    assume len: "length ts = n"
    let ?n = "[0 ..< n]"
    let ?vs = "map (\<lambda> n. Var (v,n)) ?n"
    let ?rule = "(Fun f1 ?vs, Fun f2 ?vs)"
    have "find (instance_rule ?rule) R \<noteq> None" (is "?f \<noteq> None")
      using check[unfolded check_decr_present_aux_1_def Let_def] by auto
    then obtain rule where "?f = Some rule" by (cases ?f, auto)
    from find_SomeE[OF this] have rule: "rule \<in> set R" 
      and inst: "instance_rule ?rule rule" by auto
    from inst[unfolded instance_rule_def] obtain \<sigma> where
      id: "?rule = (fst rule \<cdot> \<sigma>, snd rule \<cdot> \<sigma>)" by auto
    obtain \<delta> where delta: "\<delta> = Subst (\<lambda> (_,i). ts ! i)" by auto
    obtain \<gamma> where gamma: "\<gamma> = \<sigma> \<circ>s \<delta>" by auto
    have ts_delta: "map (\<lambda> t. t \<cdot> \<delta>) ?vs = ts" unfolding delta
      by (rule nth_equalityI, auto simp add: len)
    hence "(Fun f1 ts, Fun f2 ts) = (Fun f1 ?vs \<cdot> \<delta>, Fun f2 ?vs \<cdot> \<delta>)" by auto
    with id have id: "(Fun f1 ts, Fun f2 ts) = (fst rule \<cdot> \<gamma>, snd rule \<cdot> \<gamma>)"
      by (auto simp: gamma)
    have "(Fun f1 ts, Fun f2 ts) \<in> CS (set R)"
      unfolding id by (auto simp: rule)
  }
  thus ?thesis by auto
qed

definition check_decr_present_aux_2 :: "('f,'v)rules \<Rightarrow> 'v \<Rightarrow> ('f \<times> 'f \<times> nat)list \<Rightarrow> ('f,'v \<times> nat)rule check"
where "check_decr_present_aux_2 R v req \<equiv> 
           let (add_nats :: ('f,'v)term \<Rightarrow> ('f,'v \<times> nat)term) = map_vars (\<lambda> v. (v, 0));
               R' = map (\<lambda> (l,r). (add_nats l, add_nats r)) R
             in check_allm (\<lambda> (f1 :: 'f,f2,n). check_decr_present_aux_1 R' v f1 f2 n) req"

lemma check_decr_present_aux_2: 
  fixes R :: "('f,'v)rules"
  assumes check: "isOK(check_decr_present_aux_2 R v req)"
  shows "{(Fun f1 ts, Fun f2 ts) | f1 f2 n ts. (f1,f2,n) \<in> set req \<and> length ts = n} \<subseteq> CS (set R)"
proof -
  {
    fix f1 f2 :: 'f and n :: nat and ts :: "('f,'v)term list"
    let ?add_nats = "map_vars (\<lambda> v. (v, 0)) :: ('f,'v)term \<Rightarrow> ('f,'v \<times> nat)term"
    let ?rem_nats = "map_vars (\<lambda> (v,n). v) :: ('f,'v \<times> nat)term \<Rightarrow> ('f,'v)term"
    let ?R = "map (\<lambda> (l,r). (?add_nats l, ?add_nats r)) R"
    let ?ts = "map ?add_nats ts"
    assume "(f1,f2,n) \<in> set req" and len: "length ts = n"
    from this check[unfolded check_decr_present_aux_2_def Let_def] 
    have "isOK(check_decr_present_aux_1 ?R v f1 f2 n)" by auto
    from check_decr_present_aux_1[OF this] len
    have "(Fun f1 ?ts, Fun f2 ?ts) \<in> CS (set ?R)" by auto
    hence "(?rem_nats (Fun f1 ?ts), ?rem_nats (Fun f2 ?ts)) \<in> (\<lambda> (l,r). (?rem_nats l, ?rem_nats r)) ` CS (set ?R)" (is "_ \<in> ?RR") by force
    hence mem: "(Fun f1 ts, Fun f2 ts) \<in> ?RR" unfolding map_vars.simps(2)[symmetric] map_vars_compose
      by (simp add: o_def map_vars_id[simplified id_def])
    then obtain l r where mem: "(l,r) \<in> CS (set ?R)"
      and "Fun f1 ts = ?rem_nats l" and "Fun f2 ts = ?rem_nats r"  by force
    with subst_closed_decompose[OF mem]  obtain l r \<sigma> where mem: "(l,r) \<in> set R"
      and l: "Fun f1 ts = ?rem_nats (?add_nats l \<cdot> \<sigma>)" and r: "Fun f2 ts = ?rem_nats (?add_nats r \<cdot> \<sigma>)" by force
    obtain \<delta> where delta: "\<delta> = Subst (\<lambda> v. ?rem_nats (Var (v,0) \<cdot> \<sigma>))" by auto
    {
      fix t
      have "?rem_nats (?add_nats t \<cdot> \<sigma>) = t \<cdot> \<delta>"
      proof (induct t)
        case (Var x)
        show ?case unfolding delta by (cases \<sigma>, auto)
      next
        case (Fun f ts)
        thus ?case by auto
      qed
    }
    with l r have "Fun f1 ts = l \<cdot> \<delta>" and "Fun f2 ts = r \<cdot> \<delta>" by auto
    with mem have "(Fun f1 ts, Fun f2 ts) \<in> CS (set R)" by auto
  }
  thus ?thesis by auto
qed

definition check_decr_present :: "('f \<times> nat)list \<Rightarrow> ('f \<Rightarrow> nat list \<Rightarrow> 'f) \<Rightarrow> 'v \<Rightarrow> nat \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v \<times> nat) rule check"
where "check_decr_present sig l v c R \<equiv>
   let C = [0 ..< Suc c];
       ls = \<lambda> (f,n). concat (map (\<lambda> cs. 
                concat (map (\<lambda> i. let ci = cs ! i in if ci < c then [(l f (cs[ i:= Suc ci]),l f cs,n)] else []) [0 ..< n])
                            ) (enum_vectors_nat C n))
       in check_decr_present_aux_2 R v (concat (map ls sig))"

(* soundness of check_decr_present is proven is three steps,
   first, that one increase one component of a vector by 1,
   second, that one can increase one component of a vector arbitrarily,
   third, that one can increase all components of the vector arbitrarily (pointwise_ext is covered)
*)

lemma check_decr_present_1: 
  assumes check: "isOK(check_decr_present sig l v c R)"
  and f: "(f,n) \<in> set sig"
  and i: "i < n"
  and d: "d < c"
  and cs: "set cs \<subseteq> {0 ..< Suc c}"
  and ts: "length ts = n"
  and lcs: "length cs = n"
  shows "(Fun (l f (cs[i := Suc d])) ts, Fun (l f (cs[i := d])) ts) \<in> CS (set R)"
proof -
  let ?C = "[0 ..< Suc c]"
  let ?ls = "\<lambda> (f,n). concat (map (\<lambda> cs. 
                concat (map (\<lambda> i. let ci = cs ! i in if ci < c then [(l f (cs[ i:= Suc ci]),l f cs,n)] else []) [0 ..< n])
                            ) (enum_vectors_nat ?C n))"
  let ?ll = "l f (cs[i := Suc d])"
  let ?lr = "l f (cs[i := d])"
  let ?req = "set (concat (map ?ls sig))"
  have "set (cs[i := d]) \<subseteq> {0 ..< Suc c}" using cs 
  proof (induct cs arbitrary: i)
    case (Cons e cs)
    from d have "d < Suc c" by simp
    with Cons show ?case by (cases i, auto)
  qed simp
  moreover have "\<exists>a\<in>{0..<n}.
       (?ll, ?lr, n) \<in> set (let ci = cs[i := d] ! a
              in if ci < c
                 then [(l f (cs[i := d, a := Suc ci]), l f (cs[i := d]), n)]
                 else [])"
    by (rule bexI[of _ i], auto simp: Let_def i lcs d)
  moreover have "length (cs[i := d]) = n" using lcs by simp
  ultimately have "(?ll,?lr,n) \<in> ?req"
    by (simp del: upt_Suc, intro bexI[OF _ f], simp del: upt_Suc, intro exI[of _ "cs[i := d]"], auto)
  thus ?thesis using check_decr_present_aux_2[OF check[unfolded check_decr_present_def Let_def]] ts
    unfolding Let_def by blast
qed

lemma check_decr_present_2:   
  assumes check: "isOK(check_decr_present sig Lab v c R)"
  and f: "(f,n) \<in> set sig"
  and i: "i < n"
  and d: "d < e"
  and e: "e \<le> c"
  and cs: "set cs \<subseteq> {0 ..< Suc c}"
  and ts: "length ts = n"
  and lcs: "length cs = n"
  shows "(Fun (Lab f (cs[i := e])) ts, Fun (Lab f (cs[i := d])) ts) \<in> (CS (set R) \<inter> (decr_of_ord (lge_to_lgr_rel qmodel_lge (qmodel_LS sig [0..<Suc c])) label (qmodel_LS sig [0..<Suc c])))^+" (is "_ \<in> (_ \<inter> ?Ord)^+")
proof -
  from d obtain diff where "e - d = Suc diff" by (cases "e - d", auto)
  with d e have e: "e = Suc (d + diff)" and diff: "Suc (d + diff) \<le> c" by auto
  have ord: "\<And> d e :: nat. \<lbrakk>d \<le> c; d > e\<rbrakk>  \<Longrightarrow> (Fun (Lab f (cs[i := d])) ts, Fun (Lab f (cs[i := e])) ts) \<in> ?Ord"
  proof -
    fix d e :: nat
    assume ed: "e < d" and dc: "d \<le> c"
    hence ec: "e < Suc c" and dc: "d < Suc c" by auto
    have csd: "set (cs[i := d]) \<subseteq> {0 ..< Suc c}"
      by (rule set_update_subsetI[OF cs], simp add: dc)
    have cse: "set (cs[i := e]) \<subseteq> {0 ..< Suc c}"
      by (rule set_update_subsetI[OF cs], simp add: ec)    
    have csde: "cs[i := d] \<noteq> cs[i := e]" (is "?d \<noteq> ?e")
    proof
      assume "?d = ?e"
      hence "?d ! i = ?e ! i" by simp
      thus False using i[unfolded lcs[symmetric]] ed by auto
    qed
    show "(Fun (Lab f (cs[i := d])) ts, Fun (Lab f (cs[i := e])) ts) \<in> ?Ord"
      unfolding decr_of_ord_def lge_to_lgr_rel_def lge_to_lgr_def Let_def qmodel_lge_def qmodel_LS_def
    proof (intro CollectI, rule exI[of _ f], rule exI[of _ "Inl (cs[i := d])"], rule exI[of _ "Inl (cs[i := e])"], rule exI[of _ ts],
      simp add: f ts lcs pointwise_ext_iff csd cse csde del: upt_Suc)
      show "\<forall> j < n. ?e ! j \<le> ?d ! j"
      proof (intro allI impI)
        fix j
        assume "j < n"
        with ed show "?e ! j \<le> ?d ! j"
          unfolding lcs[symmetric] by (cases "j = i", auto)
      qed
    qed
  qed
  from diff have "(Fun (Lab f (cs[i := Suc (d + diff)])) ts, Fun (Lab f (cs[i := d])) ts) \<in> (CS (set R) \<inter> ?Ord)^+"
  proof (induct diff)
    case 0
    hence dc1: "d < c" and dc2: "Suc d \<le> c" by auto
    from check_decr_present_1[OF check f i dc1 cs ts lcs]
    have one: "(Fun (Lab f (cs[i := Suc d])) ts, Fun (Lab f (cs[i := d])) ts) \<in> CS (set R)" (is "?pair \<in> ?CS") .
    have "?pair \<in> ?Ord" by (rule ord[OF dc2], simp)
    with one show ?case by force
  next
    case (Suc diff)
    hence rec: "(Fun (Lab f (cs[i := Suc (d + diff)])) ts, Fun (Lab f (cs[i := d])) ts) \<in> (CS (set R) \<inter> ?Ord)^+" (is "(?m,?r) \<in> _") by auto
    from Suc(2) have dc1: "Suc (d + diff) < c" and dc2: "Suc (Suc (d + diff)) \<le> c" by auto
    from check_decr_present_1[OF check f i dc1 cs ts lcs] 
    have cs: "(Fun (Lab f (cs[i := Suc (Suc (d + diff))])) ts, ?m) \<in> CS (set R)" (is "(?l,_) \<in> ?CS") .
    obtain rel where rel: "?CS \<inter> ?Ord = rel" by auto
    have "(?l,?m) \<in> ?Ord" by (rule ord[OF dc2], simp)
    with cs have "(?l,?m) \<in> ?CS \<inter> ?Ord" by auto
    with rec show ?case unfolding rel by auto 
  qed
  with e show ?thesis by simp
qed

(* TODO: move the following two lemmas to util (or search in standard library for something like this) *)
lemma take_drop_update_first: assumes "j < length ds" and "length cs = length ds"
  shows "(take j ds @ drop j cs)[j := ds ! j] = take (Suc j) ds @ drop (Suc j) cs"
using assms
proof (induct j arbitrary: ds cs)
  case 0
  then obtain d dds c ccs where ds: "ds = d # dds" and cs: "cs = c # ccs" by (cases ds, simp, cases cs, auto)
  show ?case unfolding ds cs by auto
next
  case (Suc j)
  then obtain d dds c ccs where ds: "ds = d # dds" and cs: "cs = c # ccs" by (cases ds, simp, cases cs, auto)
  from Suc(1)[of dds ccs] Suc(2) Suc(3) show ?case unfolding ds cs by auto
qed

lemma take_drop_update_second: assumes "j < length ds" and "length cs = length ds"
  shows "(take j ds @ drop j cs)[j := cs ! j] = take j ds @ drop j cs"
using assms
proof (induct j arbitrary: ds cs)
  case 0
  then obtain d dds c ccs where ds: "ds = d # dds" and cs: "cs = c # ccs" by (cases ds, simp, cases cs, auto)
  show ?case unfolding ds cs by auto
next
  case (Suc j)
  then obtain d dds c ccs where ds: "ds = d # dds" and cs: "cs = c # ccs" by (cases ds, simp, cases cs, auto)
  from Suc(1)[of dds ccs] Suc(2) Suc(3) show ?case unfolding ds cs by auto
qed
  
lemma check_decr_present: assumes check: "isOK(check_decr_present sig Lab v c R)"
  and f: "(f,n) \<in> set sig"
  and cs: "set cs \<subseteq> {0 ..< Suc c}"
  and ds: "set ds \<subseteq> {0 ..< Suc c}"
  and ts: "length ts = n"
  and lcs: "length cs = n"
  and gt: "fst (pointwise_ext (\<lambda> x y :: nat. (y < x, y \<le> x)) cs ds)"
  shows "(Fun (Lab f cs) ts, Fun (Lab f ds) ts) \<in> (CS (set R) \<inter> (decr_of_ord (lge_to_lgr_rel qmodel_lge (qmodel_LS sig [0..<Suc c])) label (qmodel_LS sig [0..<Suc c])))^+" (is "_ \<in> (_ \<inter> ?Ord)^+")
proof -
  note check = check_decr_present_2[OF check f _ _ _ _ ts]
  let ?CS = "CS (set R)"
  let ?R = "?CS \<inter> ?Ord"
  let ?C = "{0 ..< Suc c}"
  from gt[unfolded pointwise_ext_iff] lcs obtain i where
    lds: "length ds = n" and ids: "i < length ds" and ics: "i < length cs" and gt: "ds ! i < cs ! i"
    and ge: "\<And> i. i < length ds \<Longrightarrow> ds ! i \<le> cs ! i"  and i: "i < n" by auto
  let ?csds = "\<lambda> i. take i ds @ drop i cs"
  let ?l = "Fun (Lab f cs) ts"
  let ?r = "\<lambda> j. Fun (Lab f (?csds j)) ts"
  from lcs lds have lcsds: "\<And> j. length (?csds j) = n" by auto
  {
    fix j
    assume "j \<le> length ds"
    hence "(?l, ?r j) \<in> ?R^* \<and> set (?csds j) \<subseteq> ?C \<and> (j > i \<longrightarrow> (?l, ?r j) \<in> ?R^+)"
    proof (induct j)
      case 0
      show ?case by (simp add: cs)
    next
      case (Suc j)
      with lds
      have ind1: "(?l, ?r j) \<in> ?R^*" and ind2: "set (?csds j) \<subseteq> ?C" and ind3: "j > i \<longrightarrow> (?l, ?r j) \<in> ?R^+" and j: "j < n" by auto
      from j lcs have idd: "(?csds j)[ j := ds ! j] = ?csds (Suc j)" unfolding lds[symmetric] by (rule take_drop_update_first)
      from j lcs have idc: "(?csds j)[ j := cs ! j] = ?csds j" unfolding lds[symmetric] by (rule take_drop_update_second)
      from j cs lcs have csj: "cs ! j \<le> c" unfolding set_conv_nth by auto
      {
        assume "cs ! j > ds ! j"
        from check[OF j this csj ind2 lcsds] 
        have "(Fun (Lab f ((?csds j)[ j := cs ! j])) ts, Fun (Lab f ((?csds j)[ j := ds ! j])) ts) \<in> ?R^+" .
        hence "(?r j, ?r (Suc j)) \<in> ?R^+" by (simp only: idd idc)
      } note strict = this
      {
        assume "\<not> (cs ! j > ds ! j)"
        with j ge lds have "cs ! j = ds ! j" by force
        hence "(?csds j)[j := ds ! j] = (?csds j)[j := cs ! j]" by simp
        hence "?csds (Suc j) = ?csds j" unfolding idd idc .
        hence "(?r j, ?r (Suc j)) \<in> ?R^*" by simp
      } note non_strict = this
      from strict non_strict have non_strict: "(?r j, ?r (Suc j)) \<in> ?R^*" by (cases "cs ! j > ds ! j", auto)      
      with ind1 have "(?l, ?r (Suc j)) \<in> ?R^*" by auto
      moreover have "set (?csds (Suc j)) \<subseteq> ?C"
      proof 
        fix x
        assume "x \<in> set (?csds (Suc j))"
        hence "x \<in> set (take (Suc j) ds) \<or> x \<in> set (drop (Suc j) cs)" by auto
        thus "x \<in> ?C"
        proof
          assume "x \<in> set (drop (Suc j) cs)"
          with set_drop_subset[of "Suc j" cs] cs 
          show "x \<in> ?C" by auto
        next
          assume "x \<in> set (take (Suc j) ds)"
          with set_take_subset[of "Suc j" ds] ds
          show "x \<in> ?C" by auto
        qed
      qed
      moreover have "Suc j > i \<longrightarrow> (?l, ?r (Suc j)) \<in> ?R^+"
      proof
        assume sj: "Suc j > i"
        show "(?l, ?r (Suc j)) \<in> ?R^+"
        proof (cases "i = j")
          case False
          with sj have "j > i" by auto
          with ind3 non_strict show ?thesis by auto
        next
          case True
          from strict[OF gt[unfolded True]] ind1 show ?thesis by auto
        qed
      qed
      ultimately show ?case by blast
    qed
  }
  from this[of "length ds"] ids have "(?l, ?r (length ds)) \<in> ?R^+" by simp
  with lcs lds show ?thesis by auto
qed
      

definition qmodel_check_decr where "qmodel_check_decr sig v c \<equiv> \<lambda> lR. check_decr_present sig Lab v c lR 
    <+? (\<lambda> r. let display = map_vars (\<lambda> (_,n). (shows ''x'' +@+ shows n) []) in 
            shows ''decreasing rule '' +@+ shows_rule (display (fst r), display (snd r)) +@+ shows '' missing'')" 

lemma qmodel_check_decr: 
  assumes ok: "isOK(qmodel_check_decr sig v n lR)"
  shows "decr_of_ord (lge_to_lgr_rel qmodel_lge (qmodel_LS sig [0 ..< Suc n])) label (qmodel_LS sig [0 ..< Suc n])
   \<subseteq> (CS (set lR) \<inter> decr_of_ord (lge_to_lgr_rel qmodel_lge (qmodel_LS sig [0 ..< Suc n])) label (qmodel_LS sig [0 ..< Suc n]))^+" (is "?Decr \<subseteq> _")
proof -
  let ?g = "\<lambda> x y :: nat. (y < x, y \<le> x)"
  let ?C = "{0 ..< Suc n}"    
  have C: "?C = insert n {0 ..< n}" by auto
  let ?labs = "\<lambda> f ts. if (f, length ts) \<in> set sig then {Inl xs |xs. set xs \<subseteq> set [0..<Suc n] \<and> length xs = length ts} else {Inl []}"
  {
    fix l r :: "(('a,nat list)lab,'b)term"
    assume mem: "(l,r) \<in> ?Decr"
    then obtain f ts lab1 lab2 where l: "l = Fun (label f (length ts) lab1) ts" and r: "r = Fun (label f (length ts) lab2) ts" 
      and neq: "lab1 \<noteq> lab2" and lab1: "lab1 \<in> ?labs f ts" and lab2: "lab2 \<in> ?labs f ts" and ll: "(lab1,lab2) \<in> {(l,l') | l l'. case (l, l') of
      (Inl cs1, Inl cs2) \<Rightarrow>
      snd (pointwise_ext ?g cs1 cs2)
      | (Inl cs1, Inr ba) \<Rightarrow> False | (Inr bb, b) \<Rightarrow> False}" (is "_ \<in> ?ll")
      unfolding qmodel_LS_def lge_to_lgr_rel_def lge_to_lgr_def qmodel_lge_def decr_of_ord_def Let_def by auto
    note mem = mem[unfolded l r] 
    let ?labs = "?labs f ts"
    from lab1 obtain cs1 where cs1: "lab1 = Inl cs1" by (cases lab1, simp, cases "(f,length ts) \<in> set sig", auto) 
    from lab2 obtain cs2 where cs2: "lab2 = Inl cs2" by (cases lab2, simp, cases "(f,length ts) \<in> set sig", auto)
    have sig: "(f,length ts) \<in> set sig"
    proof (rule ccontr)
      assume "\<not> ?thesis"
      with lab1 lab2 neq show False unfolding cs1 cs2 by auto
    qed
    note mem = mem[unfolded cs1 cs2] 
    from ll neq
    have snd: "snd (pointwise_ext ?g cs1 cs2)" and neq: "cs1 \<noteq> cs2" unfolding cs1 cs2 by auto
    let ?pair = "(Fun (Lab f cs1) ts, Fun (Lab f cs2) ts)"
    from pointwise_ext_snd_neq_imp_fst[OF _ snd neq]
    have fst: "fst (pointwise_ext ?g cs1 cs2)" by auto
    from lab1[unfolded cs1] have cs1C: "set cs1 \<subseteq> ?C" and lcs1: "length cs1 = length ts" by (auto simp: sig)
    from lab2[unfolded cs2] have cs2C: "set cs2 \<subseteq> ?C" and lcs2: "length cs2 = length ts" by (auto simp: sig)
    from ok[unfolded qmodel_check_decr_def] 
    have "isOK(check_decr_present sig Lab v n lR)" by simp
    from check_decr_present[OF this sig cs1C cs2C refl lcs1 fst]
    have "(l,r) \<in> (CS (set lR) \<inter> ?Decr)^+" unfolding l r cs1 cs2 by simp
  }
  thus ?thesis ..
qed

lemma qmodel_SN: "SN (lge_to_lgr_rel qmodel_lge (qmodel_LS sig list) f n)" (is "SN ?r")
proof 
  fix S
  assume "\<forall> i. (S i, S (Suc i)) \<in> ?r"
  hence steps: "\<And> i. (S i, S (Suc i)) \<in> ?r" ..
  {
    fix i
    from steps[of i] obtain xs where Si: "S i = Inl xs" 
      unfolding lge_to_lgr_rel_def qmodel_lge_def lge_to_lgr_def Let_def 
      by (cases "S i", auto)
    hence "\<exists> xs. S i = Inl xs" ..
  }
  hence "\<forall> i. \<exists> xs. S i = Inl xs" ..
  from choice[OF this] obtain z where "\<And> i. S i = Inl (z i)" by auto
  with steps have steps: "\<And> i. (Inl (z i), Inl (z (Suc i))) \<in> ?r" by auto
  let ?g = "\<lambda> x y :: nat. (y < x, y \<le> x)"
  {
    fix i
    from steps[of i]
    have "snd (pointwise_ext ?g (z i) (z (Suc i)))" and "z i \<noteq> z (Suc i)"
      unfolding lge_to_lgr_rel_def qmodel_lge_def lge_to_lgr_def Let_def 
      by auto
    from pointwise_ext_snd_neq_imp_fst[OF _ this]
    have "fst (pointwise_ext ?g (z i) (z (Suc i)))" by auto
  } note steps = this
  have "SN {(ys, xs). fst (pointwise_ext ?g ys xs)}"
    by (rule pointwise_ext_SN_2[of ?g], auto simp: SN_iff_wf converse_def wf_less)
  with steps show False by auto
qed


lemma qmodel_SN_union: "SN (lge_to_lgr_rel qmodel_lge 
  (\<lambda>f n. qmodel_LS sig list f n \<union> qmodel_LS sig2 list2 f n)
 f n)" (is "SN ?r")
proof 
  fix S
  assume "\<forall> i. (S i, S (Suc i)) \<in> ?r"
  hence steps: "\<And> i. (S i, S (Suc i)) \<in> ?r" ..
  {
    fix i
    from steps[of i] obtain xs where Si: "S i = Inl xs" 
      unfolding lge_to_lgr_rel_def qmodel_lge_def lge_to_lgr_def Let_def 
      by (cases "S i", auto)
    hence "\<exists> xs. S i = Inl xs" ..
  }
  hence "\<forall> i. \<exists> xs. S i = Inl xs" ..
  from choice[OF this] obtain z where "\<And> i. S i = Inl (z i)" by auto
  with steps have steps: "\<And> i. (Inl (z i), Inl (z (Suc i))) \<in> ?r" by auto
  let ?g = "\<lambda> x y :: nat. (y < x, y \<le> x)"
  {
    fix i
    from steps[of i]
    have "snd (pointwise_ext ?g (z i) (z (Suc i)))" and "z i \<noteq> z (Suc i)"
      unfolding lge_to_lgr_rel_def qmodel_lge_def lge_to_lgr_def Let_def 
      by auto
    from pointwise_ext_snd_neq_imp_fst[OF _ this]
    have "fst (pointwise_ext ?g (z i) (z (Suc i)))" by auto
  } note steps = this
  have "SN {(ys, xs). fst (pointwise_ext ?g ys xs)}" 
    by (rule pointwise_ext_SN_2[of ?g], auto simp: SN_iff_wf converse_def wf_less)
  with steps show False by auto
qed

lemma qmodel_lge_wm: "lge_wm I (qmodel_L sig) C qmodel_cge qmodel_lge"
  unfolding lge_wm qmodel_cge_def qmodel_lge_def qmodel_L_def pointwise_ext_iff
proof (simp, intro allI impI, clarify)
  fix f bef and c d :: nat and  aft i
  assume dc: "d \<le> c"
  show "(bef @ d # aft) ! i \<le> (bef @ c # aft) ! i"
  proof (cases "i = length bef")
    case True thus ?thesis by (simp add: dc)
  next
    case False note oFalse = this
    thus ?thesis
    proof (cases "i < length bef")
      case True
      thus ?thesis by (simp add: nth_append)
    next
      case False 
      with oFalse have "i - length bef > 0" by auto
      then obtain j where "i - length bef = Suc j" by (cases "i - length bef", auto)
      thus ?thesis by (simp add: nth_append) 
    qed
  qed
qed

fun qmodel_gen :: "nat list \<Rightarrow> nat list list" where
  "qmodel_gen [] = [[]]"
| "qmodel_gen (x # xs) = (
    let ys = [0 ..< Suc x];
        zs = qmodel_gen xs in
    concat (map (\<lambda> y. map (op # y) zs) ys))"

fun qmodel_gen2 where "qmodel_gen2 (Inl l) = map Inl (qmodel_gen l)"

lemma qmodel_gen:  "set (qmodel_gen l) = {l'. snd (pointwise_ext (\<lambda> x y. (y < x, y \<le> x)) l l')}"
proof (induct l)
  case Nil 
  thus ?case by (simp add: pointwise_ext_iff, auto)
next
  case (Cons x l)
  have "(\<Union>a\<in>{0..<Suc x}. op # a `
        {l'. length l = length l' \<and> (\<forall>i<length l'. l' ! i \<le> l ! i)}) =
        {a # l' | a l'. a < Suc x \<and> length l = length l' \<and> (\<forall>i<length l'. l' ! i \<le> l ! i)}"
    by auto 
  also have "\<dots> = {a # l' | a l'. a \<le> x \<and> length l = length l' \<and> (\<forall>i<length l'. l' ! i \<le> l ! i)}" by auto
  also have "\<dots> = {l'. Suc (length l) = length l' \<and> (\<forall>i<length l'. l' ! i \<le> (x # l) ! i)}" (is "?l = ?r")
  proof (rule set_eqI, intro iffI)
    fix list
    assume "list \<in> ?l"
    then obtain a l' where cond: "list = a # l' \<and> a \<le> x \<and> length l = length l' \<and> (\<forall> i < length l'. l' ! i \<le> l ! i)" by auto
    have "\<forall> i < length list. list ! i \<le> (x # l) ! i" (is "\<forall> i < _. ?p i")
    proof (intro allI impI)
      fix i
      assume "i < length list"
      thus "?p i" using cond by (cases i, auto)
    qed
    with cond 
    show "list \<in> ?r" by auto
  next
    fix list
    assume "list \<in> ?r"
    hence cond: "Suc (length l) = length list \<and> (\<forall> i<length list. list ! i \<le> (x # l) ! i)" by auto 
    then obtain a l' where "list = a # l'" by (cases list, auto)
    with cond show "list \<in> ?l" by auto
  qed
  finally 
  show ?case 
    by (simp add: Cons del: upt_Suc, simp add: pointwise_ext_iff)
qed

interpretation q_sl_finite_impl: sl_interpr_root_impl 
  "check_sl_lab_trs (sl_inter_to_inter sli) (qmodel_L sig) [0 ..< Suc (get_largest_element sli)] qmodel_cge"
  "check_sl_model_lab_trs (sl_inter_to_inter sli) (qmodel_L sig) [0 ..< Suc (get_largest_element sli)] qmodel_cge" 
  "qmodel_check_decr sig v (get_largest_element sli)"
  "qmodel_check_valid sli"
  "{x | x . x < Suc (get_largest_element sli)}" 
  "get_largest_element sli" 
  "sl_inter_to_inter sli" 
  qmodel_cge
  qmodel_lge
  "qmodel_L sig"
  "qmodel_LS sig [0 ..< Suc (get_largest_element sli)]"
  "qmodel_L sig2" 
  "qmodel_LS sig2 [0 ..< Suc (get_largest_element sli)]"
  qmodel_gen
  "check_sl_lab_all_trs (sl_inter_to_inter sli) (qmodel_L sig) (qmodel_L sig2) [0 ..< Suc (get_largest_element sli)] qmodel_gen2"
proof (unfold_locales,
  simp only: sl_helper check_sl_lab_trs_sound,
  simp only: sl_helper check_sl_model_lab_trs_sound, simp)
  show "wf_inter (sl_inter_to_inter sli) {x | x. x < Suc (get_largest_element sli)}"
    by (rule wf_sl_inter)
next
  show "get_largest_element sli \<in> {x |x. x < Suc (get_largest_element sli)}" by (simp add: sl_helper)
next
  fix f n l
  show "label_decomp (label f n l) = (f, l)" by (rule label_decomp_label)
next
  show "wf_label (qmodel_L sig) (qmodel_LS sig [0 ..< Suc (get_largest_element sli)]) {x |x. x < Suc (get_largest_element sli)}"
    unfolding wf_label_def qmodel_L_def qmodel_LS_def 
    by (auto)
next
  show "wf_label (qmodel_L sig2) (qmodel_LS sig2 [0 ..< Suc (get_largest_element sli)]) {x |x. x < Suc (get_largest_element sli)}"
    unfolding wf_label_def qmodel_L_def qmodel_LS_def 
    by (auto)
next
  fix f n
  show "SN (lge_to_lgr_rel qmodel_lge (qmodel_LS sig [0 ..< Suc (get_largest_element sli)]) f n)"
    by (rule qmodel_SN)
next
  show "lge_wm (sl_inter_to_inter sli) (qmodel_L sig) {x |x. x < Suc (get_largest_element sli)} qmodel_cge qmodel_lge"
    by (rule qmodel_lge_wm)
next
  show "lge_wm (sl_inter_to_inter sli) (qmodel_L sig2) {x |x. x < Suc (get_largest_element sli)} qmodel_cge qmodel_lge"
    by (rule qmodel_lge_wm)
next
  assume "isOK (qmodel_check_valid sli)"  
  thus "cge_wm (sl_inter_to_inter sli) {x | x. x < Suc (get_largest_element sli)} qmodel_cge"
    by (rule qmodel_check_valid)
next
  fix f n 
  show "SN (lge_to_lgr_rel qmodel_lge
               (\<lambda>f n. qmodel_LS sig [0 ..< Suc (get_largest_element sli)] f
                       n \<union>
                      qmodel_LS sig2 [0 ..< Suc (get_largest_element sli)] f
                       n)
               f n)"
    by (rule qmodel_SN_union)
next
  fix f n x y z
  assume one: "qmodel_lge f n x y" and two: "qmodel_lge f n y z"
  from one obtain xx where x: "x = Inl xx" by (unfold qmodel_lge_def, cases x,simp, cases y, auto)
  from one obtain yy where y: "y = Inl yy" by (unfold qmodel_lge_def, cases x,simp, cases y, auto)
  from two obtain zz where z: "z = Inl zz" by (unfold qmodel_lge_def, cases y,simp, cases z, auto)
  from one two 
  show "qmodel_lge f n x z"
    unfolding qmodel_lge_def x y z
    using pointwise_snd_trans[of _ xx yy zz] by force
next
  fix f n x
  assume x: "x \<in> qmodel_LS sig2 [0..<Suc (get_largest_element sli)] f n"
  then obtain xx where x: "x = Inl xx" unfolding qmodel_LS_def by (cases "(f,n) \<in> set sig2", auto)
  show "qmodel_lge f n x x" unfolding x qmodel_lge_def using pointwise_ext_refl[of _ xx] by auto    
next
  fix f n l l'
  assume "qmodel_lge f n (Inl l) (Inl l')"
  hence "snd (pointwise_ext (\<lambda> x y. (y < x, y \<le> x)) l l')"  unfolding qmodel_lge_def
    by auto
  thus "l' \<in> set (qmodel_gen l)" unfolding qmodel_gen by simp
next
  fix lR
  assume "isOK (qmodel_check_decr sig v (get_largest_element sli) lR)"
  thus "decr_of_ord
          (lge_to_lgr_rel qmodel_lge
            (qmodel_LS sig [0..<Suc (get_largest_element sli)]))
          label (qmodel_LS sig [0..<Suc (get_largest_element sli)])
         \<subseteq> (CS (set lR) \<inter>
            decr_of_ord
             (lge_to_lgr_rel qmodel_lge
               (qmodel_LS sig [0..<Suc (get_largest_element sli)]))
             label (qmodel_LS sig [0..<Suc (get_largest_element sli)]))^+"
    by (rule qmodel_check_decr)
next
  fix lP P
  assume ok: "isOK
        (check_sl_lab_all_trs (sl_inter_to_inter sli) (qmodel_L sig)
          (qmodel_L sig2) [0..<Suc (get_largest_element sli)] qmodel_gen2 lP P)"
  show "lab_root_all_trs (sl_inter_to_inter sli) (qmodel_L sig)
        (qmodel_L sig2) label qmodel_lge
        {x |x. x < Suc (get_largest_element sli)} (set P)
       \<subseteq> set lP"
    unfolding sl_helper
  proof (rule check_sl_lab_all_trs_sound[OF ok])
    fix f n l l'
    assume lge: "qmodel_lge f n l l'"
    then obtain ll ll' where l: "l = Inl ll" and l': "l' = Inl ll'" unfolding qmodel_lge_def by (cases l, auto, cases l', auto)
    from lge[unfolded l l'] have "snd (pointwise_ext (\<lambda> x y. (y < x, y \<le> x)) ll ll')"  unfolding qmodel_lge_def
      by auto
    hence "ll' \<in> set (qmodel_gen ll)" unfolding qmodel_gen by simp
    thus "l' \<in> set (qmodel_gen2 l)" unfolding l l' by auto
  qed
qed 

  
type_synonym label_type = "nat list"

fun
  create_sem_lab_tt ::
    "('f::show,label_type)sl_inter \<Rightarrow> (('f,label_type)lab,nat,label_type + ('f,label_type)lab list)label + 'v \<Rightarrow> 
     (('f,label_type)lab,'v::show)rules \<Rightarrow> 
     (('f,label_type)lab,'v)rules \<Rightarrow> 
     shows check" 
where
  "create_sem_lab_tt sli (Inl L) r lr = do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     check_sem_lab_tt    
       model_check_valid
       model_check_decr
       (check_sl_lab_trs I L C model_cge) 
       (check_sl_model_lab_trs I L C model_cge) 
       model_lge
       model_LS
       r lr
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"
 | "create_sem_lab_tt sli (Inr v) r lr = do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let sig = funas_trs_list r;
     let I = sl_inter_to_inter sli;
     let L = qmodel_L sig;
     check_sem_lab_tt    
       (qmodel_check_valid sli)
       (qmodel_check_decr sig v c)
       (check_sl_lab_trs I L C qmodel_cge) 
       (check_sl_model_lab_trs I L C qmodel_cge) 
       qmodel_lge
       (qmodel_LS sig C)
       r lr
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"



fun
  create_sem_lab_rel_tt ::
    "('f::show,label_type)sl_inter \<Rightarrow> (('f,label_type)lab,nat,label_type + ('f,label_type)lab list)label + 'v \<Rightarrow> 
     (('f,label_type)lab,'v::show)reltrsL \<Rightarrow> 
     (('f,label_type)lab,'v)reltrsL \<Rightarrow> 
     shows check" 
where
  "create_sem_lab_rel_tt sli (Inl L) rs lrs = do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     check_sem_lab_rel_tt    
       model_check_valid
       model_check_decr
       (check_sl_lab_trs I L C model_cge) 
       (check_sl_model_lab_trs I L C model_cge) 
       model_LS
       rs lrs
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"
 | "create_sem_lab_rel_tt sli (Inr v) rs lrs = do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     let sig = funas_trs_list (fst rs @ snd rs);
     let L = qmodel_L sig;
     check_sem_lab_rel_tt    
       (qmodel_check_valid sli)
       (qmodel_check_decr sig v c)
       (check_sl_lab_trs I L C qmodel_cge) 
       (check_sl_model_lab_trs I L C qmodel_cge) 
       (qmodel_LS sig C)
       rs lrs
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"

fun funas_root_term :: "('f,'v)term \<Rightarrow> ('f \<times> nat) list" where 
   "funas_root_term (Fun f ts) = [(f,length ts)]"
 | "funas_root_term (Var _) = []"

fun funas_root_rule :: "('f,'v)rule \<Rightarrow> ('f \<times> nat) list" where 
  "funas_root_rule (l,r) = remdups (funas_root_term l @ funas_root_term r)"

fun funas_root_rules :: "('f,'v)rules \<Rightarrow> ('f \<times> nat)list" where
  "funas_root_rules r = remdups (concat (map funas_root_rule r))"

(* three variants: 
1)  model version (with arbitrary labeling function), (Inl)
2)  qmodel-version where all decreasing rules are put to R (also those for tuple-symbols), (Inr False)
3)  qmodel-version where no decreasing rules for tuple symbols are generated, but where for each labeled
    rhs  (Fun f_l ...) of P directly all smaller labels (Fun f_l' ...) with l \<ge> l' must be present (Inr True)

for 3 it is additionally demanded that one has standard DP-problem with tuple-symbols at roots of P
which is not required for 1 and 2
*)
fun
  create_sem_lab_proc ::
    "('f::show,label_type)sl_inter \<Rightarrow> (('f,label_type)lab,nat,label_type + ('f,label_type)lab list)label + ('v \<times> bool) \<Rightarrow> 
     (('f,label_type)lab,'v::show)rule_map \<Rightarrow> 
     (('f,label_type)lab,'v)rules \<Rightarrow> 
     (('f,label_type)lab,'v)rules \<Rightarrow> 
     (('f,label_type)lab,'v)rules \<Rightarrow> 
     (('f,label_type)lab,'v)rules \<Rightarrow> 
     shows check" 
where
  "create_sem_lab_proc sli (Inl L) rm r1 r2 r3 r4 = do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     check_sem_lab_proc 
       model_check_valid
       model_check_decr
       (check_sl_lab_trs I L C model_cge) 
       (check_sl_model_lab_trs I L C model_cge) 
       model_lge
       model_LS
       r1 r2 r3 r4
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"
| "create_sem_lab_proc sli (Inr (v,False)) rm r1 r2 r3 r4 = do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     let sig = funas_trs_list (r1 @ r2);
     let L = qmodel_L sig;
     check_sem_lab_proc 
       (qmodel_check_valid sli)
       (qmodel_check_decr sig v c)
       (check_sl_lab_trs I L C qmodel_cge) 
       (check_sl_model_lab_trs I L C qmodel_cge) 
       qmodel_lge
       (qmodel_LS sig C)
       r1 r2 r3 r4
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"
| "create_sem_lab_proc sli (Inr (v,True)) rm r1 r2 r3 r4 = do {
     let c = get_largest_element sli;
     let C = [0 ..< Suc c];
     let I = sl_inter_to_inter sli;
     let sig1 = list_union (funas_below_trs_list r1) (funas_trs_list r2);
     let sig2 = funas_root_rules r1;
     let L = qmodel_L sig1;
     let L' = qmodel_L sig2;
     check_sem_lab_root_proc 
       (qmodel_check_valid sli)
       (qmodel_check_decr sig1 v c)
       (check_sl_lab_all_trs I L L' C qmodel_gen2) 
       (check_sl_model_lab_trs I L C qmodel_cge) 
       qmodel_lge
       (qmodel_LS sig1 C)
       rm
       r1 r2 r3 r4
       <+? (\<lambda>s. shows sli +@+ shows_nl +@+ s)
   }"



lemma create_sem_lab_tt_sound:
  "sound_lab_tt_impl (\<lambda> R m lR rm. rm = Suc m \<and> isOK(create_sem_lab_tt sli vL R lR))"
proof (cases vL)
  case (Inl L)
  show ?thesis using sl_finite_impl.check_sem_lab_tt_sound
    unfolding sound_lab_tt_impl_def
    by (simp add: Let_def Inl, blast)
next
  case (Inr v)
  show ?thesis 
    unfolding sound_lab_tt_impl_def 
  proof (intro allI impI)
    fix R m rR rm
    assume "rm = Suc m \<and> isOK(create_sem_lab_tt sli vL R rR)" and "decrSN (set rR) rm"
    with q_sl_finite_impl.check_sem_lab_tt_sound[unfolded sound_lab_tt_impl_def, 
      THEN spec[of _ R], THEN spec[of _ m], THEN spec[of _ rR], THEN spec[of _ rm]]
    show "decrSN (set R) m"
      unfolding Inr by (simp add: Let_def) blast
  qed
qed


lemma create_sem_lab_proc_sound:
  "sound_lab_proc_impl (\<lambda> P R m rmap lP lR rm. rm = Suc m \<and> isOK(create_sem_lab_proc sli vL rmap P R lP lR))"
proof (cases vL)
  case (Inl L)
  show ?thesis
    using sl_finite_impl.check_sem_lab_proc_sound 
    by (simp add: Let_def Inl, blast)
next
  case (Inr vb)
  then obtain v b where vL: "vL = Inr (v,b)" by (cases vb, auto)
  show ?thesis 
    unfolding sound_lab_proc_impl_def 
  proof (intro allI impI)
    fix P R m rP rR rm
    assume cond: "rm = Suc m \<and> isOK(create_sem_lab_proc sli vL (computeRuleMap R) P R rP rR)" and fin: "finite_lab_dpp (set rP, set rR) rm"
    show "finite_lab_dpp (set P, set R) m"
    proof (cases b)
      case False
      with vL have vL: "vL = Inr (v,False)" by simp
      from cond fin q_sl_finite_impl.check_sem_lab_proc_sound[unfolded sound_lab_proc_impl_def, 
        THEN spec[of _ P], THEN spec[of _ R], THEN spec[of _ m], THEN spec[of _ rP], THEN spec[of _ rR], THEN spec[of _ rm]]
      show "finite_lab_dpp (set P, set R) m"
        unfolding vL by (simp add: Let_def) blast
    next
      case True
      with vL have vL: "vL = Inr (v,True)" by simp      
      from cond fin q_sl_finite_impl.check_sem_lab_root_proc_sound[unfolded sound_lab_proc_impl_def, 
        THEN spec[of _ P], THEN spec[of _ R], THEN spec[of _ m], THEN spec[of _ rP], THEN spec[of _ rR], THEN spec[of _ rm]]
      show "finite_lab_dpp (set P, set R) m"
        unfolding vL by (simp add: Let_def) blast
    qed
  qed
qed

lemma create_sem_lab_rel_tt_sound:
  "sound_rel_tt_impl (\<lambda> RS lRS. isOK(create_sem_lab_rel_tt sli vL RS lRS))"
proof (cases vL)
  case (Inl L)
  show ?thesis using sl_finite_impl.check_sem_lab_rel_tt_sound
    unfolding sound_rel_tt_impl_def
    by (simp add: Let_def Inl, blast)
next
  case (Inr v)
  show ?thesis 
    unfolding sound_rel_tt_impl_def 
  proof (intro allI impI)
    fix R S rR rS
    assume "isOK(create_sem_lab_rel_tt sli vL (R,S) (rR, rS))" and "rel_SN_alt (rstep (set rR), rstep (set rS))"
    with q_sl_finite_impl.check_sem_lab_rel_tt_sound[unfolded sound_rel_tt_impl_def,
      THEN spec[of _ R], THEN spec[of _ S], THEN spec[of _ rR], THEN spec[of _ rS]]
    show "rel_SN_alt (rstep (set R), rstep (set S))"
      unfolding Inr by (simp add: Let_def) 
  qed
qed

end
